/* loop.c - This is a quick-and-dirty derivative of latticeg.c to
      list all possible loops in a greechie diagram. */

/*****************************************************************************/
/* latticeg.c - Lattice tester for orthomodular lattice algebras             */
/*                                                                           */
/*       Copyright (C) 2003  NORMAN D. MEGILL  <nm@alum.mit.edu>             */
/*             Copyright terms:  GNU General Public License                    */
/*****************************************************************************/
/*34567890123456 (79-character line to adjust text window width) 678901234567*/

#define VERSION "1.6 27-Sep-09"

#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <time.h>
#include <ctype.h>

/* Largest lattice size plus 1 */
#define MAX_NODES 101
/* Binary infix operator list - i is only for Polish conversion */
#define BIN_OPERS "^v#OI2345=<>["
/* MAX_BIN_OPERS is the length of BIN_OPERS */
#define MAX_BIN_OPERS 13
/* Negation prefix operator */
#define NEG_OPER '-'
#define UNIV_IMPL 'i'
/* List of legal variable names - note v,i,o skipped because of operators */
#define VAR_LIST "abcdefghjklmnpqrstuwxyz"
/* MAX_VARS is the length of VAR_LIST */
#define MAX_VARS 23
/* MAX_STACK is length of longest formula expressed in Polish notation */
#define MAX_STACK 10000
/* Largest number of user hypotheses */
#define MAX_HYPS 50
/* Largest number of lattices */
#define MAX_LATTICES 1000

/* Mapping for Greechie diagram atoms */
#define ATOM_MAP "123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrs" \
    "tuvwxyz!\"#$%&'()*-/:;<=>?@[\\]^_`{|}~"
/* Maximum number of blocks */
#define MAX_BLOCKS 64
/* Minimum block size - only 3 and 4 are implemented */
#define MIN_BLOCK_SIZE 3
/* Maximum block size - only 3 and 4 are implemented */
#define MAX_BLOCK_SIZE 4

/******************** Start of string handling prototypes ********************/
typedef char* vstring;

/* String assignment - MUST be used to assign vstrings */
void let(vstring *target,vstring source);
/* String concatenation - last argument MUST be NULL */
vstring cat(vstring string1,...);

/* Emulate BASIC linput statement; returns NULL if EOF */
/* Note that linput assigns target string with let(&target,...) */
  /*
    BASIC:  linput "what";a$
    c:      linput(NULL,"what?",&a);

    BASIC:  linput #1,a$                        (error trap on EOF)
    c:      if (!linput(file1,NULL,&a)) break;  (break on EOF)

  */
vstring linput(FILE *stream,vstring ask,vstring *target);

/* Emulation of BASIC string functions */
vstring seg(vstring sin, long p1, long p2);
vstring mid(vstring sin, long p, long l);
vstring left(vstring sin, long n);
vstring right(vstring sin, long n);
vstring edit(vstring sin, long control);
vstring space(long n);
vstring string(long n, char c);
vstring chr(long n);
vstring xlate(vstring sin, vstring control);
vstring date(void);
vstring time_(void);
vstring num(double x);
vstring num1(double x);
vstring str(double x);
long len(vstring s);
long instr(long start, vstring sin, vstring s);
long ascii_(vstring c);
double val(vstring s);
/* Emulation of PROGRESS string functions added 11/25/98 */
vstring entry(long element, vstring list);
long lookup(vstring expression, vstring list);
long numEntries(vstring list);
long entryPosition(long element, vstring list);
/* Print to log file as well as terminal if fplog opened */
void print2(char* fmt,...);
FILE *fplog = NULL;
/* Opens files with error message; opens output files with
   backup of previous version.   Mode must be "r" or "w". */
FILE *fSafeOpen(vstring fileName, vstring mode);


/******* Special pupose routines for better
      memory allocation (use with caution) *******/
/* Make string have temporary allocation to be released by next let() */
/* Warning:  after makeTempAlloc() is called, the vstring may NOT be
   assigned again with let() */
void makeTempAlloc(vstring s);   /* Make string have temporary allocation to be
                                    released by next let() */
/* Remaining prototypes (outside of mmvstr.h) */
char *tempAlloc(long size);     /* String memory allocation/deallocation */

#define MAX_ALLOC_STACK 100
int tempAllocStackTop=0;        /* Top of stack for tempAlloc functon */
int startTempAllocStack=0;      /* Where to start freeing temporary allocation
                                    when let() is called (normally 0, except in
                                    special nested vstring functions) */
char *tempAllocStack[MAX_ALLOC_STACK];

/* Bug check error */
void bug(int bugNum);

/******************** End of string handling prototypes **********************/

/******************** Prototypes *********************************************/
void usersGreechieDiagrams(void);
void printhelp(void);
void testAllLattices(void);
void findCommutingPairs(void);
void init(void);
void initLattice(long latticeNum);

void greechie3(vstring name, vstring glattice);
void loop(vstring sofar);
unsigned char test(void);
vstring unabbreviate(vstring eqn);
vstring subformulaList(vstring eqn);
vstring toPolish(vstring equation);
vstring fromPolish(vstring polishEqn);
unsigned char eval(vstring trialEqn);
unsigned char fastEval(vstring trialEqn, long startChar);
vstring subformula(vstring eqn);
long subformulaLen(vstring eqn, long startChar);
unsigned char lookupCompl(unsigned char arg);
unsigned char lookupBinOp(unsigned char operation, unsigned char arg1,
    unsigned char arg2);
vstring unPrintableString(vstring sin);
vstring printableString(vstring sin);

/******************** Global variables ***************************************/

/* For loop.c */
#define MAX_LOOP_SIZE 100
vstring loops = ""; /* All loops so far for duplicate detection */

vstring nodeList[MAX_NODES];
vstring sup[MAX_NODES][MAX_NODES]; /* Supremum (disjunction) table */
vstring nodeNames = ""; /* List of node names in nodeList order */
long nodes;
vstring latticeName = ""; /* Name of lattice being worked with */
vstring equation = ""; /* Expression being tested */
vstring hypList[MAX_HYPS]; /* List of user's hypotheses (0 or more) */
vstring polHypList[MAX_HYPS]; /* List of user's hypotheses in Polish */
long hypotheses;
vstring conclusion = ""; /* User's conclusion to test */
vstring polConclusion = ""; /* Conclusion in Polish */
vstring failingAssignment = ""; /* Assignment that violated lattice */
long userArg; /* -1 = test all; n = test only nth lattice */
long startArg = 0; /* Lattice # to start from */
long endArg = 0; /* Lattice # to end at */
/* Special for Greechie diagram program latticeg.c */
vstring lletterlist = ""; /* Node names for atoms */
vstring uletterlist = ""; /* Node names for complemented atoms */
long glatticeNum;
long glatticeCount;
vstring greechieStmt = ""; /* Greechie lattice C statement */
char showGreechie = 0; /* Set to 1 by -g option */
char removeLegs = 0; /* Set to 1 by -l option */
FILE *fp = NULL;  /* Input file of greechie diagrams */
vstring *fileLattice; /* Storage of input file */
long fileLattices = 0; /* Number of input file diagrams */
  long block[MAX_BLOCKS + 1][MAX_BLOCK_SIZE + 1];
  long blockSize[MAX_BLOCKS + 1];
  long block4Offset[MAX_BLOCKS + 1];
  long blocks;


unsigned char negMap[256]; /* Map of negatives for speedup */
unsigned char opMap[256]; /* Map of operators for speedup */
unsigned char nodeNameMap[256]; /* Map from node name to lattice entry for speedup */

/* Binary operation table for fast lookup */
unsigned char binOpTable[MAX_BIN_OPERS][MAX_NODES][MAX_NODES];

/* Special flags for commuting pair option */
char commuteOption; /* Flag that user specified -c program option */
char OADoneFlag; /* Flag for last lattice passing the OA law */
char OMDoneFlag; /* Flag for last lattice passing the OM law */

/* For -v (show lattice points visited) option */
char showVisits = 0; /* Set to 1 to show all
                        lattice points visited by failure.  Useful for
                        identifying unnecessary (redundant) lattice points. */
char visitFlag = 0; /* Used internally when showVisits is 1 */
vstring visitList = ""; /* List of nodes visited */

/* For -f (show all failures) option */
char showAllFailures = 0;  /* Set to 1 to show
                        all possible lattice failures.  Useful for comparing
                        lattice behavior of two different equations. */
void usersGreechieDiagrams(void)
{
  /*** Start of user's Greechie diagrams ***/
  /* The first argument of greechie3, if non-blank, is the lattice
     name - otherwise a default name is generated */
  /* The second argument uses the matrix notation in
     Kalmbach, "Orthomodular Lattices", p. 319, with the matrix
     represented as a comma-separated linear string. */

  greechie3("L42 (OM, OA)",
  "1 2 3  3 4 5  5 6 7  7 8 9  9 10 11  11 12 13  13 14 1"
  "  3 15 10  6 16 13  8 17 18  13 19 18  3 20 17");
  /* Alternate numbering for L42 (in oapass.oml): */
  /*
  "1 2 3  1 4 5  1 6 7  1 8 9  2 10 11  4 12 13  6 14 15"
    "  8 16 17  10 12 14  11 16 18  13 16 19  15 16 20");
  */

  greechie3("L40 (OM, OA)",
  "1 2 11  1 3 12  1 4 13  5 6 7  8 9 10  2 5 14"
  "  2 8 15  3 6 16  3 9 17 4 7 18  4 10 19");

  greechie3("Mayet Fig. 5 (OM, OA)",
   "1 2 3  1 4 7  2 5 8  3 6 9  7 12 14  8 10 12  8 11 13  9 13 16  14 15 16");

  greechie3("L38 (OM, non-OA)",
      "1,2,3,"
      "3,4,5,"
      "5,6,7,"
      "7,8,9,"
      "9,10,11,"
      "11,12,13,"
      "13,14,1,"
      "12,15,4,"
      "15,16,17,"
      "17,18,6");

  greechie3("L36 (OM, non-OA)",
  " 1 2 11  1 3 12  4 5 6  4 7 8  5 9 10  1 6 13"
  "  2 7 14  2 9 15  3 8 16  3 10 17");

  greechie3("L36b (OM, non-OA)",
  "1 2 3  1 4 5  2 6 7  3 8 9  4 10 11  5 12 13"
  "  6 10 14  7 12 15 8 11 16  9 13 14  15 16 17");

  greechie3("L$ (OM, non-OA)",
  "1 2 3  3 4 5  5 6 7  7 8 9  9 10 11  11 12 1  10 13 4");

  /*** End of user's Greechie diagrams ***/
} /* usersGreechieDiagrams() */



/******************** Main program *******************************************/

int main(int argc, char *argv[])
{

  vstring str1 = "";
  vstring str2 = "";
  long i, j, argOffset;
  char argOffsetChanged;
  vstring contLine = ""; /* Continuation line */
  char printLatticeFlag = 0;

  /* argc is the number of arguments; argv points to array containing them */

  /* Line continuation for DOS: "$x" means get argument x from prompt */
  for (i = 1; i < argc; i++) {
    let(&str1, argv[i]);
    if (str1[0] == '$') {
      if (str1[0] == '\'' || str1[0] == '\"') {
        /* Strip quotes */
        let(&str1, seg(str1, 2, strlen(str1) - 1));
      }
      /* Strip off successive "$x"s and replace with continuation lines */
      while (str1[0] == '$') {
        let(&str2, cat(left(str1, 2), "> ", NULL));
        linput(NULL, str2, &contLine);
        let(&str1, cat(right(str1, 3), contLine, NULL));
      }
      /* Note that we'll just give up and not free the memory argv[i] points
         to; it seems the safest thing to do since it may not be a real malloc
         object depending on compiler */
      /* To be extra safe we might want to clone the argv pointer array first
         before reassigning it; maybe in a future version */
      argv[i] = str1;
      str1 = ""; /* Relinquish string ownership to argv[i] (instead of let()) */
    }
  }
  let(&str1, ""); /* Deallocate string */
  let(&str2, ""); /* Deallocate string */
  let(&contLine, ""); /* Deallocate string */

  if (argc <= 1) {
    /* No arguments means print help */
    printhelp();
    return 0;
  }

  /* Get command line options */
  /* Important: none of the options have the syntax of a legal wff.   This
     should be the case for any future options as well. */
  userArg = 0;
  commuteOption = 0;
  argOffset = 0;
  argOffsetChanged = 1;
  while (argOffsetChanged) {
    argOffsetChanged = 0;
    if (argc - argOffset > 1 && !strcmp(argv[1 + argOffset], "--help")) {
      printhelp();
      return 0;
    }
    if (argc - argOffset > 1 && !strcmp(argv[1 + argOffset], "-a")) {
      argOffset++;
      argOffsetChanged = 1;
      userArg = -1;
    }
    if (argc - argOffset > 1 && !strcmp(argv[1 + argOffset], "-c")) {
      argOffset++;
      argOffsetChanged = 1;
      commuteOption = 1;
      /* This is not implement for latticeg.c, only lattice.c */
      print2("-c is implemented in lattice.c only\n");
      return 0;
    }
    if (argc - argOffset > 1 && !strcmp(argv[1 + argOffset], "-v")) {
      argOffset++;
      argOffsetChanged = 1;
      showVisits = 1;
    }
    if (argc - argOffset > 1 && !strcmp(argv[1 + argOffset], "-f")) {
      argOffset++;
      argOffsetChanged = 1;
      showAllFailures = 1;
    }
    if (argc - argOffset > 1 && !strcmp(argv[1 + argOffset], "-g")) {
      argOffset++;
      argOffsetChanged = 1;
      showGreechie = 1;
    }
    if (argc - argOffset > 1 && !strcmp(argv[1 + argOffset], "-l")) {
      argOffset++;
      argOffsetChanged = 1;
      removeLegs = 1;
    }
    if (argc - argOffset > 1 && !strcmp(argv[1 + argOffset], "-i")) {
      if (argc - argOffset > 2) {
        /* Open the output file */
        fp = fopen(argv[2 + argOffset], "r");
        if (fp == NULL) {
          print2("?Error: Couldn't open the file \"%s\".\n",
              argv[2 + argOffset]);
          return 0;
        }
      } else {
        print2("?Error: No input file specified.\n");
        return 0;
      }
      argOffset += 2;
      argOffsetChanged = 1;
    }
    if (argc - argOffset > 1 && !strcmp(argv[1 + argOffset], "-o")) {
      if (argc - argOffset > 2) {
        if (fplog != NULL) {
          print2("?Error: Cannot specify more than one output file.\n");
          return 0;
        }
        /* Open the output file */
        fplog = fSafeOpen(argv[2 + argOffset], "w");
        if (fplog == NULL) {
          return 0;
        }
        /* Disable buffering of the output file so that all partial results
           will be there in case a run is aborted before completion */
        setbuf(fplog, NULL);
      } else {
        print2("?Error: No output log file specified.\n");
        return 0;
      }
      argOffset += 2;
      argOffsetChanged = 1;
    }
    if (argc - argOffset > 1 && !strcmp(argv[1 + argOffset], "--o")) {
      if (argc - argOffset > 2) {
        if (fplog != NULL) {
          print2("?Error: Cannot specify more than one output file.\n");
          return 0;
        }
        /* Open the output file - append mode */
        fplog = fopen(argv[2 + argOffset], "a");
        if (fplog == NULL) {
          print2("?Error: Couldn't open the file \"%s\".\n",
              argv[2 + argOffset]);
          return 0;
        }
        /* Disable buffering of the output file so that all partial results
           will be there in case a run is aborted before completion */
        setbuf(fplog, NULL);
      } else {
        print2("?Error: No output log file specified.\n");
        return 0;
      }
      argOffset += 2;
      argOffsetChanged = 1;
    }
    if (argc - argOffset > 1 && !strcmp(argv[1 + argOffset], "-n")) {
      if (argc - argOffset > 2) userArg = val(argv[2 + argOffset]);
      if (userArg <= 0  || strcmp(argv[2 + argOffset], str(userArg))) {
        print2("?Error: Expected positive integer after -n\n");
        return 0;
      }
      argOffset += 2;
      argOffsetChanged = 1;
    }
    if (argc - argOffset > 1 && !strcmp(argv[1 + argOffset], "-s")) {
      if (argc - argOffset > 2) startArg = val(argv[2 + argOffset]);
      if (startArg <= 0  || strcmp(argv[2 + argOffset], str(startArg))) {
        print2("?Error: Expected positive integer after -s\n");
        return 0;
      }
      argOffset += 2;
      argOffsetChanged = 1;
    }
    if (argc - argOffset > 1 && !strcmp(argv[1 + argOffset], "-e")) {
      if (argc - argOffset > 2) endArg = val(argv[2 + argOffset]);
      if (endArg <= 0  || strcmp(argv[2 + argOffset], str(endArg))) {
        print2("?Error: Expected positive integer after -e\n");
        return 0;
      }
      argOffset += 2;
      argOffsetChanged = 1;
    }
    if (argc - argOffset > 1 && !strcmp(argv[1 + argOffset], "-p")) {
      if (argc - argOffset > 2) userArg = val(argv[2 + argOffset]);
      if (userArg <= 0  || strcmp(argv[2 + argOffset], str(userArg))) {
        print2("?Error: Expected positive integer after -p\n");
        return 0;
      }
      printLatticeFlag = 1;
      argOffset += 2;
      argOffsetChanged = 1;
    }
  }

  /* loop.c */
  if (argc - argOffset != 1) {
    print2("?You have specified unknown options.\n");
    exit(0);
  }

  init(); /* One-time initialization */

  if (printLatticeFlag) {
    /* Print lattice for user */
    i = 0;
    if (fp != NULL) i = i + 2;
    if (fplog != NULL) i = i + 2;
    if (argc != 3 + i) {
      print2(
          "?Error: Only the -i and -o (or --o) options may be used with -p\n");
      return 0;
    }
    initLattice(userArg);
    if (!nodes && userArg > 0) {
      print2("?Error: Lattice %ld does not exist\n", userArg);
      return 0;
    }
    print2("Name of lattice %ld: %s\n", userArg, latticeName);
    print2("Number of nodes: %ld\n", nodes);
    print2("Greechie diagram:\n");
    print2(" %s\n", greechieStmt);
    print2("The lattice nodes map to Greechie atoms: a=1, b=2, etc.\n");
    print2("The lattice node names are capitalized to indicate complement.\n");
    print2("\n");
    print2("  Node  Nodes directly below it\n");
    print2("  ----  -----------------------\n");
    for (i = 1; i <= nodes; i++) {
      let(&str1, right(nodeList[i], 2));
      for (j = strlen(str1); j > 1; j--) {
        let(&str1, cat(left(str1, j - 1), ", ", right(str1, j), NULL));
      }
      let(&str1, cat("  ", left(nodeList[i], 1), "     ",
          str1, NULL));
      print2("%s\n", printableString(str1));
    }
    return 0;
  } /* if printLatticeFlag */

  /* Assign user's hypotheses and conclusion */
  hypotheses = argc - 2 - argOffset;
  for (i = 1; i < argc - argOffset; i++) {
    let(&str2, argv[i + argOffset]);
    /* Strip quotes */
    for (j = 0; j < strlen(str2); j++) {

      if (str2[j] == '\'' || str2[j] == '\"') {
        let(&str2, cat(left(str2, j), right(str2, j + 2), NULL));
      }
    }
    let(&str2, edit(str2, 1 + 2 + 4)); /* Strip spaces & garbage chars */
    if (i != argc - 1 - argOffset) {
      /* Assign hypothesis */
      let(&(hypList[i - 1]), str2);
    } else {
      /* Assign conclusion */
      let(&conclusion, str2);
    }
  }
  let(&str2, ""); /* Deallocate */

  /****** loop.c ******/
  if (1) goto skip_equation;

  /* Repeat the hypotheses and conclusion for user verification */
  for (i = 0; i < hypotheses; i++) {
    if (i > 0) {
      print2("& ");
    }
    print2("%s ", hypList[i]);
  }
  if (hypotheses > 0) {
    print2("=> ");
  }
  print2("%s\n", conclusion);

  /* Convert to Polish */
  for (i = 0; i < hypotheses; i++) {
    /* Add outer parentheses */
    let(&(hypList[i]), cat("(", hypList[i], ")", NULL));
    let(&(polHypList[i]), "");
    polHypList[i] = toPolish(hypList[i]);
  }
  /* Add outer parentheses */
  let(&conclusion, cat("(", conclusion, ")", NULL));
  let(&polConclusion, "");
  polConclusion = toPolish(conclusion);

  /****** loop.c ********/
 skip_equation:

  if (!commuteOption) {
    testAllLattices(); /* Run the program */
  } else {
    /* Run -c special option to find potentially commuting pairs  */
    findCommutingPairs();
  }

  return 0;
} /* main */

void printhelp(void)
{
vstring a = "";

/* loop.c */
print2("\n");
print2("loop.c - Finds loops in Greechie diagrams\n");
print2("\n");
print2("Usage: loop -i <file1> -o <file2>\n");
print2("Notes:  <file1> must contain exactly one Greechie diagram\n");
print2(
"  It is recommended that <file2> be sorted by user to be more useful\n");
print2("\n");
print2("Example:  If <file1> contains \n");
print2("  '123,345,567,789,9AB,BC1,2E8,4FA,6DC,DEF.' (Peterson OML)\n");
print2("then one of the lines in <file2> will be:\n");
print2("  6 123,345,567,789,9AB,BC1.  2*E.8* 4*F.A* 6*D.C* D.E.F.\n");
print2("  a bbbbbbbbbbbbbbbbbbbbbbbb  cdcecd cdcecd cdcecd cecece\n");
print2("where\n");
print2("  a = loop size\n");
print2("  b = blocks around the outer loop\n");
print2("  c = remaining blocks\n");
print2("  d = * means the atom is connected to the loop\n");
print2("  e = . means the atom is not connected to the loop\n");
print2("\n");
print2(
   "Copyright (C) 2000 Norman D. Megill <nm@alum.mit.edu> Version %s\n",
   VERSION);
print2("License terms:  GPL http://www.fsf.org/copyleft/gpl.html\n");
if (1) return;

print2("\n");
print2("latticeg.c - Orthomodular Lattice Evaluator for Greechie Diagrams\n");
print2("Usage: lattice [options] <hyp> <hyp> ... <conclusion>\n");
print2("         options:\n");
print2("           -a - test all lattices (don't stop after first failure)\n");
print2(
"           -n <integer> - test only the program's <integer>th lattice\n");
print2(
"           -s <integer> - start at the program's <integer>th lattice\n");
print2(
"           -e <integer> - end at the program's <integer>th lattice\n");
print2("           -v - show all visits to lattice points in a failure\n");
print2("           -f - show all failures in failing lattice\n");
print2(
"           -g - categorize Greechie diagrams (legs, pass/fail, etc.)\n");
print2("           -l - remove legs from Greechie lattice before testing\n");
print2(
"           -i <file> - use Greechie lattices from <file> instead of the\n");
print2(
"                       built-in ones\n");
print2(
"           -o <file> - write output to <file> as well as terminal\n");
print2(
"           --o <file> - append output to <file> as well as terminal\n");
print2(
"       lattice -p <integer> - print the program's <integer>th lattice\n");
print2(
"           (you may use the -i and -o [or --o] options with -p)\n");
print2("       lattice --help (or no argument) - print this message\n");
print2("Notes: Only one of -a and -n should be used.  -s and -e are applied\n");
print2("before any others.\n");
print2("\n");
print2("\n");
print2("\n");
linput(NULL,"Press Enter to continue, q to quit...",&a);
if (toupper(a[0]) == 'Q') goto returnPoint;
print2("\n");
print2("This program checks built-in lattices (hard-coded into the\n");
print2("function 'initLattice') for violation of the user's input\n");
print2("conjecture.  Input formulas may have up to 23 variables\n");
print2("a,b,c,d,e,f,g,h,j,k,l,m,n,p,q,r,s,t,u,w,x,y,z (i,o,v omitted)\n");
print2("and constants 0,1.\n");
print2("\n");
print2("The output is of the form \"<result> <name>\" where <name> is\n");
print2("the lattice name.  An assignment that violates the lattice is\n");
print2("given when the <result> is FAILED.\n");
print2("\n");
print2("To compile for Unix:  Use 'gcc latticeg.c -o latticeg'.  The\n");
print2("entire program is contained in the single file latticeg.c .\n");
print2("\n");
print2("To run:  Type, at the Unix or DOS prompt, \'latticeg <hyp> \n");
print2("<hyp> ... <conclusion>\' where there are zero or more hypotheses\n");
print2("followed by one conclusion.  Each argument may be enclosed in\n");
print2("double or (Unix) single quotes if there is ambiguity.  Example:\n");
print2("\'latticeg \"1<(x#y)\" \"x=y\"\' tests the orthomodular law.\n");
print2("\n");
print2("\n");
print2("\n");
print2("\n");
print2("\n");
linput(NULL,"Press Enter to continue, q to quit...",&a);
if (toupper(a[0]) == 'Q') goto returnPoint;
print2("\n");
print2("Each <hyp> and the <conclusion> must be a <wff>:\n");
print2("    <atom> := a | b | c | d | e | f | g | h | j | k | l | m | n |\n");
print2("                  p | q | r | s | t | u | w | x | y | z | 0 | 1\n");
print2("    <opr> := ^ | v | # | O | I | 2 | 3 | 4 | 5 \n");
print2("    <uopr> := -\n");
print2("    <term> := <atom> | <uopr> <term> | ( <term> <opr> <term> )\n");
print2(
"    <wff> := <term> = <term> | <term> < <term> | <term> > <term> |\n");
print2(
"             <term> [ <term>\n");
print2("where a,b,c,... are variables (no i,o,v); 0,1 are constants; and\n");
print2("    - = negation (orthocomplement, NOT)\n");
print2("    ^ = conjunction (cap, meet, infimum, AND)\n");
print2("    v = disjunction (cup, join, supremum, OR)\n");
print2("    # = biimplication: ((x^y)v(-x^-y))\n");
print2("    O = ->0 = classical arrow: (-xvy)\n");
print2("    I = ->1 = Sasaki arrow: (-xv(x^y))\n");
print2("    2 = ->2 = Dishkant arrow: (-yI-x)\n");
print2("    3 = ->3 = Kalmbach arrow: (((-x^y)v(-x^-y))v(x^(-xvy)))\n");
print2("    4 = ->4 = non-tollens arrow: (-y3-x)\n");
print2("    5 = ->5 = relevance arrow: (((x^y)v(-x^y))v(-x^-y))\n");
print2(
"and = is equality, < is less-than-or-equal, > is g.e., [ is commutes:\n");
print2("    x<y is (xvy)=y; x>y is y<x; x[y is x=((x^y)v(x^-y)).\n");
print2("\n");
print2("\n");
linput(NULL,"Press Enter to continue, q to quit...",&a);
if (toupper(a[0]) == 'Q') goto returnPoint;
print2("\n");
print2("A FAILED result shows the failing assignments with a,b,c... as\n");
print2("lattice points and A,B,C,... as complemented lattice points.\n");
print2("Use the 'latticeg -p' program option to see lattice contents.\n");
print2("\n");
print2("For lattices with more than 48 lattice points, the failing\n");
print2("assignments beyond ...x,y,z are shown as &a,&b,&c,... with\n");
print2("complements &A,&B,&C,...\n");
print2("\n");
print2("The Greechie lattice atom numbers 1,2,3,...,23 correspond to nodes\n");
print2("a,b,c,... from the list:\n");
print2("\n");
print2("     a,b,c,d,e,f,g,h,j,k,l,m,n,p,q,r,s,t,u,w,x,y,z\n");
print2("\n");
print2("(with i,o,v omitted) and 24,25,26,... by:\n");
print2("\n");
print2("     &a,&b,&c,...\n");
print2("\n");
print2("with no omissions in the letter sequence.\n");
print2("\n");
print2("\n");
print2("\n");
print2("\n");
print2("\n");
linput(NULL,"Press Enter to continue, q to quit...",&a);
if (toupper(a[0]) == 'Q') goto returnPoint;
print2("\n");
print2("How to handle long equations:\n");
print2("You may specify strings to be replaced by continuation lines using\n");
print2("arguments of the form '$x' where x is a single character.  The\n");
print2("program will prompt with '$x> '; in response you should enter the\n");
print2("corresponding argument (quotes are optional).  Hint:  In Unix\n");
print2("enclose '$x' in single quotes to suppress shell interpretation.\n");
print2("\n");
print2("  latticeg $1 $2 \"x=z\"\n");
print2("  $1> x=y\n");
print2("  $2> y=z\n");
print2("\n");
print2("is the same as\n");
print2("\n");
print2("  latticeg \"x=y\" \"y=z\" \"x=z\"\n");
print2("\n");
print2("To break up very long formulas, use '$x$y$z...'; the above is also\n");
print2("the same as:\n");
print2("\n");
print2("  latticeg $1$2 $3 \"x=z\"\n");
print2("  $1> x=\n");
print2("  $2> y\n");
print2("  $3> y=z\n");
print2("\n");
linput(NULL,"Press Enter to continue, q to quit...",&a);
if (toupper(a[0]) == 'Q') goto returnPoint;
print2("\n");
print2("Using an input file (-i option):\n");
print2("\n");
print2("The input file should have one Greechie diagram per line, with each\n");
print2("atom name a character from the list:\n");
print2("  %s\n", left(ATOM_MAP, 54));
print2("  %s\n", right(ATOM_MAP, 55));
print2("where atom 1 = 1, atom 9 = 9, atom 10 = A, atom 35 = Z,\n");
print2("atom 36 = a, etc. with no gaps in numbering.  Blocks are separated\n");
print2("with a comma, and the last block ends with a period.  Each block\n");
print2("must have 3 or 4 atoms.  For example the diagram on p. 319 of\n");
print2("Kalmbach's _Orthomodular Lattices_ would be\n");
print2("  124,235,167,389.\n");
print2("Use the -p option to see the resulting lattice.\n");
print2("\n");
print2("\n");
print2("\n");
print2("\n");
print2("\n");
print2("\n");
print2("\n");
print2("\n");
print2("\n");
returnPoint:
let(&a, ""); /* Deallocate string */
return;
} /* printhelp() */


void testAllLattices(void)
{
  long latticeCase = 0;
  if (userArg > 0) latticeCase = userArg - 1;

  while (1) {
    latticeCase++;
    if (latticeCase > MAX_LATTICES    /* Built-in lattice limit */
        && latticeCase > fileLattices /* External file lattices */
        && userArg <= 0) break;                   /* Exhausted lattice cases */
    initLattice(latticeCase);
    if (!nodes && userArg > 0) {
      print2("?Error: program does not have lattice %ld\n", userArg);
      return;
    }
    if (!nodes) continue; /* Ignore gap in lattice numbering */
    if (test() == 'T') {
      print2("Passed %s\n", latticeName);
      if (showGreechie) print2("%s passed: %s\n", latticeName, greechieStmt);
    } else {
      if (!showAllFailures) {
        /* (If showAllFaiure, failure was printed already by test()) */
        print2("FAILED %s at %s\n", latticeName, failingAssignment);
      }
      if (showGreechie) print2("%s failed: %s\n", latticeName, greechieStmt);
      if (userArg == 0) break; /* Default: stop on first failure */
    }
    if (userArg > 0) break; /* User wants to try only one lattice */
  }

  return;
} /* testAllLattices() */

/* This function is called when the user specifies the -c option.
   The hypotheses and conclusion are broken down into subformulas,
   then all subformula pairs are tested to see if they potentially
   commute (i.e. if assuming they commute does not fail the OM
   lattices).  The pairs that potentially commute are printed out. */
void findCommutingPairs(void)
{
  long latticeCase;
  long i, j, subformulas;
  unsigned char testResult;
  vstring str1 = "";
  vstring exprList = "";
  vstring subformList = "";
  vstring fromPol1 = "";
  vstring fromPol2 = "";
  char OAOnlyFlag;

  print2(
"This is a list of potentially commuting pairs i.e. that pass all OM\n");
  print2(
"lattices with the hypotheses assumed.  Additional pairs that potentially\n");
  print2(
"commute assuming OA are prefixed with '(OA)'.\n");

  /* Build a list with all expressions - abbreviated */
  for (i = 0; i < hypotheses; i++) {
    /* Strip leading '=' before adding to subformula list */
    let(&exprList, cat(exprList, right(polHypList[i], 2), NULL));
  }
  let(&exprList, cat(exprList, right(polConclusion, 2), NULL));

  /* Unabbreviate the hypotheses and conclusion */
  let(&str1, "");
  for (i = 0; i < hypotheses; i++) {
    str1 = unabbreviate(polHypList[i]);
    let(&polHypList[i], str1);
    let(&str1, ""); /* Deallocate from unabbreviate fn call */
  }
  str1 = unabbreviate(polConclusion);
  let(&polConclusion, str1);
  let(&str1, ""); /* Deallocate from unabbreviate fn call */

  /* Build a list with all expressions - unabbreviated */
  /* The unabbreviated expressions obtain additional commute pairs */
  for (i = 0; i < hypotheses; i++) {
    /* Strip leading '=' before adding to subformula list */
    let(&exprList, cat(exprList, right(polHypList[i], 2), NULL));
  }
  let(&exprList, cat(exprList, right(polConclusion, 2), NULL));

  /* Build a list with all subformulas */
  let(&subformList, "");
  subformList = subformulaList(exprList);

  subformulas = numEntries(subformList);
  for (i = 1; i <= subformulas - 1; i++) {
    for (j = i + 1; j <= subformulas; j++) {
      /* Set conclusion to "subformula pair commutes" */
      let(&polConclusion, cat("[", entry(i, subformList),
          entry(j, subformList), NULL));
      latticeCase = 0;
      OAOnlyFlag = 0;
      while (1) {
        latticeCase++;
        initLattice(latticeCase);
        if (latticeCase > MAX_LATTICES)
          bug(10);/* Exhausted lattice cases - means OMDoneFlag
                     was never set */
        if (!nodes) continue; /* Ignore gap in lattice numbering */
        testResult = test();
        if (testResult != 'T') break;
           /* Failed - definitely doesn't commute */
        if (OADoneFlag) OAOnlyFlag = 1;
        if (OMDoneFlag) break; /* Did all OM cases */
      }
      if (testResult == 'T' ||
          OAOnlyFlag) {
        /* The pair potentially commutes - print it out */
        let(&fromPol1, "");
        fromPol1 = fromPolish(entry(i, subformList));
        /* Strip leading and trailing parenths */
        /* if (fromPol1[0] == '(')
          let(&fromPol1, seg(fromPol1, 2, strlen(fromPol1) - 1));*/
        let(&fromPol2, "");
        fromPol2 = fromPolish(entry(j, subformList));
        /* Strip leading and trailing parenths */
        /*if (fromPol2[0] == '(')
          let(&fromPol2, seg(fromPol2, 2, strlen(fromPol2) - 1));*/
        if (testResult != 'T')
          let(&fromPol1, cat("(OA) ", fromPol1, NULL));
        if (strlen(fromPol1) + strlen(fromPol2) <= 74) {
          /* Print on one line */
          print2("%s and %s\n", fromPol1, fromPol2);
        } else {
          /* Print on two lines */
          print2("%s and\n  %s\n", fromPol1, fromPol2);
        }
      }
    } /* next j */
  } /* next i */

  let(&str1, "");
  let(&exprList, "");
  let(&subformList, "");
  let(&fromPol1, "");
  let(&fromPol2, "");
  return;
} /* findCommutingPairs() */



void init(void) /* Should be called only once!! */
{
  long i,j;
  vstring tmpStr = "";

  if (fp != NULL) {
    /* The user specified an input file */
    /* Count the lines - each line is a lattice */

    fileLattices = 0;
    while (linput(fp, NULL, &tmpStr)) {
      fileLattices++;
    }
    print2("The input file has %ld lattices.\n", fileLattices);
    /* Put the lines into an array of strings */
    rewind(fp);

    fileLattice = malloc((fileLattices + 1) * sizeof(vstring));
    if (fileLattice == NULL) {
      print2("?Error: Out of memory\n");
      exit(0);
    }
    for (i = 1; i <= fileLattices; i++) {
      fileLattice[i] = "";
    }
    i = 0;
    while (linput(fp, NULL, &tmpStr)) {
      i++;
      fileLattice[i] = tmpStr;
      /* Make tmpStr point to empty string so next linput won't deallocate
         the string that fileLattice[i] now points to. */
      tmpStr = "";
    }
    fclose(fp);
    if (i != fileLattices) bug(11);
  }


  if (MAX_BIN_OPERS != strlen(BIN_OPERS)) bug(1);
  if (MAX_VARS != strlen(VAR_LIST)) bug(2);

  /* vstring initialization */
  for (i = 0; i < MAX_NODES; i++) {
    nodeList[i] = "";
    for (j = 0; j < MAX_NODES; j++) {
      sup[i][j] = "";
    }
  }
  for (i = 0; i < MAX_HYPS; i++) {
    hypList[i] = "";
    polHypList[i] = "";
  }

  /* Initialize speedup table for negative */
  for (i = 'A'; i <= 'z'; i++) {
    if (isupper(i)) {
      negMap[i] = tolower(i);
      /* Extended characters for >23 atoms */
      negMap[i + 128] = 128 + tolower(i);
    }
    if (islower(i)) {
      negMap[i] = toupper(i);
      /* Extended characters for >23 atoms */
      negMap[i + 128] = 128 + toupper(i);
    }
  }
  negMap['0'] = '1';
  negMap['1'] = '0';

  /* Initialize speedup table for operator */
  for (i = 1; i < 256; i++) {
    j = instr(1, BIN_OPERS, chr(i));
    if (j != 0) {
      opMap[i] = j - 1; /* Points to location of char in BIN_OPERS string */
    } else {
      opMap[i] = 127; /* Not a binary operation */
    }
    let(&tmpStr, ""); /* Deallocate temporary stack to prevent overflow */
  }
  let(&lletterlist, cat("abcdefghjklmnpqrstuwxyz", space(26), NULL));
  let(&uletterlist, cat("ABCDEFGHJKLMNPQRSTUWXYZ", space(26), NULL));
  for (i = 0; i < 26; i++) {
    lletterlist[i + 23] = (i + 'a' + 128) & 0xFF;
    uletterlist[i + 23] = (i + 'A' + 128) & 0xFF;
  }

  /* To debug extended characters:  This code makes the extended characters
     the first ones that are used, so the "&" notation can be debugged with
     smaller Greechie diagrams. */
  /*
  let(&lletterlist, cat(space(26), "abcdefghjklmnpqrstuwxyz", NULL));
  let(&uletterlist, cat(space(26), "ABCDEFGHJKLMNPQRSTUWXYZ", NULL));
  for (i = 0; i < 26; i++) {
    lletterlist[i ] = (i + 'a' + 128) & 0xFF;
    uletterlist[i ] = (i + 'A' + 128) & 0xFF;
  }
  */

} /* init */

void initLattice(long latticeNum)
{
  /* latticeNum:  1 = Boolean lattice, 2 = MO2 lattice, ... */
  long i ,j ,k, p, q, changed;
  vstring nname = "";
  vstring nbranch = "";
  vstring big = "";
  vstring small1 = "";
  vstring small2 = "";
  vstring oldsup = "";
  vstring tmpNodeList[MAX_NODES];

  nodes = 0; /* If 0 is returned, lattice is undefined */

  if (endArg && latticeNum > endArg) return;

  /* Initialize flags for special lattices */
  OADoneFlag = 0;
  OMDoneFlag = 0;

  /* Global variables used by greechie3() */
  glatticeNum = latticeNum;
  glatticeCount = 0;

  /* Get the Greechie diagram corresponding to glatticeNum */
  if (fp == NULL) {
    /* Use the ones hard-coded into this program */
    usersGreechieDiagrams();
  } else {
    /* We read from an input file */
    /* Trick greechie3() into thinking we've already done latticeNum cases */
    glatticeCount = latticeNum - 1;
    /* Call greechie3() as if this were a hard-coded user lattice */
    if (latticeNum <= fileLattices) greechie3("", fileLattice[latticeNum]);
  }

  if (!nodes) return;
  if (nodes >= MAX_NODES) bug(8);

  /* Node name list */
  let(&nodeNames, "");
  for (i = 1; i <= nodes; i++) {
    let(&nodeNames, cat(nodeNames, left(nodeList[i], 1), NULL));
    /* Initialize speed-up lookup table */
    nodeNameMap[ascii_(left(nodeList[i], 1))] = i;
  }

  /* Fill out ordering table */
  for (i = 1; i <= nodes; i++) {
    tmpNodeList[i] = "";
    let(&(tmpNodeList[i]), nodeList[i]);
  }
  changed = 1;
  while (changed) {
    changed = 0;
    for (i = 1; i <= nodes; i++) {
      for (j = 1; j <= nodes; j++) {
        if (i != j) {
          let(&nname, left(tmpNodeList[j], 1));
          p = instr(2, tmpNodeList[i], nname);
          if (p != 0) {
            for (k = 2; k <= len(tmpNodeList[j]); k++) {
              let(&nbranch, mid(tmpNodeList[j], k, 1));
              q = instr(2, tmpNodeList[i], nbranch);
              if (q == 0) {
                changed = 1;
                let(&(tmpNodeList[i]), cat(tmpNodeList[i], nbranch, NULL));
              } /* end if */
            } /* next k */
          } /* end if */
        } /* end if */
      } /* next j */
    } /* next i */
  } /* next while */

  /* Build supremum (disjunction) table */
  for (i = 1; i <= nodes; i++) {
    for (j = 1; j <= nodes; j++) {
      let(&(sup[i][j]), "1");
    } /* next j */
  } /* next i */
  for (i = 1; i <= nodes; i++) {
    let(&big, left(tmpNodeList[i], 1));
    for (j = 1; j <= len(tmpNodeList[i]); j++) {
      for (k = 1; k <= len(tmpNodeList[i]); k++) {
        let(&small1, mid(tmpNodeList[i], j, 1));
        let(&small2, mid(tmpNodeList[i], k, 1));
        let(&oldsup, sup[instr(1, nodeNames, small1)]
            [instr(1, nodeNames, small2)]);
        /* oldsup > big */
        if (instr(2, tmpNodeList[instr(1, nodeNames, oldsup)], big)) {
          let(&(sup[instr(1, nodeNames, small1)]
              [instr(1, nodeNames, small2)]), big);
        } /* end if */
      } /* next k */
    } /* next j */
  } /* next i */

  /* Debug */
  /*
  for (i = 1; i <= nodes; i++) {
    for (j= 1; j <= nodes; j++) {
      print2("%s",sup[i][j]);
    }
    print2("\n");
  }
  */

  /* Build table with all binary operations for fast lookup */
  for (i = 0; i < strlen(BIN_OPERS); i++) {
    for (j = 1; j <= nodes; j++) {
      for (k = 1; k <= nodes; k++) {
        binOpTable[i][j][k] = lookupBinOp(BIN_OPERS[i],
            (unsigned char)(nodeNames[j - 1]),
            (unsigned char)(nodeNames[k - 1]));
      } /* next k */
    } /* next j */
  } /* next i */

  /* Deallocate vstrings */
  for (i = 1; i <= nodes; i++) {
    let(&(tmpNodeList[i]), "");
  }
  let(&nname, "");
  let(&nbranch, "");
  let(&big, "");
  let(&small1, "");
  let(&small2, "");
  let(&oldsup, "");

} /* initLattice() */


/* Converts a Greechie height 3 lattice, using the matrix notation in
   Kalmbach, "Orthomodular Lattices", p. 319, and populates the
   nodes variable and the nodeList array.  glattice is a comma-separated
   list of all nodes, with rows (assumed to be width 3) placed after
   each other. */
void greechie3(vstring name, vstring glattice) {
  long atoms, atom, atom2, msize, legs, i, j, k, m, n;
  vstring glattice1 = "";
  vstring str1 = "";
  vstring str2 = "";

  glatticeCount++;  /* Global variable */
  if (glatticeCount != glatticeNum) return; /* Early exit if not the one */
  if (startArg && glatticeNum < startArg) return;

  /* In case of temp alloc of glattice; also trim leading, trailing spaces */
  /* Also trim CR in case of DOS file read in Unix */
  let(&glattice1, edit(glattice, 8 + 128 + 4));

  /* Assign the lattice name */
  if (name[0] == 0) {
    /* Blank argument - assign default name */
    let(&latticeName, cat("#", str(glatticeNum), NULL));
  } else {
    let(&latticeName, name); /* User's name */
  }

  if (showGreechie || removeLegs)
    print2("%s original: %s\n", latticeName, glattice1);

  /* Parse the input string describing the lattice */

  if (instr(1, glattice1, ".") == 0) {
    /* No period - assume old standard:  3-atom blocks */

    /* The following code allows free-formatted lists with spaces and/or
       commas as delimiters */
    i = 0;
    while (glattice1[i]) { /* Convert commas to spaces */
      if (glattice1[i] == ',') glattice1[i] = ' ';
      i++;
    }
    let(&glattice1, edit(glattice1, 8 + 16 + 128)); /* Reduce & trim spaces */
    i = 0;
    while (glattice1[i]) { /* Convert spaces to commas */
      if (glattice1[i] == ' ') glattice1[i] = ',';
      i++;
    }

    msize = numEntries(glattice1); /* Total matrix entries (nx3 matrix) */
    if ((msize / 3) * 3 != msize) {
      print2("%s: %s\n", latticeName, glattice1);
      print2("?Error: Number of Greechie matrix entries must be a factor of 3\n");
      exit(0);
    }
    blocks = msize / 3;
    atoms = 0;
    for (i = 1; i <= msize; i++) {
      j = val(entry(i, glattice1));
      let(&str1, ""); /* Purge temp alloc in entry() */
      if (j <= 0) {
        print2("%s: %s\n", latticeName, glattice1);
        print2("?Error: Greechie matrix nodes must be numbers > 1\n");
        exit(0);
      }
      if (j > atoms) atoms = j;
    }

    for (i = 1; i < msize; i = i + 3) {
      blockSize[(i + 2) / 3] = 3; /* Fixed block size of 3 */
      for (j = 0; j < 3; j++) {
        block[(i + 2) / 3][j + 1] = val(entry(i + j, glattice1));
      }
    }
  } else {
    /* glattice1 has period - assume new (Brendan) compact standard */
    if (glattice1[0] == '+') {
      print2("%s: %s\n", latticeName, glattice1);
      print2("?Error: '+' notation for large diagrams is not implemented\n");
      exit(0);
    }
    let(&glattice1, edit(glattice1, 2)); /* Remove spaces */
    n = strlen(glattice1);
    if (glattice1[n - 1] != '.') {
      print2("%s: %s\n", latticeName, glattice1);
      print2("?Error: Last character should be a period\n");
      exit(0);
    }
    if (instr(1, left(glattice1, n - 1), ".") != 0) {
      print2("%s: %s\n", latticeName, glattice1);
      print2("?Error: Period can only be last character\n");
      exit(0);
    }
    if (n == 1) {
      print2("%s: %s\n", latticeName, glattice1);
      print2("?Error: Diagram must have at least one block\n");
      exit(0);
    }
    atoms = 0;
    blocks = 1;
    blockSize[blocks] = 0;
    for (i = 0; i < n; i++) {
      if (glattice1[i] == ',' || glattice1[i] == '.') {
        /* End of block */
        if (blockSize[blocks] < MIN_BLOCK_SIZE) {
          print2("%s: %s\n", latticeName, glattice1);
          print2("?Error: Minimum block size is %ld\n", (long)MIN_BLOCK_SIZE);
          exit(0);
        }
        if (glattice1[i] == ',') {
          /* Start of new block */
          blocks++;
          if (blocks > MAX_BLOCKS) {
            print2("%s: %s\n", latticeName, glattice1);
            print2("?Error: Maximum blocks allowed is %ld\n",
                (long)MAX_BLOCKS);
            exit(0);
          }
          blockSize[blocks] = 0;
        }
        continue;
      }
      /* Get the atom number */
      j = instr(1, ATOM_MAP, chr(glattice1[i]));
      let(&str1, ""); /* Deallocate chr call */
      if (j == 0) {
        print2("%s: %s\n", latticeName, glattice1);
        print2("?Error: Illegal character '%c' in diagram\n", glattice1[i]);
        exit(0);
      }
      blockSize[blocks]++;
      if (blockSize[blocks] > MAX_BLOCK_SIZE) {
        print2("%s: %s\n", latticeName, glattice1);
        print2("?Error: Maximum block size is %ld\n", (long)MAX_BLOCK_SIZE);
        exit(0);
      }
      /* Assign the atom */
      block[blocks][blockSize[blocks]] = j;
      if (j > atoms) atoms = j; /* Maximum atom number */
    } /* next i */
  } /* glattice1 has a period (new compact standard) */

  for (i = 1; i <= blocks; i++) {
    for (j = 1; j <= blockSize[i] - 1; j++) {
      for (k = j + 1; k <= blockSize[i]; k++) {
        if (block[i][j] == block[i][k]) {
          print2("%s: %s\n", latticeName, glattice1);
          print2("?Error: Duplicate Greechie matrix node numbers in a block\n");
          exit(0);
        }
      }
    }
  }

  /* Special function:  see if diagram has 1 or more "legs" i.e. a block
     of rank 1. */
  /* This function also checks to see that 2-atom blocks are disconnected. */
  legs = 0;
  if (showGreechie || removeLegs) {
    for (i = 1; i <= blocks; i++) {
      n = 0; /* rank of block */
      for (j = 1; j <= blockSize[i]; j++) {
        /* Scan all other blocks for atom block[i][j] */
        for (k = 1; k <= blocks; k++) {
          if (k == i) continue;
          for (m = 1; m <= blockSize[k]; m++) {
            if (block[i][j] == block[k][m]) {
              if (blockSize[i] == 2 || blockSize[k] == 2) {
                print2("%s: %s\n", latticeName, glattice1);
                print2("?Error: Connected 2-atom blocks are not allowed\n");
                exit(0);
              }
              n++; /* Increase rank */
              goto exit_scan;
            }
          }
         exit_scan:
          break;
        }
      }
      if (n == 1) {
        /* This is a foot */
        legs = 1; /* Means at least one foot */
        break;
      }
    } /* next i */
  }

  /* Count gaps in atom numbering */
  let(&str1, string(atoms, '0'));
  for (i = 1; i <= blocks; i++) {
    for (j = 1; j <= blockSize[i]; j++) {
      str1[block[i][j] - 1] = '1';
    }
  }
  n = 0; /* Number of gaps in atom numbering */
  for (i = 0; i < atoms; i++) {
    if (str1[i] == '0') n++;
  }
  /* Error if gaps in numbering */
  if (n > 0) {
    print2("%s: %s\n", latticeName, glattice1);
    print2("?Error: There are gaps in atom numbering\n");
    exit(0);
  }

  /* This is the global greechie3 diagram for use wherever
     we want to print it out.  It is in the new compact form; must
     be enhanced later for huge diagrams ("+" notation) */
  let(&greechieStmt, "");
  for (i = 1; i <= blocks; i++) {
    for (j = 1; j <= blockSize[i]; j++) {
      let(&greechieStmt, cat(greechieStmt, mid(ATOM_MAP, block[i][j], 1),
          NULL));
    }
    if (i < blocks) {
      let(&greechieStmt, cat(greechieStmt, ",", NULL));
    } else {
      let(&greechieStmt, cat(greechieStmt, ".", NULL));
    }
  }

  nodes = (atoms * 2) + 2;
  n = atoms;
  for (i = 1; i <= blocks; i++) {
    /* Add 6 nodes for each 4-atom block */
    block4Offset[i] = 0;
    if (blockSize[i] == 4) {
      block4Offset[i] = n;
      nodes = nodes + 6;
      n = n + 3;
    }
  }
  /* Enhance lattice name with atom/block/node count */
  let(&latticeName, cat(latticeName, " (", str(atoms), "/", str(blocks),
      "/", str(nodes), ")", NULL));

  if (removeLegs && legs > 0) {
    print2("%s >0 legs, skipped: %s\n", latticeName,
        greechieStmt);
    nodes = 0; /* Force lattice to be skipped */
    goto returnPoint;
  }

  /********* loop.c program ************/
  print2("%s original\n", greechieStmt);
  for (i = 1; i <= blocks; i++) {
    print2("Starting block = %ld\n", i);
    let(&str1, chr(i));
    loop(str1);
  }
  exit(0);

  if (nodes > 100) {
    print2("%s: %s\n", latticeName, greechieStmt);
    print2("?Error: Greechie diagram %s has more than 100 nodes\n",
        latticeName);
    nodes = 0;  /* Force lattice to be skipped */
    goto returnPoint;
  }


  /* Assign the unit node - all complemented atoms go below it */
  let(&(nodeList[1]), cat("1", left(uletterlist, atoms), NULL));

  /* Assign the zero node below all uncomplemented atoms */
  for (i = 1; i <= atoms; i++) {
    let(&(nodeList[i + 1]), cat(mid(lletterlist, i, 1), "0", NULL));
  }

  /* Assign the zero node */
  let(&(nodeList[atoms + 2]), "0");

  /* Create complemented atom list */
  for (i = 1; i <= atoms; i++) {
    let(&(nodeList[atoms + 2 + i]), mid(uletterlist, i, 1));
  }

  /* Create extra nodes for 4-atom blocks */
  for (i = 1; i <= blocks; i++) {
    if (blockSize[i] == 4) {
      for (j = 0; j < 3; j++) {
        let(&(nodeList[2 * block4Offset[i] + 2 + 2*j + 1]),
            mid(uletterlist, block4Offset[i] + j + 1, 1));
        let(&(nodeList[2 * block4Offset[i] + 2 + 2*j + 2]),
            mid(lletterlist, block4Offset[i] + j + 1, 1));
      }
      /* t */
      let(&(nodeList[2 * block4Offset[i] + 2 + 1]), cat(
          mid(lletterlist, block4Offset[i] + 1, 1),
          mid(lletterlist, block[i][1], 1),
          mid(lletterlist, block[i][2], 1), NULL));
      /* u */
      let(&(nodeList[2 * block4Offset[i] + 2 + 2]), cat(
          mid(lletterlist, block4Offset[i] + 2, 1),
          mid(lletterlist, block[i][1], 1),
          mid(lletterlist, block[i][3], 1), NULL));
      /* v */
      let(&(nodeList[2 * block4Offset[i] + 2 + 3]), cat(
          mid(lletterlist, block4Offset[i] + 3, 1),
          mid(lletterlist, block[i][1], 1),
          mid(lletterlist, block[i][4], 1), NULL));
      /* t' */
      let(&(nodeList[2 * block4Offset[i] + 2 + 4]), cat(
          mid(uletterlist, block4Offset[i] + 1, 1),
          mid(lletterlist, block[i][3], 1),
          mid(lletterlist, block[i][4], 1), NULL));
      /* u' */
      let(&(nodeList[2 * block4Offset[i] + 2 + 5]), cat(
          mid(uletterlist, block4Offset[i] + 2, 1),
          mid(lletterlist, block[i][2], 1),
          mid(lletterlist, block[i][4], 1), NULL));
      /* v' */
      let(&(nodeList[2 * block4Offset[i] + 2 + 6]), cat(
          mid(uletterlist, block4Offset[i] + 3, 1),
          mid(lletterlist, block[i][2], 1),
          mid(lletterlist, block[i][3], 1), NULL));
    }
  }

  /* Process Boolean blocks */
  for (i = 1; i <= blocks; i++) {
    if (blockSize[i] == 3) {
      for (j = 1; j <= 3; j++) {
        atom = block[i][j];
        for (k = 1; k <= 3; k++) {
          if (k == j) continue;
          atom2 = block[i][k];
          let(&(nodeList[atoms + atom + 2]), cat(nodeList[atoms + atom + 2],
              mid(lletterlist, atom2, 1), NULL));
        }
      }
      continue;
    }
    if (blockSize[i] == 4) {
      let(&(nodeList[atoms + block[i][1] + 2]),
          cat(nodeList[atoms + block[i][1] + 2],
          mid(uletterlist, block4Offset[i] + 1, 1),
          mid(uletterlist, block4Offset[i] + 2, 1),
          mid(uletterlist, block4Offset[i] + 3, 1), NULL));
      let(&(nodeList[atoms + block[i][2] + 2]),
          cat(nodeList[atoms + block[i][2] + 2],
          mid(uletterlist, block4Offset[i] + 1, 1),
          mid(lletterlist, block4Offset[i] + 2, 1),
          mid(lletterlist, block4Offset[i] + 3, 1), NULL));
      let(&(nodeList[atoms + block[i][3] + 2]),
          cat(nodeList[atoms + block[i][3] + 2],
          mid(lletterlist, block4Offset[i] + 1, 1),
          mid(uletterlist, block4Offset[i] + 2, 1),
          mid(lletterlist, block4Offset[i] + 3, 1), NULL));
      let(&(nodeList[atoms + block[i][4] + 2]),
          cat(nodeList[atoms + block[i][4] + 2],
          mid(lletterlist, block4Offset[i] + 1, 1),
          mid(lletterlist, block4Offset[i] + 2, 1),
          mid(uletterlist, block4Offset[i] + 3, 1), NULL));
      continue;
    }
    bug(9);
  }

 returnPoint:
  let(&str1, ""); /* Deallocate */
  let(&str2, ""); /* Deallocate */
  let(&glattice1, ""); /* Deallocate */

} /* greechie3 */

/***** loop.c program *******/
void loop(vstring sofar) {
  /* Test case:

    loop -i 1.tmp x=x

    123,345,561.

  */
  long i,j,k,f,g,m,n,trialBlock,intersection,a1,a2,m1,m2;
  long sharedAtoms; /* 10-Oct-03 */
  char foundShared; /* 10-Oct-03 */
  vstring str1 = "";
  static long biggestSoFar = 0;
  char s = 0; /* 0 = loops, 1 = straight line test */
  static char s1 = 1;
  vstring dummy = "";
  vstring printstr = "";
  vstring loopstr = "";

  /* 27-Sep-2009 */
  /* Blocks that will be re-ordered by atom number */
  long ordBlock[MAX_BLOCKS + 1][MAX_BLOCK_SIZE + 1];
  long ordBlockSize[MAX_BLOCKS + 1];
  long ordBlocks = 0;
  long ordAtomInBlock = 0;
#define MAX_ATOMS 1000
  long ordRemap[MAX_ATOMS + 1];
  long ordChars;
  vstring ordGreechieStmt = "";
  vstring ordGreechieStmtPrefix = "";


  intersection = 0; /* Prevent compiler warning */

  if (s1) {  /* First time */
    /*
    if (s == 0) print2("Loop test - change s to 1 for straight line test\n");
    if (s == 1) print2("Straight line test - change s to 0 for loop test\n");
    */
    s1 = 0;
  }

  for (trialBlock = 1; trialBlock <= blocks; trialBlock++) {

    if (s == 0) {
      /* Look for loop */

      /* Look for duplicate blocks */
      f = 0;
      for (k = 0; k < strlen(sofar); k++) {
        if (trialBlock == (long)(sofar[k])) {
          f = 1;
          break;
        }
      }
      if (f == 1) continue;

      /* See if trialBlock intersects last block */
      f = 0;
      for (i = 1; i<= blockSize[trialBlock]; i++) {
        for (j = 1; j <= blockSize[(long)(sofar[strlen(sofar)-1])]; j++) {
          if (block[trialBlock][i] == block[(long)(sofar[strlen(sofar)-1])][j]) {
            f = 1;
            intersection = block[trialBlock][i];
            break;
          }
        }
        if (f == 1) break;
      }
      if (f == 0) continue;

      /* Make sure trialBlock doesn't intersect middle blocks, or 1st
         block at intersection point of last block */
      f = 0;
      for (k = 0; k < strlen(sofar) - 1; k++) {
        for (i = 1; i<= blockSize[trialBlock]; i++) {
          for (j = 1; j <= blockSize[(long)(sofar[k])]; j++) {
            if (block[trialBlock][i] == block[(long)(sofar[k])][j]) {
              if (k > 0 ||  block[trialBlock][i] == intersection) {
                f = 1;
                break;
              }
            }
          }
          if (f == 1) break;
        }
      }
      if (f == 1) continue;

      /* See if trialBlock intersects 1st block */
      f = 0;
      for (k = 0; k < 1; k++) {
        for (i = 1; i<= blockSize[trialBlock]; i++) {
          for (j = 1; j <= blockSize[(long)(sofar[k])]; j++) {
            if (block[trialBlock][i] == block[(long)(sofar[k])][j]) {
              f = 1;
              break;
            }
          }
          if (f == 1) break;
        }
      }

      let(&str1, cat(sofar, chr(trialBlock), NULL));

      /* For better diagram - make sure that no other block shares
         all atoms with the blocks so far */
      /* 10-Oct-03 nm - The above only works for 3-atom blocks.  For the
         general case, we make sure that no block (not in the blocks
         collected so far) shares 3 or more atoms with the blocks so far. */
      g = 0;
      for (m = 1; m <= blocks; m++) { /* For each remaining block */
        sharedAtoms = 0; /* 10-Oct-03 */
        if (instr(1, str1, chr(m)) != 0) continue; /* Skip blocks so far */
        for (n = 1; n <= blockSize[m]; n++) { /* For each atom in it */
          foundShared = 0;
          for (k = 0; k < strlen(str1); k++) { /* For each block so far */
            for (i = 1; i<= blockSize[(long)(str1[k])]; i++) {
                                                      /* For each atom in it */
              if (block[(long)(str1[k])][i] == block[m][n]) {
                foundShared = 1; /* A shared atom was found */
                break;
              }
            } /* next i */
            if (foundShared == 1) break;
                               /* Skip the rest of this block if a shared
                                  atom was found (two can't be shared in
                                  a Greechie diagram) */
                               /* 10-Oct-03 - may have to revisit this for
                                  MMP diagrams? */
          } /* next k */
          /* if (g == 0) break; */  /* (removed 10-Oct-03) */
          /* 10-Oct-03 */
          if (foundShared == 1) sharedAtoms++;  /* Count the shared atom found */
          if (sharedAtoms >= 3) {
            g = 1; /* Can't draw straight line */
            break;
          } else {
            g = 0; /* Still possible to draw straight line; continue search */
          }
          /* End of 10-Oct-03 */
        } /* next n */
        if (g == 1) break;  /* If g=1 it means we can't draw w/ only straight
                               lines */
      } /* next m */
      /*if (g == 1) continue;*/  /* This is if we don't even want to print it */


      if (f == 1 && strlen(sofar) > 1) { /* Loop found */
        /*
        for (k = 0; k < strlen(str1); k++) {
          for (i = 1; i<= blockSize[(long)(str1[k])]; i++) {
            print2(" %ld", block[(long)(str1[k])][i]);
          }
          print2(" ");
        }
        if (g) print2(" g");
        print2("\n");
        */

        /* Print right-justified loop size for later sorting by user */
        let(&printstr, cat(space(3 - strlen(str(strlen(str1)))),
            str(strlen(str1)), " ", NULL));
        let(&loopstr, "");

        /* 27-Sep-2009 - all references to variables ordxxx are new */
        let(&ordGreechieStmtPrefix, printstr);
        ordBlocks = 0;

        /* str1 has a list of the blocks in the loop */
        for (k = 0; k < strlen(str1); k++) {
          /* Get the previous block */
          if (k == 0) {
            m1 = strlen(str1) - 1;
          } else {
            m1 = k - 1;
          }
          /* Find the atom intersecting the previous block */
          a1 = 0;
          for (i = 1; i <= blockSize[(long)(str1[k])]; i++) {
            for (j = 1; j <= blockSize[(long)(str1[m1])]; j++) {
              if (block[(long)(str1[k])][i] == block[(long)(str1[m1])][j]) {
                a1 = block[(long)(str1[k])][i];
                break;
              }
            }
            if (a1) break;
          }
          if (!a1) bug(2001);
          /* Get the next block */
          if (k == strlen(str1) - 1) {
            m2 = 0;
          } else {
            m2 = k + 1;
          }
          /* Find the atom intersecting the next block */
          a2 = 0;
          for (i = 1; i <= blockSize[(long)(str1[k])]; i++) {
            for (j = 1; j <= blockSize[(long)(str1[m2])]; j++) {
              if (block[(long)(str1[k])][i] == block[(long)(str1[m2])][j]) {
                a2 = block[(long)(str1[k])][i];
                break;
              }
            }
            if (a2) break;
          }
          if (!a2) bug(2002);

          /* Add the first atom */
          let(&printstr, cat(printstr,
              chr(ATOM_MAP[a1 - 1]), NULL));
          let(&loopstr, cat(loopstr,
              chr(ATOM_MAP[a1 - 1]), NULL));

          ordBlocks++;
          ordAtomInBlock = 1;
          ordBlockSize[ordBlocks] = blockSize[(long)(str1[k])];
          ordBlock[ordBlocks][ordAtomInBlock] = a1;

          /* Add the middle atoms */
          for (i = 1; i<= blockSize[(long)(str1[k])]; i++) {
            /*print2("%c", ATOM_MAP[block[(long)(str1[k])][i] - 1]);*/
            if (a1 == block[(long)(str1[k])][i]) continue; /* Skip first */
            if (a2 == block[(long)(str1[k])][i]) continue; /* Skip last */
            let(&printstr, cat(printstr,
                chr(ATOM_MAP[block[(long)(str1[k])][i] - 1]), NULL));
            let(&loopstr, cat(loopstr,
                chr(ATOM_MAP[block[(long)(str1[k])][i] - 1]), NULL));

            ordAtomInBlock++;
            ordBlock[ordBlocks][ordAtomInBlock] = block[(long)(str1[k])][i];

          }
          /* Add the last atom */
          let(&printstr, cat(printstr,
              chr(ATOM_MAP[a2 - 1]), NULL));
          let(&loopstr, cat(loopstr,
              chr(ATOM_MAP[a2 - 1]), NULL));

          ordAtomInBlock++;
          if (ordAtomInBlock != ordBlockSize[ordBlocks]) bug(123);
          ordBlock[ordBlocks][ordAtomInBlock] = a2;

          /*if (k < strlen(str1) - 1) print2(",");*/
          if (k < strlen(str1) - 1) let(&printstr, cat(printstr, ",", NULL));
          if (k < strlen(str1) - 1) let(&loopstr, cat(loopstr, ",", NULL));
          /*if (k == strlen(str1) - 1) print2(".");*/
          if (k == strlen(str1) - 1) let(&printstr, cat(printstr, ".", NULL));
        }

        if (!instr(1, loops, loopstr)) {

          /* It is a new loop */
          /* Add to loop list */
          let(&loops, cat(loops, " ", loopstr, ",", loopstr, " ", NULL));
          /* Create reversed loop */
          let(&dummy, loopstr);
          for (k = 0; k < strlen(loopstr); k++) {
            dummy[strlen(loopstr) - k - 1] = loopstr[k];
          }
          /* Add reversed loop to list */
          let(&loops, cat(loops, " ", dummy, ",", dummy, " ", NULL));

          print2("%s", printstr);

          print2("  ");

          /* Print remaining blocks for reference */
          for (k = 1; k <= blocks; k++) {
            if (strchr(str1, k) == NULL) {

              ordBlocks++;
              ordBlockSize[ordBlocks] = blockSize[k];

              for (i = 1; i<= blockSize[k]; i++) {
                print2("%c", ATOM_MAP[block[k][i] - 1]);

                ordBlock[ordBlocks][i] = block[k][i];

                /* Show if it occurs in loop */
                if (strchr(loopstr, ATOM_MAP[block[k][i] - 1]) != NULL)
                  print2("*");  /* Occurs */
                else
                  print2(".");  /* Does not occur */
              }
              print2(" ");
            }
          }

          if (g) print2(" can't be drawn with all straight lines");
          print2("\n");


          /* Added 27-Sep-2009 - print Greechie diag. w/ 123,456, etc. */

          if (ordBlocks != blocks) bug(124);

          /* First, re-order the ordBlock atom numbers */

          /* Renumber the atoms starting at 1 from left to right */
          /* This code is roughly from mmpsubset.c */
          /* See wideString warning in mmpsubset.c */
          for (i = 0; i <= MAX_ATOMS; i++) {
               /* It seems "atoms" is not avail. so use MAX_ATOMS */
            ordRemap[i] = 0;
          }
          k = 0;
          for (i = 1; i <= ordBlocks; i++) {
            for (j = 1; j <= ordBlockSize[i]; j++) {
              if (ordRemap[ordBlock[i][j]] == 0) {
                k++;
                if (k > MAX_ATOMS) bug(126);
                if (ordBlock[i][j] >  MAX_ATOMS) bug(125);
                ordRemap[ordBlock[i][j]] = k;
              }
              ordBlock[i][j] = ordRemap[ordBlock[i][j]];
            }
          }

          /* First, compute the size of the new diagram */
          /* (extended notation code is from mmpsubset.c but not relevant
             since loop.c does not handle extended notation) */
          ordChars = 0;
          for (i = 1; i <= ordBlocks; i++) {
            for (j = 1; j <= ordBlockSize[i]; j++) {
              k = ordBlock[i][j];
              /*
              while (k > atomMapLen) {
                / * Handle extended notation * /
                ordChars++;
                k -= atomMapLen;
              }
              */
              ordChars++;
            }
            ordChars++;
          }
          let(&ordGreechieStmt, space(ordChars)); /* Preallocate string to
                                                     computed size */
          /* Next, fill in the characters */
          ordChars = 0;
          for (i = 1; i <= ordBlocks; i++) {
            for (j = 1; j <= ordBlockSize[i]; j++) {
              k = ordBlock[i][j];
              /*
              while (k > atomMapLen) {
                / * Handle extended notation * /
                greechieStmt[ordChars] = '+';
                ordChars++;
                k -= atomMapLen;
              }
              */
              ordGreechieStmt[ordChars] = ATOM_MAP[k - 1];
              ordChars++;
            }
            if (i < ordBlocks) {
              ordGreechieStmt[ordChars] = ',';
              ordChars++;
            } else {
              ordGreechieStmt[ordChars] = '.';
              ordChars++;
            }
          }
          if (ordGreechieStmt[ordChars] != 0) bug(127); /* Must be end of
                                                           string */
          ordGreechieStmtPrefix[0] = '=';
          print2("%s%s\n", ordGreechieStmtPrefix, ordGreechieStmt);
          let(&ordGreechieStmt, ""); /* Deallocate memory */
          let(&ordGreechieStmtPrefix, ""); /* Deallocate memory */

          /* End of 27-Sep-2009 mod */




        } /* if if (!instr(1, loops, loopstr)) */
        continue;
      } else {
        loop(str1);
      }
      let(&str1, "");
    } /* if (s == 0) - if we're looking for loop */


    /*******************/
    if (s == 1) {
      /* we're looking for biggest set drawable with
                       straight lines only */

      /* Look for duplicate blocks */
      f = 0;
      for (k = 0; k < strlen(sofar); k++) {
        if (trialBlock == (long)(sofar[k])) {
          f = 1;
          break;
        }
      }
      if (f == 1) continue;


      let(&str1, cat(sofar, chr(trialBlock), NULL));

      /* For straight lines - make sure that no other block shares
         all atoms with the blocks so far */
      g = 0;
      for (m = 1; m <= blocks; m++) {
        if (instr(1, str1, chr(m)) != 0) continue;
        for (n = 1; n <= blockSize[m]; n++) {
          g = 0;
          for (k = 0; k < strlen(str1); k++) {
            for (i = 1; i<= blockSize[(long)(str1[k])]; i++) {
              if (block[(long)(str1[k])][i] == block[m][n]) {
                g = 1;
                break;
              }
            }
            if (g == 1) break;
          }
          if (g == 0) break;
        }
        if (g == 1) break;
      }
      if (g == 1) continue;



      if (strlen(sofar) > biggestSoFar) { /* Found a biggest one drawable with
          straight lines */
        /*
        for (k = 0; k < strlen(str1); k++) {
          for (i = 1; i<= blockSize[(long)(str1[k])]; i++) {
            print2(" %ld", block[(long)(str1[k])][i]);
          }
          print2(" ");
        }
        if (g) print2(" g");
        print2("\n");
        */
        for (k = 0; k < strlen(str1); k++) {
          for (i = 1; i<= blockSize[(long)(str1[k])]; i++) {
            print2("%c", ATOM_MAP[block[(long)(str1[k])][i] - 1]);
          }
          if (k < strlen(str1) - 1) print2(",");
          if (k == strlen(str1) - 1) print2(".");
        }

        print2(" %ld", strlen(str1));
        print2("\n");
        biggestSoFar = strlen(sofar);
        continue;
      } else {
/*D*/if (blocks - strlen(sofar)>25)
/*D*/print2("%ld %ld %ld\n",(long)(sofar[0]),strlen(sofar),trialBlock);
        if (strlen(sofar) < blocks) loop(str1);
      }
      let(&str1, "");

    } /* if (s == 1) - if we're looking for biggest set drawable with
                       straight lines only */


  } /* next trialblock */

  /* Deallocate strings */
  let(&str1, "");
  let(&dummy, "");
  let(&printstr, "");
  let(&loopstr, "");
}



/* Returns 'T' if matrix test passed, 'F' if failed */
/* When 'F' is returned, the string failingAssignment has the assignment
   that violates the lattice */
unsigned char test(void)
{
  long i, j, k, n, conclusionLen, p, e, maxDepth, loopDepth;
  vstring trialConclusion = "";
  vstring trialHypList[MAX_HYPS];
  long loopVar[MAX_VARS];
  vstring fromPol = "";
  long hypLen[MAX_HYPS];
  long maxHypVar[MAX_HYPS];
  long maxConclusionVar;
  vstring tmpStr = "";
  vstring tmpStr2 = "";
  vstring tmpStr3 = "";
  vstring tmpStr4 = "";
  vstring tmpStr5 = "";
  vstring varOrder = "";
  char atLeastOneFailure = 0; /* Used in conjunction with showAllFailures */

  /* Local string array initialization */
  for (i = 0; i < hypotheses; i++) {
    trialHypList[i] = "";
  }

  /* Initialization */
  let(&trialConclusion, polConclusion);
  conclusionLen = strlen(polConclusion);
  for (i = 0; i < hypotheses; i++) {
    let(&(trialHypList[i]), polHypList[i]);
    hypLen[i] = strlen(polHypList[i]);
  }

  /* Build a list of all variables found to determine best scanning order */
  /* This way, failures in early hypotheses with few variables can be used
     to skip evaluations of later hypotheses and conclusion for speed-up */
  /* (We assume the user places hypotheses in order of fewest variables
     or most likelihood of failure first) */
  let(&varOrder, "");
  for (i = 0; i < hypotheses; i++) {
    maxHypVar[i] = -1;
    for (p = 1; p <= hypLen[i]; p++) {
      let(&tmpStr, mid(polHypList[i], p, 1));
      j = instr(1, VAR_LIST, tmpStr);
      if (j != 0) { /* It's a variable */
        if (instr(1, varOrder, tmpStr) == 0) { /* but not in list */
          let(&varOrder, cat(varOrder, tmpStr, NULL)); /* then add to list */
        }
        /* Find out where it is in the list */
        j = instr(1, varOrder, tmpStr) - 1;
        if (j < 0) bug(3);
        /* Save the highest variable in the hypothesis for later speedup */
        if (j > maxHypVar[i]) maxHypVar[i] = j;
      }
    }
  }
  maxConclusionVar = -1;
  for (p = 1; p <= conclusionLen; p++) {
    let(&tmpStr, mid(polConclusion, p, 1));
    j = instr(1, VAR_LIST, tmpStr);
    if (j != 0) { /* It's a variable */
      if (instr(1, varOrder, tmpStr) == 0) { /* but not in list */
        let(&varOrder, cat(varOrder, tmpStr, NULL)); /* then add to list */
      }
      /* Find out where it is in the list */
      j = instr(1, varOrder, tmpStr) - 1;
      if (j < 0) bug(4);
      /* Save the highest variable in the conclusion for later speedup */
      if (j > maxConclusionVar) maxConclusionVar = j;
    }
  }
  maxDepth = strlen(varOrder) - 1;


  /* Perform all possible evaluations */
  loopDepth = 0;
  loopVar[loopDepth] = 0;
  while (1) {
    if (loopVar[loopDepth] < nodes) { /* Loop not exhausted */
      /* Evaluate any hypotheses with less than current # variables */
      e = 'T';
      for (i = 0; i < hypotheses; i++) {
        if (maxHypVar[i] >= loopDepth) {
          /* Substitute lattice points for current variable in the
             hypothesis */
          for (p = 0; p < hypLen[i]; p++) {
            /* (If there are no variables at all, varOrder[0] will be 0
               i.e. end of string so we're still OK) */
            if ((polHypList[i])[p] == varOrder[loopDepth]) {
              (trialHypList[i])[p] = nodeNames[loopVar[loopDepth]];
            }
          }
        }
        if (maxHypVar[i] == loopDepth
            /* Special case: make sure we don't skip hypotheses
               with constants only, using the following OR condition: */
            || (maxHypVar[i] < 0 && loopDepth == 0) /* For hyp w/ no vars */
            ) {
          e = eval(trialHypList[i]);
          if (e != 'T') break; /* Out of i loop */
        }
      } /* Next i */
      /* Skip deeper iterations if hypothesis is false */
      if (e != 'T') goto nextIter; /* A hypothesis failed; next iteration */
      if (maxConclusionVar >= loopDepth) {
        /* Substitute lattice points for current variable in the conclusion */
        for (p = 0; p < conclusionLen; p++) {
          if (polConclusion[p] == varOrder[loopDepth]) {
            trialConclusion[p] = nodeNames[loopVar[loopDepth]];
          }
        }
      }
      /* Make iteration deeper unless deepest level */
      if (loopDepth < maxDepth) {
        loopDepth++;
        loopVar[loopDepth] = 0; /* Init var for next deeper level */
        continue;
      } else { /* loopDepth == maxDepth (or special case: loopDepth = 0
          and maxDepth = -1 if both hyp's & conclusion are all constants) */
        /* We're at the maximum variable depth; evaluate conclusion */
        e = eval(trialConclusion);
        if (e != 'T') {
          if (e != 'F') {print2("Bug - should be T or F\n"); exit(0);}
          /* Conclusion is false */
          atLeastOneFailure = 1;

          if (showVisits) {
            /* Show details of lattice failure evaluation (debugging mode) */
            printf("Details of lattice evaluation at failure:\n");
            visitFlag = 1;
            let(&visitList, ""); /* Initialize (redundant but just in case) */
            for (i = 0; i < hypotheses; i++) {
              print2("Hypothesis %ld\n", i + 1);

              /*e = eval(trialHypList[i]);*/ /* bad - doesn't show all nodes */
              /* Use unabbreviated version to see detailed visits */
              let(&tmpStr, "");
              tmpStr = unabbreviate(trialHypList[i]);
              e = eval(tmpStr);
              let(&tmpStr, "");

            } /* Next i */
            print2("Conclusion\n");

            /*e = eval(trialConclusion);*/ /* bad - doesn't show all nodes */
            /* Use unabbreviated version to see detailed visits */
            let(&tmpStr, "");
            tmpStr = unabbreviate(trialConclusion);
            e = eval(tmpStr);
            let(&tmpStr, "");

            /* Show nodes not visited (useful for identifying redundant
               lattice points) */
            let(&tmpStr2, "");
            for (i = 0; i < strlen(nodeNames); i++) {
              let(&tmpStr, chr(nodeNames[i]));
              if (!instr(1, visitList, tmpStr)) {
                let(&tmpStr2, cat(tmpStr2, " ", tmpStr, NULL));
              }
            }
            print2("Nodes not visited: %s\n", printableString(tmpStr2));
            /* For Greechie diagrams: show atoms not visited (useful for
               indentifying redundant atoms) */
            let(&tmpStr2, "");
            /* Form comma-separated greechie matrix - tmpStr3 */
            let(&tmpStr3, edit(greechieStmt, 8 + 16 + 128));
            i = 0;
            while (tmpStr3[i]) { /* Convert spaces to commas */
              if (tmpStr3[i] == ' ') tmpStr3[i] = ',';
              i++;
            }
            /* Form binary greechie matrix - tmpStr4 */
            j = numEntries(tmpStr3);
            let(&tmpStr4, space(j));
            for (i = 0; i < j; i++) {
              tmpStr4[i] = val(entry(i + 1, tmpStr3));
            }
            /* Form atom visit flag list - tmpStr5 */
            let(&tmpStr5, string(j, 1));
            /* Scan node visits to get Greechie diagram info */
            for (i = 0; i < strlen(lletterlist); i++) {
              let(&tmpStr, chr(lletterlist[i]));
              if (!instr(1, nodeNames, tmpStr)) break;
              if (!instr(1, visitList, tmpStr) &&
                  !instr(1, visitList, chr(uletterlist[i]))) {
                let(&tmpStr2, cat(tmpStr2, " ", str(i + 1), NULL));
                /* Flag unvisited atoms in Greechie matrix */
                for (k = 0; k < j; k++) {
                  if (tmpStr4[k] == i + 1) {
                    tmpStr5[k] = 2;
                  }
                }
              }
            }
            print2("Greechie lattice atoms not visited: %s\n", tmpStr2);
            let(&tmpStr2, "");
            for (i = 0; i < j; i = i + 3) {
              n = 0;
              for (k = i; k < i + 3; k++) {
                if (tmpStr5[k] == 2) n++;
              }
              if (n > 1) {
                if (tmpStr2[0] != 0) {
                  let(&tmpStr2, cat(tmpStr2, " ", NULL));
                }
                for (k = i; k < i + 3; k++) {
                  let(&tmpStr2, cat(tmpStr2, " ", entry(k + 1, tmpStr3),
                      NULL));
                }
              }
            }
            print2("Greechie blocks with 0 or 1 visits:%s\n", tmpStr2);

            visitFlag = 0;
            let(&visitList, ""); /* Deallocate */
          } /* if (showVisits) */

          /* Construct failing assignment info string */
          let(&failingAssignment, ""); /* Clear any earlier failure */
          for (i = 0; i < hypotheses; i++) {
            if (i > 0)
              let(&failingAssignment, cat(failingAssignment, " & ", NULL));
            let(&fromPol, "");
            fromPol = fromPolish(trialHypList[i]);
            /* Strip leading and trailing parenths */
            let(&fromPol, seg(fromPol, 2, strlen(fromPol) - 1));
            let(&failingAssignment, cat(failingAssignment, fromPol,
                NULL));
          }
          if (hypotheses > 0)
            let(&failingAssignment, cat(failingAssignment, " => ", NULL));
          let(&fromPol, "");
          fromPol = fromPolish(trialConclusion);
          /* Strip leading and trailing parenths */
          let(&fromPol, seg(fromPol, 2, strlen(fromPol) - 1));
          let(&failingAssignment, cat(failingAssignment, fromPol,
              NULL));
          /* Convert unprintable characters */
          let(&failingAssignment, printableString(failingAssignment));

          /* Normal (non-debugging) operation:  exit on first failure */
          if (!showAllFailures) goto done; /* Failed */

          /* Debugging mode: print the failure and try for another one */
          print2("FAILED %s at %s\n", latticeName, failingAssignment);

        } /* end if (e != 'T') */
      } /* end of loopDepth >= maxDepth case */
    } else { /* loopVar[loopDepth] >= nodes) - loop exhausted */
      /* End of inner loop; pop up one level */
      loopDepth--;
      if (loopDepth < 0) break; /* Completely done - out of while loop */
    } /* if (loopVar[loopDepth] < nodes) */
    /* Increment the loop variable */
   nextIter:
    loopVar[loopDepth]++;
  } /* while (1) */
 done:
  if (atLeastOneFailure) {
    e = 'F';
  } else {
    e = 'T';
  }
  let(&trialConclusion, ""); /* Deallocate string */
  let(&fromPol, ""); /* Deallocate string */
  let(&tmpStr, ""); /* Deallocate string */
  let(&tmpStr2, ""); /* Deallocate string */
  let(&tmpStr3, ""); /* Deallocate string */
  let(&tmpStr4, ""); /* Deallocate string */
  let(&tmpStr5, ""); /* Deallocate string */
  let(&varOrder, ""); /* Deallocate string */
  for (i = 0; i < hypotheses; i++) {
    let(&(trialHypList[i]), "");  /* Deallocate string */
  }
  return e;
} /* test() */


/* Converts all operators to v,^,- */
/* The caller must deallocate the result */
vstring unabbreviate(vstring eqn)
{
  long i;
  long sblen1, sblen2;
  vstring subEqn1 = "";
  vstring subEqn2 = "";
  vstring newSubEqn = "";
  vstring subform = "";
  vstring eqn1 = "";
  let(&eqn1, eqn); /* In case temp allocation passed in */
  /* Note: don't precompute strlen(eqn1), since eqn1 grows */
  for (i = 1; i <= strlen(eqn1); i++) {
    if (opMap[(unsigned char)(eqn1[i - 1])] < 127) {
      /* It's a binary operation - unabbreviate it */
      subEqn1 = subformula(right(eqn1, i + 1));
      sblen1 = strlen(subEqn1);
      subEqn2 = subformula(right(eqn1, i + 1 + sblen1));
      sblen2 = strlen(subEqn2);
      switch ((unsigned char)(eqn1[i - 1])) {
        /* I'm not sure whether it's helpful to do < & [ */
        case '>': /* greater than or equal:  x>y is y<x */
        case '<': /* less than or equal:  x<y is (xvy)=y */
          /*
          let(&newSubEqn, cat("=v", subEqn1, subEqn2, subEqn2, NULL));
          break;
          */
        case '[': /* commutes:  x[y is x=((x^y)v(x^-y)) */
          /*
          let(&newSubEqn, cat("=", subEqn1, "v^", subEqn1, subEqn2, "^",
              subEqn1, "-", subEqn2, NULL));
          break;
          */
        case 'v':
        case '^':
        case '-':
        case '=':
          let(&newSubEqn, cat(mid(eqn1, i, 1), subEqn1, subEqn2, NULL));
          break;
        case '#': /* # = biimplication: ((x^y)v(-x^-y))*/
          let(&newSubEqn, cat("v^", subEqn1, subEqn2, "^-",
              subEqn1, "-", subEqn2, NULL));
          break;
        case 'O': /* O = ->0 = classical arrow: (-xvy) */
          let(&newSubEqn, cat("v-", subEqn1, subEqn2, NULL));
          break;
        case 'I': /* I = ->1 = Sasaki arrow: (-xv(x^y)) */
          let(&newSubEqn, cat("v-", subEqn1, "^", subEqn1, subEqn2, NULL));
          break;
        case '2': /* 2 = ->2 = Dishkant arrow: (yv(-x^-y)) */
          let(&newSubEqn, cat("v", subEqn2, "^-", subEqn1, "-", subEqn2, NULL));
          break;
        case '3': /* 3 = ->3 = Kalmbach arrow: (((-x^y)v(-x^-y))v(x^(-xvy))) */
          let(&newSubEqn, cat("vv^-", subEqn1, subEqn2, "^-", subEqn1, "-",
              subEqn2, "^", subEqn1, "v-", subEqn1, subEqn2, NULL));
          break;
        case '4': /* 4 = ->4 = non-tollens arrow: (((x^y)v(-x^y))v((-xvy)^-y)) */
          let(&newSubEqn, cat("vv^", subEqn1, subEqn2, "^-", subEqn1, subEqn2,
              "^v-", subEqn1, subEqn2, "-", subEqn2, NULL));
          break;
        case '5': /* 5 = ->5 = relevance arrow: (((x^y)v(-x^y))v(-x^-y)) */
          let(&newSubEqn, cat("vv^", subEqn1, subEqn2, "^-", subEqn1, subEqn2,
              "^-", subEqn1, "-", subEqn2, NULL));
          break;
        default:
          bug(5);
      } /* end switch */
      let(&eqn1, cat(left(eqn1, i - 1), newSubEqn, right(eqn1, i + sblen1
          + sblen2 + 1), NULL));
      let(&subEqn1, ""); /* Deallocate from subformula() call */
      let(&subEqn2, ""); /* Deallocate from subformula() call */
    } /* if (opMap[eqn1[i]] < 127) */
  } /* next i */
  let(&newSubEqn, ""); /* Deallocate */
  let(&subform, ""); /* Deallocate */
  return eqn1;
} /* unabbreviate */

/* Returns a comma-separated list of all subformulas in an
   equation */
/* The caller must deallocate the result */
vstring subformulaList(vstring eqn)
{
  long i, j, k, p;
  long length;
  vstring subform = "";
  vstring list = "";
  vstring eqn1 = "";
  vstring e = "";
  vstring unabbrList = "";
  let(&eqn1, eqn); /* In case temp allocation passed in */
  length = strlen(eqn1);
  for (i = 1; i <= length; i++) {
    let(&subform, ""); subform = subformula(right(eqn1, i));
    /* Ignore negated subformulas to make list cleaner */
    while (subform[0] == NEG_OPER)
      let(&subform, right(subform, 2));
    /* If the subformula isn't in the list, add it */
    if (lookup(subform, list) == 0
        /* Also look at unabbreviated list, so we won't have abbreviated
           and unabbreviated duplicates (this assume abbreviated form of
           expression appears first in eqn1 - see findCommutingPairs()
           function) */
        && lookup(subform, unabbrList) == 0
        ) {
      if (list[0] == 0) {
        let(&list, subform);
      } else {
        /*let(&list, cat(list, ",", subform, NULL));*/
        /* Figure out where it should go - shortest first */
        j = numEntries(list);
        for (k = 1; k <= j; k++) {
          let(&e, entry(k, list));
          if (strlen(e) > strlen(subform) ||
              (strlen(e) == strlen(subform) && strcmp(e, subform) > 0)) {
            /* A longer one found - put it before */
            p = entryPosition(k, list);
            let(&list, cat(left(list, p - 1), subform, ",", right(list, p),
                NULL));
            break;
          }
        } /* next k */
        if (k > j) {
          /* It's the biggest - put at end of list */
          let(&list, cat(list, ",", subform, NULL));
        }
      }

      /* Add to unabbreviated list */
      let(&e, "");
      e = unabbreviate(subform);
      if (unabbrList[0] == 0) {
        let(&unabbrList, e);
      } else {
        let(&unabbrList, cat(unabbrList, ",", e, NULL));
      }

    }
  }
  let(&subform, "");
  let(&eqn1, "");
  let(&e, "");
  return list;
} /* subformulaList() */



/* Returns the char value of the nodeName character that the
   trial equation evaluates to */
unsigned char eval(vstring trialEqn)
{
  unsigned char e;
  vstring subEqn1 = "";
  vstring subEqn2 = "";

  /* Speedup - use the fast version */
  if (1) {
    return fastEval(trialEqn, 0);
  }

  if ((unsigned char)(trialEqn[0]) == NEG_OPER) {
    let(&subEqn1, right(trialEqn, 2));
    /*
    e = lookupCompl(eval(subEqn1));
    */
    /* Speedup */
    e = negMap[eval(subEqn1)];
  } else {
    /*
    if (instr(1, BIN_OPERS, chr(trialEqn[0])) != 0) {
    */
    /* Speedup */
    if (opMap[(unsigned char)(trialEqn[0])] < 127) {
      subEqn1 = subformula(right(trialEqn, 2));
      subEqn2 = subformula(right(trialEqn, strlen(subEqn1) + 2));
      /*
      e = lookupBinOp(trialEqn[0], eval(subEqn1), eval(subEqn2));
      */
      /* Speedup */
      e = binOpTable[opMap[(unsigned char)(trialEqn[0])]]
         [nodeNameMap[eval(subEqn1)]][nodeNameMap[eval(subEqn2)]];
    } else {
      /* Must be a node */
      e = (unsigned char)(trialEqn[0]);
    }
  }
  let(&subEqn1, "");  /* Deallocate vstring */
  let(&subEqn2, "");  /* Deallocate vstring */
  return e;

}

/* Returns the char value of the nodeName character that the
   trial equation evaluates to */
/* startChar = 0 is 1st char in string, 1 is 2nd, etc. */
unsigned char fastEval(vstring trialEqn, long startChar)
{
  unsigned char e;
  long i;
  vstring s1; /* Don't initialize for speedup */
  vstring s2; /* Don't initialize for speedup */

  if ((unsigned char)(trialEqn[startChar]) == NEG_OPER) {
    e = negMap[fastEval(trialEqn, startChar + 1)];
  } else {
    if (opMap[(unsigned char)(trialEqn[startChar])] < 127) {
      i = subformulaLen(trialEqn, startChar + 1);
      e = binOpTable[opMap[(unsigned char)(trialEqn[startChar])]]
         [nodeNameMap[fastEval(trialEqn, startChar + 1)]]
         [nodeNameMap[fastEval(trialEqn, startChar + i + 1)]];
    } else {
      /* Must be a node */
      e = (unsigned char)(trialEqn[startChar]);
    }
  }
  if (visitFlag) {
    /* Show user details of failing lattice visit */
    s1 = subformula(right(trialEqn, startChar + 1));
    s2 = fromPolish(s1);
    if (startChar == 0) {
      /* Ignore final result (will be 'F', which means "false" rather than
         lattice node F -- or 'T' for "true" in the case of hypothesis) */
      if (e != 'F' && e != 'T') bug(201);
      if (e == 'F') print2("(false) = %s\n", printableString(s2));
      if (e == 'T') print2("(true) = %s\n", printableString(s2));
    } else {
      print2("%s = %s\n", printableString(chr(e)), printableString(s2));
      let(&visitList, cat(visitList, chr(e), NULL));
    }
    let(&s1, ""); /* Deallocate */
    let(&s2, ""); /* Deallocate */
  }
  return e;
}

/* Returns the shortest subformula from beginning of equation */
/* The caller must deallocate the result */
vstring subformula(vstring eqn)
{
  vstring result = "";
  long i, p;
  let(&result, eqn); /* In case temp allocation passed in */
  i = 0;
  p = 1;
  while (p > 0) {
    /*
    if (instr(1, BIN_OPERS, chr(result[i])) != 0) {
    */
    /* Speedup */
    if (opMap[(unsigned char)(result[i])] < 127) {
      p++;
    } else {
      if ((unsigned char)(result[i]) != NEG_OPER) p--;  /* It is a node name */
    }
    i++;
  }
  let(&result, left(result, i));
  return result;
}


/* Returns the length of the shortest subformula from startChar */
long subformulaLen(vstring eqn, long startChar)
{
  long i, p;
  i = 0;
  p = 1;
  while (p > 0) {
    if (opMap[(unsigned char)(eqn[startChar + i])] < 127) {
      p++;
    } else {
      if ((unsigned char)(eqn[startChar + i]) != NEG_OPER) p--;
                                                        /* It is a node name */
    }
    i++;
  }
  return i;
}


unsigned char lookupCompl(unsigned char arg) {
  /* abc... = node names; ABC.. = complemented node names */
  /*
  if (arg == '0') return '1';
  if (arg == '1') return '0';
  if (isupper(arg)) return tolower(arg);
  if (islower(arg)) return toupper(arg);
  bug(6);
  return 0;
  */
  /* Speedup */
  return negMap[arg];
}

unsigned char lookupBinOp(unsigned char operation, unsigned char arg1,
    unsigned char arg2) {
    /*
    ^ = conjunction
    v = disjunction
    # = biimplication: ((x^y)v(-x^-y))
    O = ->0 = classical arrow: (-xvy)
    I = ->1 = Sasaki arrow: (-xv(x^y))
    2 = ->2 = Dishkant arrow: (-yI-x)
    3 = ->3 = Kalmbach arrow: (((-x^y)v(-x^-y))v(x^(-xvy)))
    4 = ->4 = non-tollens arrow: (-y3-x)
    5 = ->5 = relevance arrow: (((x^y)v(-x^y))v(-x^-y))

    = = equals predicate - returns T or F
    < = less-than-or-equal predicate:  ((xvy)=y) - returns T or F
    [ = commutes predicate:  (x=((x^y)v(x^-y))) - returns T or F
    */
  unsigned char result;
  switch (operation) {
    case 'v':
      /* Speedup:  eliminate instr() w/ char map */
      if (1) return (sup[nodeNameMap[arg1]][nodeNameMap[arg2]])[0];
      break;
    case '^':
      result = lookupCompl(lookupBinOp('v', lookupCompl(arg1),
          lookupCompl(arg2)));
      break;
    case '#':
      result = lookupBinOp('v',
                lookupBinOp('^', arg1, arg2),
                lookupCompl(lookupBinOp('v', arg1, arg2)));
      break;
    case 'O':
      result = lookupBinOp('v', lookupCompl(arg1), arg2);
      break;
    case 'I':
      result = lookupBinOp('v', lookupCompl(arg1),
                lookupBinOp('^', arg1, arg2));
      break;
    case '2':
      result = lookupBinOp('I', lookupCompl(arg2), lookupCompl(arg1));
      break;
    case '3':
      result = lookupBinOp('v',
               lookupBinOp('v',
                 lookupBinOp('^', lookupCompl(arg1), arg2),
                 lookupBinOp('^', lookupCompl(arg1), lookupCompl(arg2))),
               lookupBinOp('^', arg1,
                 lookupBinOp('v', lookupCompl(arg1), arg2)));
      break;
    case '4':
      result = lookupBinOp('3', lookupCompl(arg2), lookupCompl(arg1));
      break;
    case '5':
      result = lookupBinOp('v',
               lookupBinOp('v',
                 lookupBinOp('^', arg1, arg2),
                 lookupBinOp('^', lookupCompl(arg1), arg2)),
               lookupBinOp('^', lookupCompl(arg1), lookupCompl(arg2)));
      break;

    case '=':
      if (arg1 == arg2) {
        result = 'T';
      } else {
        result = 'F';
      }
      break;
    case '<': /* Less than or equal to */
      result = lookupBinOp('=', lookupBinOp('v', arg1, arg2), arg2);
      break;
    case '>': /* Greater than or equal to */
      result = lookupBinOp('<', arg2, arg1);
      break;
    case '[': /* Commutes */
      result = lookupBinOp('=', arg1, lookupBinOp('v',
                 lookupBinOp('^', arg1, arg2),
                 lookupBinOp('^', arg1, lookupCompl(arg2))));
      break;
    default:
      result = 0;
      bug(7);
  } /* switch (operation) */

  /* Uncomment for debugging */
  /*
  print2("%s %s %s = %s\n", chr(arg1), chr(operation), chr(arg2),
      chr(result));
  */
  return result;
}


/* ***Note:  The caller must deallocate returned string */
/* Convert normal to Polish notation */
vstring toPolish(vstring equation)
{
 /* This function converts a theorem in parentheses notation to
    one in Polish notation */

  vstring stack[MAX_STACK];
  long i, j, k, m, stackPtr, stackTop, level, p;
  vstring stackEntry = "";
  vstring polEqn = "";

  /* Initialize vstring array */
  /* Done below only as needed for speedup
  for (i = 0; i < MAX_STACK; i++) {
    stack[i] = "";
  }
  */

  /* Count parentheses for better user information */
  j = 0; k = 0; m = strlen(equation);
  for (i = 0; i < m; i++) {
    if (equation[i] == '(') j++;
    if (equation[i] == ')') k++;
  }
  if (j != k) {
    print2("?Error: There are %ld left but %ld right parentheses in %s\n",
        j, k, equation);
    exit(0);
  }

  stackPtr = 1;
  stackTop = 1;
  stack[stackTop] = ""; /* Initialize vstring array */
  let(&(stack[stackPtr]), equation);
  while (stackPtr <= stackTop) {
    let(&stackEntry, stack[stackPtr]);

    if (len(stackEntry) == 1) {
      /* The stack entry is a single character.  Only a variable
         or constant is allowed. */
      if (instr(1, cat(VAR_LIST, "01", NULL), stackEntry) == 0) {
        print2("%s\n", cat("?Error 1: Bad syntax in ",
            equation, NULL));
        exit(0);
      } /* end if */
      stackPtr++;
    } else { /* 1 */

    if (stackEntry[0] == NEG_OPER) {  /* Negation */
      if (stackTop >= MAX_STACK - 1) {
        print2(
            "?Error 2: Stack overflow - please increase MAX_STACK\n");
        exit(0);
      }
      stackTop++;
      stack[stackTop] = ""; /* Initialize vstring array */
      for (i = stackTop - 1; i >= stackPtr + 1; i--) { /* Push stack */
        let(&(stack[i + 1]), stack[i]);
      }
      let(&(stack[stackPtr]), chr(NEG_OPER));
      let(&(stack[stackPtr + 1]), right(stackEntry, 2));
      stackPtr++;
    } else { /* 2 */

    if (stackEntry[0] == '(') {  /* 2-argument operator */
      if (stackTop >= MAX_STACK - 2) {
        print2(
            "?Error 3: Stack overflow - please increase MAX_STACK\n");
        exit(0);
      }
      stackTop++;
      stack[stackTop] = ""; /* Initialize vstring array */
      stackTop++;
      stack[stackTop] = ""; /* Initialize vstring array */
      for (i = stackTop - 2; i >= stackPtr + 1; i--) {
        /* Push stack */
        let(&(stack[i + 2]), stack[i]);
      }
      if (stackEntry[strlen(stackEntry) - 1] != ')') {
        print2("%s\n", cat("?Error 4: Bad syntax in ",
            equation, NULL));
        exit(0);
      } /* end if */
      /* Find the operator */
      level = 0;
      p = 2;
      while (stackEntry[p - 1] == NEG_OPER) p++; /* Get past negations */

      if (stackEntry[p - 1] == '(') { /* Find closing parenthesis */
        level = 1;
        while (level > 0 && p < strlen(stackEntry)) {
          p++;
          if (stackEntry[p - 1] == '(') level++;
          else
            if (stackEntry[p - 1] == ')') level--;
        } /* end while */
      }
      p++;
      if (instr(1, BIN_OPERS, mid(stackEntry, p, 1)) == 0 ||
          p > strlen(stackEntry)) {
        print2("%s\n", cat("?Error 5: Bad syntax in ",
            equation, NULL));
        exit(0);
      } /* end if */
      let(&(stack[stackPtr]), mid(stackEntry, p, 1));
      let(&(stack[stackPtr + 1]), seg(stackEntry, 2, p - 1));
      let(&(stack[stackPtr + 2]), seg(stackEntry, p + 1,
          strlen(stackEntry) - 1));
      stackPtr++;
    } else { /* 3 */
       print2("%s\n", cat("?Error 6: Bad syntax in ",
           equation, NULL));
       exit(0);
    } } } /* 3 2 1 */
  } /* end while */

  for (i = 1; i <= stackTop; i++) {
    let(&polEqn, cat(polEqn, stack[i], NULL));
    let(&(stack[i]), ""); /* Deallocate vstring array */
  }

  /* Additional special error-checking for lattice algebra */
  if (instr(2, polEqn, "=") || instr(2, polEqn, "<") ||
      instr(2, polEqn, ">") || instr(2, polEqn, "[") ||
      (polEqn[0] != '=' && polEqn[0] != '<' &&
       polEqn[0] != '>' && polEqn[0] != '[') ) {
    print2("%s\n",
        cat("?Error 7: Relation inside of wff in ", equation, NULL));
    exit(0);
  }


  return polEqn;

} /* toPolish */



/* ***Note:  The caller must deallocate returned string */
/* Convert Polish to normal notation */
vstring fromPolish(vstring polishEqn)
{
  /* This function converts a theorem in Polish notation to one
     in parentheses notation */
#define MAX_STACK2 100
  vstring stack[MAX_STACK2];
  long i, stackPtr;
  long maxStack = 0;
  vstring stackEntry = "";
  vstring ch = "";
  vstring polishEqn1 = "";
  let(&polishEqn1, polishEqn); /* In case temp alloc string is passed in */

  stackPtr = 0;
  for (i = len(polishEqn1); i >=1; i--) {
    let(&ch, mid(polishEqn1, i, 1));
    if (ch[0] == NEG_OPER) {
      if (stackPtr > 0) {
        let(&(stack[stackPtr]), cat(ch, stack[stackPtr], NULL));
      } else {
        print2("?Error 8: Stack underflow at %ld\n", i);
      }
    } else {
      if (instr(1, BIN_OPERS, ch)) {
        if (stackPtr > 1) {
          stackPtr--;
          let(&(stack[stackPtr]), cat("(", stack[stackPtr + 1], ch,
              stack[stackPtr], ")", NULL));
        } else {
          print2("?Error 9:  Stack underflow at %ld\n", i);
        }
      } else { /* Variable assumed */
        stackPtr++;
        if (stackPtr > maxStack) {
          maxStack = stackPtr;
          stack[stackPtr] = "";
        }
        let(&(stack[stackPtr]), ch);
      }
    }
  } /* next i */
  if (stackPtr != 1) print2("?Error 10: Stack not emptied\n");
  /* Deallocate vstring array */
  for (i = 2; i <= maxStack; i++) {
    let(&(stack[i]), "");
  }
  let(&ch, "");
  let(&stackEntry, "");
  let(&polishEqn1, "");
  return (stack[stackPtr]);
} /* fromPolish */

/* Convert a printable string to internal upper ASCII */
/* This function converts ascii "&a", "&b", ... to 'a'+128, 'b'+128,... */
/* Use like any vstring function (left, right,...) */
vstring unPrintableString(vstring sin)
{
  vstring sout;
  long i, j, n;
  unsigned char c;
  j = strlen(sin);
  n = 0;
  for (i = 0; i < j; i++) if (sin[i] == '&') n++;
  sout = tempAlloc(j - n + 1);
  sout[j - n] = 0;
  n = 0;
  for (i = 0; i < j; i++) {
    if (sin[i] == '&') {
      c = sin[i + 1];
      c += 128;
      sout[i - n] = c;
      n++;
      i++;
    } else {
      sout[i - n] = sin[i];
    }
  }
  return sout;
}

/* Return a printable string */
/* This function converts ascii 'a'+128, 'b'+128,... to "&a", "&b", ... */
/* Use like any vstring function (left, right,...) */
vstring printableString(vstring sin)
{
  vstring sout;
  long i, j, k, n;
  j = strlen(sin);
  n = 0;
  for (i = 0; i < j; i++) {
    /* If k is eliminated below, it introduces a bug in the lcc compiler
       with the optimization option */
    k = (unsigned char)(sin[i]);
    if (k > 127) n++;
  }
  sout = tempAlloc(j + n + 1);
  sout[j + n] = 0;
  n = 0;
  for (i = 0; i < j; i++) {
    k = (unsigned char)(sin[i]);
    if (k > 127) {
      sout[i + n] = '&';
      n++;
      sout[i + n] = k - 128;
    } else {
      sout[i + n] = sin[i];
    }
  }
  return (sout);
}



/*****************************************************************************/
/*       Copyright (C) 2000  NORMAN D. MEGILL  <nm@alum.mit.edu>             */
/*             License terms:  GNU General Public License                    */
/*****************************************************************************/

/*34567890123456 (79-character line to adjust text window width) 678901234567*/
/*
mmvstr.h - VMS-BASIC variable length string library routines header
This is a collection of useful built-in string functions available in VMS BASIC.
*/

/**************************************************************************

Variable-length string handler
------------------------------

     This collection of string-handling functions emulate most of the
string functions of VMS BASIC.  The objects manipulated by these functions
are strings of a special type called 'vstring' which
have no pre-defined upper length limit but are dynamically allocated
and deallocated as needed.  To use the vstring functions within a program,
all vstrings must be initially set to the null string when declared or
before first used, for example:

        vstring string1 = "";
        vstring stringArray[] = {"","",""};

        vstring bigArray[100][10]; /- Must be initialized before using -/
        int i,j;
        for (i=0; i<100; i++)
          for (j=0; j<10; j++)
            bigArray[i][j] = ""; /- Initialize -/


     After initialization, vstrings should be assigned with the 'let(&'
function only; for example the statements

        let(&string1,"abc");
        let(&string1,string2);
        let(&string1,left(string2,3));

all assign the second argument to 'string1'.  The 'let(&' function must
not be used to initialize a vstring for the first time.

     The 'cat' function emulates the '+' concatenation operator in BASIC.
It has a variable number of arguments, and the last argument should always
be NULL.  For example,

        let(&string1,cat("abc","def",NULL));

assigns "abcdef" to 'string1'.  Warning: 0 will work instead of NULL on the
VAX but not on the Macintosh, so always use NULL.

     All other functions are generally used exactly like their BASIC
equivalents.  For example, the BASIC statement

        let string1$=left$("def",len(right$("xxx",2)))+"ghi"+string2$

is emulated in c as

        let(&string1,cat(left("def",len(right("xxx",2))),"ghi",string2,NULL));

Note that ANSII c does not allow "$" as part of an identifier
name, so the names in c have had the "$" suffix removed.

     The string arguments of the vstring functions may be either standard c
strings or vstrings (except that the first argument of the 'let(&' function
must be a vstring).  The standard c string functions may use vstrings or
vstring functions as their string arguments, as long as the vstring variable
itself (which is a char * pointer) is not modified and no attempt is made to
increase the length of a vstring.  Caution must be excercised when
assigning standard c string pointers to vstrings or the results of
vstring functions, as the memory space may be deallocated when the
'le(&t' function is next executed.  For example,

        char *stdstr; /- A standard c string pointer -/
         ...
        stdstr=left("abc",2);

will assign "ab" to 'stdstr', but this assignment will be lost when the
next 'let(&' function is executed.  To be safe, use 'strcpy':

        char stdstr1[80]; /- A fixed length standard c string -/
         ...
        strcpy(stdstr1,left("abc",2));

Here, of course, the user must ensure that the string copied to 'stdstr1'
does not exceed 79 characters in length.

     The vstring functions allocate temporary memory whenever they are called.
This temporary memory is deallocated whenever a 'let(&' assignment is
made.  The user should be aware of this when using vstring functions
outside of 'let(&' assignments; for example

        for (i=0; i<10000; i++)
          print2("%s\n",left(string1,70));

will allocate another 70 bytes or so of memory each pass through the loop.
If necessary, dummy 'let(&' assignments can be made periodically to clear
this temporary memory:

        for (i=0; i<10000; i++)
          {
          print2("%s\n",left(string1,70));
          let(&dummy,"");
          }

It should be noted that the 'linput' function assigns its target string
with 'let(&' and thus has the same effect as 'let(&'.

************************************************************************/


vstring tempAlloc(long size)    /* String memory allocation/deallocation */
{
  /* When "size" is >0, "size" bytes are allocated. */
  /* When "size" is 0, all memory previously allocated with this */
  /* function is deallocated. */
  /* EXCEPT:  When startTempAllocStack != 0, the freeing will start at
     startTempAllocStack. */
  int i;
  if (size) {
    if (tempAllocStackTop>=(MAX_ALLOC_STACK-1)) {
      print2("?Error: Temporary string stack overflow\n");
      bug(101);
    }
    if (!(tempAllocStack[tempAllocStackTop++]=malloc(size))) {
      print2("?Error: Temporary string allocation failed\n");
      bug(102);
    }
    return (tempAllocStack[tempAllocStackTop-1]);
  } else {
    for (i=startTempAllocStack; i<tempAllocStackTop; i++) {
      free(tempAllocStack[i]);
    }
    tempAllocStackTop=startTempAllocStack;
    return (NULL);
  }
}


/* Make string have temporary allocation to be released by next let() */
/* Warning:  after makeTempAlloc() is called, the genString may NOT be
   assigned again with let() */
void makeTempAlloc(vstring s)
{
    if (tempAllocStackTop>=(MAX_ALLOC_STACK-1)) {
      print2("?Error: Temporary string stack overflow\n");
      bug(103);
    }
    tempAllocStack[tempAllocStackTop++]=s;
}


void let(vstring *target,vstring source)        /* String assignment */
/* This function must ALWAYS be called to make assignment to */
/* a vstring in order for the memory cleanup routines, etc. */
/* to work properly.  If a vstring has never been assigned before, */
/* it is the user's responsibility to initialize it to "" (the */
/* null string). */
{
  long targetLength,sourceLength;

  sourceLength=strlen(source);  /* Save its length */
  targetLength=strlen(*target); /* Save its length */
  if (targetLength) {
    if (sourceLength) { /* source and target are both nonzero length */

      if (targetLength>=sourceLength) { /* Old string has room for new one */
        strcpy(*target,source); /* Re-use the old space to save CPU time */
      } else {
        /* Free old string space and allocate new space */
        free(*target);  /* Free old space */
        *target=malloc(sourceLength+1); /* Allocate new space */
        if (!*target) {
          print2("?Error: String memory couldn't be allocated\n");
          bug(104);
        }
        strcpy(*target,source);
      }

    } else {    /* source is 0 length, target is not */
      free(*target);
      *target= "";
    }
  } else {
    if (sourceLength) { /* target is 0 length, source is not */
      *target=malloc(sourceLength+1);   /* Allocate new space */
      if (!*target) {
        print2("?Error: Could not allocate string memory\n");
        bug(105);
      }
      strcpy(*target,source);
    } else {    /* source and target are both 0 length */
      *target= "";
    }
  }

  tempAlloc(0); /* Free up temporary strings used in expression computation */

}




vstring cat(vstring string1,...)        /* String concatenation */
#define MAX_CAT_ARGS 30
{
  va_list ap;   /* Declare list incrementer */
  vstring arg[MAX_CAT_ARGS];    /* Array to store arguments */
  long argLength[MAX_CAT_ARGS]; /* Array to store argument lengths */
  int numArgs=1;        /* Define "last argument" */
  int i;
  long j;
  vstring ptr;

  arg[0]=string1;       /* First argument */

  va_start(ap,string1); /* Begin the session */
  while ((arg[numArgs++]=va_arg(ap,char *)))
        /* User-provided argument list must terminate with 0 */
    if (numArgs>=MAX_CAT_ARGS-1) {
      print2("?Error: Too many cat() arguments\n");
      bug(106);
    }
  va_end(ap);           /* End var args session */

  numArgs--;    /* The last argument (0) is not a string */

  /* Find out the total string length needed */
  j=0;
  for (i=0; i<numArgs; i++) {
    argLength[i]=strlen(arg[i]);
    j=j+argLength[i];
  }
  /* Allocate the memory for it */
  ptr=tempAlloc(j+1);
  /* Move the strings into the newly allocated area */
  j=0;
  for (i=0; i<numArgs; i++) {
    strcpy(ptr+j,arg[i]);
    j=j+argLength[i];
  }
  return (ptr);

}


/* input a line from the user or from a file */
vstring linput(FILE *stream,vstring ask,vstring *target)
{
  /*
    BASIC:  linput "what";a$
    c:      linput(NULL,"what?",&a);

    BASIC:  linput #1,a$                        (error trap on EOF)
    c:      if (!linput(file1,NULL,&a)) break;  (break on EOF)

  */
  /* This function prints a prompt (if 'ask' is not NULL), gets a line from
    the stream, and assigns it to target using the let(&...) function.
    NULL is returned when end-of-file is encountered.  The vstring
    *target MUST be initialized to "" or previously assigned by let(&...)
    before using it in linput. */
  char f[10001]; /* Allow up to 10000 characters */
  if (ask) print2("%s",ask);
  if (stream == NULL) stream = stdin;
  if (!fgets(f,10000,stream)) {
    /* End of file */
    return NULL;
  }
  f[10000]=0;     /* Just in case */
  f[strlen(f)-1]=0;     /* Eliminate new-line character */
  /* Assign the user's input line */
  let(target,f);
  return *target;
}


/* Find out the length of a string */
long len(vstring s)
{
  return (strlen(s));
}


/* Extract sin from character position start to stop into sout */
vstring seg(vstring sin,long start,long stop)
{
  vstring sout;
  long len;
  if (start<1) start=1;
  if (stop<1) stop=0;
  len=stop-start+1;
  if (len<0) len=0;
  sout=tempAlloc(len+1);
  strncpy(sout,sin+start-1,len);
  sout[len]=0;
  return (sout);
}

/* Extract sin from character position start for length len */
vstring mid(vstring sin,long start,long len)
{
  vstring sout;
  if (start<1) start=1;
  if (len<0) len=0;
  sout=tempAlloc(len+1);
  strncpy(sout,sin+start-1,len);
/*??? Should db be substracted from if len > end of string? */
  sout[len]=0;
  return (sout);
}

/* Extract leftmost n characters */
vstring left(vstring sin,long n)
{
  vstring sout;
  if (n < 0) n = 0;
  sout=tempAlloc(n+1);
  strncpy(sout,sin,n);
  sout[n]=0;
  return (sout);
}

/* Extract after character n */
vstring right(vstring sin,long n)
{
  /*??? We could just return &sin[n-1], but this is safer for debugging. */
  vstring sout;
  long i;
  if (n<1) n=1;
  i = strlen(sin);
  if (n>i) return ("");
  sout = tempAlloc(i - n + 2);
  strcpy(sout,&sin[n-1]);
  return (sout);
}

/* Emulate VMS BASIC edit$ command */
vstring edit(vstring sin,long control)
#define isblank_(c) ((c==' ') || (c=='\t'))
{
/*
EDIT$
  Syntax
         str-vbl = EDIT$(str-exp, int-exp)
     Values   Effect
     1        Trim parity bits
     2        Discard all spaces and tabs
     4        Discard characters: CR, LF, FF, ESC, RUBOUT, and NULL
     8        Discard leading spaces and tabs
     16       Reduce spaces and tabs to one space
     32       Convert lowercase to uppercase
     64       Convert [ to ( and ] to )
     128      Discard trailing spaces and tabs
     256      Do not alter characters inside quotes

     (non-BASIC extensions)
     512      Convert uppercase to lowercase
     1024     Tab the line (convert spaces to equivalent tabs)
     2048     Untab the line (convert tabs to equivalent spaces)
     4096     Convert VT220 screen print frame graphics to -,|,+ characters
*/
  vstring sout;
  long i,j,k;
  int last_char_is_blank;
  int trim_flag,discardcr_flag,bracket_flag,quote_flag,case_flag;
  int alldiscard_flag,leaddiscard_flag,traildiscard_flag,reduce_flag;
  int processing_inside_quote=0;
  int lowercase_flag, tab_flag, untab_flag, screen_flag;
  unsigned char graphicsChar;

  /* Set up the flags */
  trim_flag=control & 1;
  alldiscard_flag=control & 2;
  discardcr_flag=control & 4;
  leaddiscard_flag=control & 8;
  reduce_flag=control & 16;
  case_flag=control & 32;
  bracket_flag=control & 64;
  traildiscard_flag=control & 128;
  quote_flag=control & 256;

  /* Non-BASIC extensions */
  lowercase_flag = control & 512;
  tab_flag = control & 1024;
  untab_flag = control & 2048;
  screen_flag = control & 4096; /* Convert VT220 screen prints to |,-,+
                                   format */

  /* Copy string */
  i = strlen(sin) + 1;
  if (untab_flag) i = i * 7;
  sout=tempAlloc(i);
  strcpy(sout,sin);

  /* Discard leading space/tab */
  i=0;
  if (leaddiscard_flag)
    while ((sout[i]!=0) && isblank_(sout[i]))
      sout[i++]=0;

  /* Main processing loop */
  while (sout[i]!=0) {

    /* Alter characters inside quotes ? */
    if (quote_flag && ((sout[i]=='"') || (sout[i]=='\'')))
       processing_inside_quote=~processing_inside_quote;
    if (processing_inside_quote) {
       /* Skip the rest of the code and continue to process next character */
       i++; continue;
    }

    /* Discard all space/tab */
    if ((alldiscard_flag) && isblank_(sout[i]))
        sout[i]=0;

    /* Trim parity (eighth?) bit */
    if (trim_flag)
       sout[i]=sout[i] & 0x7F;

    /* Discard CR,LF,FF,ESC,BS */
    if ((discardcr_flag) && (
         (sout[i]=='\015') || /* CR  */
         (sout[i]=='\012') || /* LF  */
         (sout[i]=='\014') || /* FF  */
         (sout[i]=='\033') || /* ESC */
         /*(sout[i]=='\032') ||*/ /* ^Z */ /* DIFFERENCE won't work w/ this */
         (sout[i]=='\010')))  /* BS  */
      sout[i]=0;

    /* Convert lowercase to uppercase */
    if ((case_flag) && (islower(sout[i])))
       sout[i]=toupper(sout[i]);

    /* Convert [] to () */
    if ((bracket_flag) && (sout[i]=='['))
       sout[i]='(';
    if ((bracket_flag) && (sout[i]==']'))
       sout[i]=')';

    /* Convert uppercase to lowercase */
    if ((lowercase_flag) && (islower(sout[i])))
       sout[i]=tolower(sout[i]);

    /* Convert VT220 screen print frame graphics to +,|,- */
    if (screen_flag) {
      graphicsChar = sout[i]; /* Need unsigned char for >127 */
      /* vt220 */
      if (graphicsChar >= 234 && graphicsChar <= 237) sout[i] = '+';
      if (graphicsChar == 241) sout[i] = '-';
      if (graphicsChar == 248) sout[i] = '|';
      /* vt100 */
      if (graphicsChar == 218 /*up left*/ || graphicsChar == 217 /*lo r*/
          || graphicsChar == 191 /*up r*/ || graphicsChar == 192 /*lo l*/)
        sout[i] = '+';
      if (graphicsChar == 196) sout[i] = '-';
      if (graphicsChar == 179) sout[i] = '|';
    }

    /* Process next character */
    i++;
  }
  /* sout[i]=0 is the last character at this point */

  /* Clean up the deleted characters */
  for (j=0,k=0; j<=i; j++)
    if (sout[j]!=0) sout[k++]=sout[j];
  sout[k]=0;
  /* sout[k]=0 is the last character at this point */

  /* Discard trailing space/tab */
  if (traildiscard_flag) {
    --k;
    while ((k>=0) && isblank_(sout[k])) --k;
    sout[++k]=0;
  }

  /* Reduce multiple space/tab to a single space */
  if (reduce_flag) {
    i=j=last_char_is_blank=0;
    while (i<=k-1) {
      if (!isblank_(sout[i])) {
        sout[j++]=sout[i++];
        last_char_is_blank=0;
      } else {
        if (!last_char_is_blank)
          sout[j++]=' '; /* Insert a space at the first occurrence of a blank */
        last_char_is_blank=1; /* Register that a blank is found */
        i++; /* Process next character */
      }
    }
    sout[j]=0;
  }

  /* Untab the line */
  if (untab_flag || tab_flag) {

    /*
    DEF FNUNTAB$(L$)      ! UNTAB LINE L$
    I9%=1%
    I9%=INSTR(I9%,L$,CHR$(9%))
    WHILE I9%
      L$=LEFT(L$,I9%-1%)+SPACE$(8%-((I9%-1%) AND 7%))+RIGHT(L$,I9%+1%)
      I9%=INSTR(I9%,L$,CHR$(9%))
    NEXT
    FNUNTAB$=L$
    FNEND
    */

    k = strlen(sout);
    for (i = 1; i <= k; i++) {
      if (sout[i - 1] != '\t') continue;
      for (j = k; j >= i; j--) {
        sout[j + 8 - ((i - 1) & 7) - 1] = sout[j];
      }
      for (j = i; j < i + 8 - ((i - 1) & 7); j++) {
        sout[j - 1] = ' ';
      }
      k = k + 8 - ((i - 1) & 7);
    }
  }

  /* Tab the line */
  if (tab_flag) {

    /*
    DEF FNTAB$(L$)        ! TAB LINE L$
    I9%=0%
    FOR I9%=8% STEP 8% WHILE I9%<LEN(L$)
      J9%=I9%
      J9%=J9%-1% UNTIL ASCII(MID(L$,J9%,1%))<>32% OR J9%=I9%-8%
      IF J9%<=I9%-2% THEN
        L$=LEFT(L$,J9%)+CHR$(9%)+RIGHT(L$,I9%+1%)
        I9%=J9%+1%
      END IF
    NEXT I9%
    FNTAB$=L$
    FNEND
    */

    i = 0;
    k = strlen(sout);
    for (i = 8; i < k; i = i + 8) {
      j = i;
      while (sout[j - 1] == ' ' && j > i - 8) j--;
      if (j <= i - 2) {
        sout[j] = '\t';
        j = i;
        while (sout[j - 1] == ' ' && j > i - 8 + 1) {
          sout[j - 1] = 0;
          j--;
        }
      }
    }
    i = k;
    /* sout[i]=0 is the last character at this point */
    /* Clean up the deleted characters */
    for (j = 0, k = 0; j <= i; j++)
      if (sout[j] != 0) sout[k++] = sout[j];
    sout[k] = 0;
    /* sout[k]=0 is the last character at this point */
  }

  return (sout);
}


/* Return a string of the same character */
vstring string(long n, char c)
{
  vstring sout;
  long j=0;
  if (n<0) n=0;
  sout=tempAlloc(n+1);
  while (j<n) sout[j++]=c;
  sout[j]=0;
  return (sout);
}


/* Return a string of spaces */
vstring space(long n)
{
  return (string(n,' '));
}


/* Return a character given its ASCII value */
vstring chr(long n)
{
  vstring sout;
  sout=tempAlloc(2);
  sout[0]= n & 0xFF;
  sout[1]=0;
  return(sout);
}


/* Search for string2 in string 1 starting at start_position */
long instr(long start_position,vstring string1,vstring string2)
{
   char *sp1,*sp2;
   long ls1,ls2;
   long found=0;
   if (start_position<1) start_position=1;
   ls1=strlen(string1);
   ls2=strlen(string2);
   if (start_position>ls1) start_position=ls1+1;
   sp1=string1+start_position-1;
   while ((sp2=strchr(sp1,string2[0]))!=0) {
     if (strncmp(sp2,string2,ls2)==0) {
        found=sp2-string1+1;
        break;
     } else
        sp1=sp2+1;
   }
   return (found);
}


/* Translate string in sin to sout based on table.
   Table must be 256 characters long!! <- not true anymore? */
vstring xlate(vstring sin,vstring table)
{
  vstring sout;
  long len_table,len_sin;
  long i,j;
  long table_entry;
  char m;
  len_sin=strlen(sin);
  len_table=strlen(table);
  sout=tempAlloc(len_sin+1);
  for (i=j=0; i<len_sin; i++)
  {
    table_entry= 0x000000FF & (long)sin[i];
    if (table_entry<len_table)
      if ((m=table[table_entry])!='\0')
        sout[j++]=m;
  }
  sout[j]='\0';
  return (sout);
}


/* Returns the ascii value of a character */
long ascii_(vstring c)
{
  return (long)((unsigned char)(c[0]));
}

/* Returns the floating-point value of a numeric string */
double val(vstring s)
{
  /*
  return (atof(s));
  */
  /* 12/10/98 - NDM - atof may corrupt memory when processing
     random character strings.
     The implementation below makes best guess of value of any
     random string, ignoring commas, etc. and tolerating numbers
     suffixed with sign.  "E" notation is not handled. */
  double v = 0;
  char signFound = 0;
  double power = 1.0;
  long i;
  /* Scan from lsd backwards to minimize rounding errors */
  for (i = strlen(s) - 1; i >= 0; i--) {
    switch (s[i]) {
      case '0': case '1': case '2': case '3': case '4':
      case '5': case '6': case '7': case '8': case '9':
        v = v + ((double)(s[i] - '0')) * power;
        power = 10.0 * power;
        break;
      case '.':
        v = v / power;
        power = 1.0;
        break;
      case '-':
        signFound = 1;
        break;
    }
  }
  if (signFound) v = - v;
  return v;
}


/* Returns current date as an ASCII string */
vstring date()
{
        vstring sout;
        struct tm *time_structure;
        time_t time_val;
        char *month[12];

        /* (Aggregrate initialization is not portable) */
        /* (It must be done explicitly for portability) */
        month[0]="Jan";
        month[1]="Feb";
        month[2]="Mar";
        month[3]="Apr";
        month[4]="May";
        month[5]="Jun";
        month[6]="Jul";
        month[7]="Aug";
        month[8]="Sep";
        month[9]="Oct";
        month[10]="Nov";
        month[11]="Dec";

        time(&time_val);                        /* Retrieve time */
        time_structure=localtime(&time_val); /* Translate to time structure */
        sout=tempAlloc(12);
        /* "%02d" means leading zeros with min. field width of 2 */
        sprintf(sout,"%d-%s-%02d",
                time_structure->tm_mday,
                month[time_structure->tm_mon],
                (int)((time_structure->tm_year) % 100)); /* Y2K */
        return(sout);
}

/* Return current time as an ASCII string */
vstring time_()
{
        vstring sout;
        struct tm *time_structure;
        time_t time_val;
        int i;
        char *format;
        char *format1="%d:%d %s";
        char *format2="%d:0%d %s";
        char *am_pm[2];
        /* (Aggregrate initialization is not portable) */
        /* (It must be done explicitly for portability) */
        am_pm[0]="AM";
        am_pm[1]="PM";

        time(&time_val);                        /* Retrieve time */
        time_structure=localtime(&time_val); /* Translate to time structure */
        if (time_structure->tm_hour>=12) i=1;
        else                             i=0;
        if (time_structure->tm_hour>12) time_structure->tm_hour-=12;
        if (time_structure->tm_hour==0) time_structure->tm_hour=12;
        sout=tempAlloc(12);
        if (time_structure->tm_min>=10)
          format=format1;
        else
          format=format2;
        sprintf(sout,format,
                time_structure->tm_hour,
                time_structure->tm_min,
                am_pm[i]);
        return(sout);

}


/* Return a number as an ASCII string */
vstring str(double f)
{
  /* This function converts a floating point number to a string in the */
  /* same way that %f in printf does, except that trailing zeroes after */
  /* the one after the decimal point are stripped; e.g., it returns 7 */
  /* instead of 7.000000000000000. */
  vstring s;
  long i;
  s = tempAlloc(50);
  sprintf(s, "%f", f);
  if (strchr(s, '.') != 0) {              /* the string has a period in it */
    for (i = strlen(s) - 1; i > 0; i--) { /* scan string backwards */
      if (s[i] != '0') break;             /* 1st non-zero digit */
      s[i] = 0;                           /* delete the trailing 0 */
    }
    if (s[i] == '.') s[i] = 0;            /* delete trailing period */
  }
  return (s);
}


/* Return a number as an ASCII string */
vstring num1(double f)
{
  return (str(f));
}


/* Return a number as an ASCII string surrounded by spaces */
vstring num(double f)
{
  return (cat(" ",str(f)," ",NULL));
}


/*** NEW FUNCTIONS ADDED 11/25/98 ***/

/* Emulate PROGRESS "entry" and related string functions */
/* (PROGRESS is a 4-GL database language) */

/* A "list" is a string of comma-separated elements.  Example:
   "a,b,c" has 3 elements.  "a,b,c," has 4 elements; the last element is
   an empty string.  ",," has 3 elements; each is an empty string.
   In "a,b,c", the entry numbers of the elements are 1, 2 and 3 (i.e.
   the entry numbers start a 1, not 0). */

/* Returns a character string entry from a comma-separated
   list based on an integer position. */
/* If element is less than 1 or greater than number of elements
   in the list, a null string is returned. */
vstring entry(long element, vstring list)
{
  vstring sout;
  long commaCount, lastComma, i, len;
  if (element < 1) return ("");
  lastComma = -1;
  commaCount = 0;
  i = 0;
  while (list[i] != 0) {
    if (list[i] == ',') {
      commaCount++;
      if (commaCount == element) {
        break;
      }
      lastComma = i;
    }
    i++;
  }
  if (list[i] == 0) commaCount++;
  if (element > commaCount) return ("");
  len = i - lastComma - 1;
  if (len < 1) return ("");
  sout = tempAlloc(len + 1);
  strncpy(sout, list + lastComma + 1, len);
  sout[len] = 0;
  return (sout);
}

/* Emulate PROGRESS lookup function */
/* Returns an integer giving the first position of an expression
   in a comma-separated list. Returns a 0 if the expression
   is not in the list. */
long lookup(vstring expression, vstring list)
{
  long i, exprNum, exprPos;
  char match;

  match = 1;
  i = 0;
  exprNum = 0;
  exprPos = 0;
  while (list[i] != 0) {
    if (list[i] == ',') {
      exprNum++;
      if (match) {
        if (expression[exprPos] == 0) return exprNum;
      }
      exprPos = 0;
      match = 1;
      i++;
      continue;
    }
    if (match) {
      if (expression[exprPos] != list[i]) match = 0;
    }
    i++;
    exprPos++;
  }
  exprNum++;
  if (match) {
    if (expression[exprPos] == 0) return exprNum;
  }
  return 0;
}


/* Emulate PROGRESS num-entries function */
/* Returns the number of items in a comma-separated list. */
long numEntries(vstring list)
{
  long i, commaCount;
  i = 0;
  commaCount = 0;
  while (list[i] != 0) {
    if (list[i] == ',') commaCount++;
    i++;
  }
  return (commaCount + 1);
}

/* Returns the character position of the start of the
   element in a list - useful for manipulating
   the list string directly.  1 means the first string
   character. */
/* If element is less than 1 or greater than number of elements
   in the list, a 0 is returned.  If entry is null, a 0 is
   returned. */
long entryPosition(long element, vstring list)
{
  long commaCount, lastComma, i;
  if (element < 1) return 0;
  lastComma = -1;
  commaCount = 0;
  i = 0;
  while (list[i] != 0) {
    if (list[i] == ',') {
      commaCount++;
      if (commaCount == element) {
        break;
      }
      lastComma = i;
    }
    i++;
  }
  if (list[i] == 0) {
    if (i == 0) return 0;
    if (list[i - 1] == ',') return 0;
    commaCount++;
  }
  if (element > commaCount) return (0);
  if (list[lastComma + 1] == ',') return 0;
  return (lastComma + 2);
}


void print2(char* fmt,...)
{
  /* This performs the same operations as printf, except that if a log file is
    open, the characters will also be printed to the log file. */
  va_list ap;
  char printBuffer[10001];

  va_start(ap, fmt);
  vsprintf(printBuffer, fmt, ap); /* Put formatted string into buffer */
  va_end(ap);

  printf("%s", printBuffer); /* Terminal */

  if (fplog != NULL) {
    fprintf(fplog, "%s", printBuffer);  /* Print to log file */
  }
  return;
}

/* Opens files with error message; opens output files with
   backup of previous version.   Mode must be "r" or "w". */
FILE *fSafeOpen(vstring fileName, vstring mode)
{
  FILE *fp;
  vstring prefix = "";
  vstring postfix = "";
  vstring bakName = "";
  vstring newBakName = "";
  long v;

  if (!strcmp(mode, "r")) {
    fp = fopen(fileName, "r");
    if (!fp) {
      print2("?Sorry, couldn't open the file \"%s\".\n", fileName);
    }
    return (fp);
  }

  if (!strcmp(mode, "w")) {
    /* See if the file already exists. */
    fp = fopen(fileName, "r");

    if (fp) {
      fclose(fp);

#define VERSIONS 9
      /* The file exists.  Rename it. */

#if defined __WATCOMC__ /* MSDOS */
      /* Make sure file name before extension is 8 chars or less */
      i = instr(1, fileName, ".");
      if (i) {
        let(&prefix, left(fileName, i - 1));
        let(&postfix, right(fileName, i));
      } else {
        let(&prefix, fileName);
        let(&postfix, "");
      }
      let(&prefix, cat(left(prefix, 5), "~", NULL));
      let(&postfix, cat("~", postfix, NULL));
      if (0) goto skip_backup; /* Prevent compiler warning */

#elif defined __GNUC__ /* Assume unix */
      let(&prefix, cat(fileName, "~", NULL));
      let(&postfix, "");

#elif defined THINK_C /* Assume Macintosh */
      let(&prefix, cat(fileName, "~", NULL));
      let(&postfix, "");

#elif defined VAXC /* Assume VMS */
      /* For debugging on VMS: */
      /* let(&prefix, cat(fileName, "-", NULL));
         let(&postfix, "-"); */
      /* Normal: */
      goto skip_backup;

#else /* Unknown; assume unix standard */
      /*if (1) goto skip_backup;*/  /* [if no backup desired] */
      let(&prefix, cat(fileName, "~", NULL));
      let(&postfix, "");

#endif


      /* See if the lowest version already exists. */
      let(&bakName, cat(prefix, str(1), postfix, NULL));
      fp = fopen(bakName, "r");
      if (fp) {
        fclose(fp);
        /* The lowest version already exists; rename all to lower versions. */

        /* If version VERSIONS exists, delete it. */
        let(&bakName, cat(prefix, str(VERSIONS), postfix, NULL));
        fp = fopen(bakName, "r");
        if (fp) {
          fclose(fp);
          remove(bakName);
        }

        for (v = VERSIONS - 1; v >= 1; v--) {
          let(&bakName, cat(prefix, str(v), postfix, NULL));
          fp = fopen(bakName, "r");
          if (!fp) continue;
          fclose(fp);
          let(&newBakName, cat(prefix, str(v + 1), postfix, NULL));
          rename(bakName, newBakName);
        }

      }
      let(&bakName, cat(prefix, str(1), postfix, NULL));
      rename(fileName, bakName);

      /***
      printLongLine(cat("The file \"", fileName,
          "\" already exists.  The old file is being renamed to \"",
          bakName, "\".", NULL), "  ", " ");
      ***/
    } /* End if file already exists */
   /*skip_backup:*/

    fp = fopen(fileName, "w");
    if (!fp) {
      print2("?Sorry, couldn't open the file \"%s\".\n", fileName);
    }

    let(&prefix, "");
    let(&postfix, "");
    let(&bakName, "");
    let(&newBakName, "");

    return (fp);
  } /* End if mode = "w" */

  bug(1510); /* Illegal mode */
  return(NULL);

}


/* Bug check */
void bug(int bugNum)
{
  print2("?Error: Program bug # %d\n", bugNum);
  exit(0);
}
