/* states01.c */
#define VERSION "2.0 30-Oct-2010"
/* 2.0 - 30-Oct-2010 - nm - add -e (continue if error found). */
/* 1.9 - 25-Apr-2010 - nm - add -c (test for critical diagram). */
/* 1.8 - 4-Apr-2010 - nm - fix bug 1015. */
/* 1.7 - 24-Apr-2009 - nm - added extended notation handling. */
/* 1.6 - 17-Apr-2009 - nm - fixed bug where the default option (i.e. neither
   -sc, -wc, nor -1.0) incorrectly yielded "Admits no {0,1} states" for
   the case of 1 block diagram e.g. "123.". */
/* 1.5 - 2-Apr-2008 - nm - increased MAX_BLOCKS from 64 to 128 */
/* 1.1 - 27-Oct-04 - nm - slight change in sorting criteria for improved
   speed; added -1.0 option for old version; added -v verbose option */

/* To run this program, type:
      states01 < file1 > file2
   where
      file1 = input file with MMP diagrams in Brendan McKay's format
      file2 = output file with {0,1} state existence information
   See  states01 --help  for more options and explanation.
*/

/*****************************************************************************/
/*       Copyright (C) 2010  NORMAN D. MEGILL  <nm at alum.mit.edu>          */
/*             License terms:  GNU General Public License                    */
/*****************************************************************************/


#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <time.h>
#include <ctype.h>

/***********************************************************************/
/************ Start of "vstring" header stuff **************************/
/************ Do not touch anything in this section ********************/
/***********************************************************************/
typedef char* vstring;

/* String assignment - MUST be used to assign vstrings */
void let(vstring *target,vstring source);
/* String concatenation - last argument MUST be NULL */
vstring cat(vstring string1,...);

/* Emulate BASIC linput statement; returns NULL if EOF */
/* Note that linput assigns target string with let(&target,...) */
  /*
    BASIC:  linput "what";a$
    c:      linput(NULL,"what?",&a);

    BASIC:  linput #1,a$                        (error trap on EOF)
    c:      if (!linput(file1,NULL,&a)) break;  (break on EOF)

  */
vstring linput(FILE *stream,vstring ask,vstring *target);

/* Emulation of BASIC string functions */
vstring seg(vstring sin, long p1, long p2);
vstring mid(vstring sin, long p, long l);
vstring left(vstring sin, long n);
vstring right(vstring sin, long n);
vstring edit(vstring sin, long control);
vstring space(long n);
vstring string(long n, char c);
vstring chr(long n);
vstring xlate(vstring sin, vstring control);
vstring date(void);
vstring time_(void);
vstring num(double x);
vstring num1(double x);
vstring str(double x);
long len(vstring s);
long instr(long start, vstring sin, vstring s);
long ascii_(vstring c);
double val(vstring s);
/* Emulation of PROGRESS string functions added 11/25/98 */
vstring entry(long element, vstring list);
long lookup(vstring expression, vstring list);
long numEntries(vstring list);
long entryPosition(long element, vstring list);
/* Print to log file as well as terminal if fplog opened */
void print2(char* fmt,...);
FILE *fplog = NULL;
/* Opens files with error message; opens output files with
   backup of previous version.   Mode must be "r" or "w". */
FILE *fSafeOpen(vstring fileName, vstring mode);
/* Bug check error */
void bug(int bugNum);
/* End of functions you should call directly */


/* Do not call the ones below directly */
/******* Special pupose routines for better
      memory allocation (use with caution) *******/
/* Make string have temporary allocation to be released by next let() */
/* Warning:  after makeTempAlloc() is called, the vstring may NOT be
   assigned again with let() */
void makeTempAlloc(vstring s);   /* Make string have temporary allocation to be
                                    released by next let() */
/* Remaining prototypes (outside of mmvstr.h) */
char *tempAlloc(long size);     /* String memory allocation/deallocation */

#define MAX_ALLOC_STACK 100
int tempAllocStackTop=0;        /* Top of stack for tempAlloc functon */
int startTempAllocStack=0;      /* Where to start freeing temporary allocation
                                    when let() is called (normally 0, except in
                                    special nested vstring functions) */
char *tempAllocStack[MAX_ALLOC_STACK];


/*****************************************************************************/
/*********************** End of "vstring" header stuff ***********************/
/*****************************************************************************/

/* Constants */

/* Mapping for MMP diagram atoms */
#define ATOM_MAP "123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrs" \
    "tuvwxyz!\"#$%&'()*-/:;<=>?@[\\]^_`{|}~"
/* Maximum number of atoms - increase as needed, at expense of memory */
#define MAX_ATOMS 1000
/* Maximum number of blocks - increase as needed, at expense of memory */
#define MAX_BLOCKS 200
/* Minimum block size */
#define MIN_BLOCK_SIZE 2
/* Maximum block size - increase as needed, at expense of memory */
#define MAX_BLOCK_SIZE 10

/* Global variables */
char oneLineDisplay = 0;
char verboseMode = 0;
char worstCaseAlgorithm = 0;
char version1_0Algorithm = 0;
char skipClusterSortAlgorithm = 0;
char criticalTestFlag = 0;
char noErrorCheck = 0;
char noAbortOnError = 0;
long lattices = 0;
long atomMapLen;
long block[MAX_BLOCKS + 1][MAX_BLOCK_SIZE + 1];
long blockSize[MAX_BLOCKS + 1];
long blocks;
long atoms;
long totalBacktrackCount = 0; /* For user information */
/* saveBlock... stuff for the criticalTestingFlag mode */
long saveBlock[MAX_BLOCKS + 1][MAX_BLOCK_SIZE + 1];
long saveBlockSize[MAX_BLOCKS + 1];
long saveBlocks;


/* Prototypes */
vstring state01(vstring glattice);
char state01Test(long *backtrackCount);




/******************** Main program *******************************************/

int main(int argc, char *argv[])
{

  /* Integer variable declarations */

  /* This is how you declare some strings you want to work with */
  /* They MUST be initialized to the empty string, never to anything else */
  vstring str1 = "";
  vstring str2 = "";
  long arg;

  /* if (strlen(ATOM_MAP) != MAX_ATOMS) bug(1); */
  atomMapLen = strlen(ATOM_MAP); /* Do here to speed up its reuse */

  for (arg = 1; arg < argc; arg++) {
    if (!strcmp(argv[arg], "-1")) {
      oneLineDisplay = 1;
    } else if (!strcmp(argv[arg], "-e")) {
      noAbortOnError = 1;
    } else if (!strcmp(argv[arg], "-ne")) {
      noErrorCheck = 1;
    } else if (!strcmp(argv[arg], "-v")) {
      verboseMode = 1;
    } else if (!strcmp(argv[arg], "-wc")) {
      worstCaseAlgorithm = 1;
    } else if (!strcmp(argv[arg], "-1.0")) {
      version1_0Algorithm = 1;
    } else if (!strcmp(argv[arg], "-sc")) {
      skipClusterSortAlgorithm = 1;
    } else if (!strcmp(argv[arg], "-c")) {
      criticalTestFlag = 1;
    } else if (!strcmp(argv[arg], "--help")) {
printf("states01.c  Version %s\n", VERSION);
printf("To run this program, type:\n");
/*
printf("   states01 < file1 > file2\n");
*/
printf("   states01 [-1] [-ne] [-sc] [-wc] < file1 > file2\n");
printf("where:\n");
printf(
"   -1 = display 1-line output for use with Unix pipe filters (formatted\n");
printf(
"        per Mladen Pavicic: 'fails' means 'admits no {0,1} state')\n");
printf(
"   -e = do not abort on error; instead, output '...error...' and continue\n");
printf(
"   -ne = skip some error checking for speedup\n");
printf(
"   -v = verbose mode for debugging\n");
printf(
"   -sc = skip cluster sort algorithm completely\n");
printf(
"   -wc = worst-case (vs. best-case) cluster sort algorithm for debugging\n");
printf(
"   -1.0 = use Version 1.0 cluster sort algorithm\n");
printf(
"   -c = pass/fail means diagram is/is not critical i.e. it fails, and\n");
printf(
"        removing any block makes it pass.\n");
printf(
"   file1 = input file with diagrams in Brendan McKay's format\n");
printf("   file2 = output file with {0,1} state information\n");
/*
printf("Only one of -n or -s may be specified.\n");
*/
printf("For this help message, type:  states01 --help\n");
printf("\n");
printf(
"Purpose:  This program determines whether an MMP diagram admits a {0,1}\n");
printf(
"(non-dispersive) state.  The MMP input file notation is the same as\n");
printf(
"for Greechie diagrams described in the help for the program latticeg.c\n");
printf("\n");
printf("Example of use:\n");
printf("  states01 < test.o\n");
printf("where test.o contains the two lines:\n");
printf("  1234.\n");
printf("  FHJ,EGJ,DIJ,9AI,6BE,68A,5CF,579,4BC,38H,27G,14D,123.\n");
printf("corresponding to two diagrams, the first admitting {0,1}-states and\n");
printf("the second one not admitting them.\n");
      goto return_point;
    } else {
      fprintf(stderr,
          "?Unrecognized option \"%s\".  Type \"states01 --help\" for usage\n",
          argv[arg]);
      exit(1);
    }
  }


  while (1) {
    /* Get line from 1st file */
    if (linput(NULL, NULL, &str1) == NULL) break; /* NULL means EOF */
    /* Clean off carriage return (for Windows files under Cygwin) and spaces */
    let(&str1, edit(str1, 2 + 4));
    /*let(&str2, "");*/
    lattices++;
    str2 = state01(str1); /* Must always return empty string */
    if (str2[0] != 0) bug(2); /* ...to make sure */
    /*printf("%s\n", str2);*/
  }

  if (!oneLineDisplay) {
    printf("Total diagrams = %ld  Total backtrack count = %ld",
        lattices, totalBacktrackCount);
#ifdef CLOCKS_PER_SEC
    printf("  CPU time =%6.2f s", (double)((1.0 * clock())/CLOCKS_PER_SEC));
#endif
    printf("\n");
  }

 return_point:
  /* Deallocate vstring memory */
  let(&str1, "");
  let(&str2, "");

  return 0;
} /* End of main() */


/* Caller must deallocate returned string. */
vstring state01(vstring glattice1) {
  long i, j, k, n;
  vstring jptr;
  long extendedNotationIncr; /* For + notation */
  long extendedNotationOffset; /* For + notation */
  /*vstring glattice1 = "";*/
  vstring str1 = "";
  long backtrackCount = 0; /* Returned statistic from state01Test */
  char result; /* Returned value of state01Test: 0 = has state; 1 = no state;
                                                           2 = error occured */

  result = 0; /* Default to error condition until determined otherwise */
  /* extendedNotationIncr = strlen(ATOM_MAP); */ /* For + notation */
  extendedNotationIncr = atomMapLen; /*  For + notation (faster than strlen) */

  /*let(&glattice1, glattice);*/
  /*let(&glattice1, edit(glattice, 2));*/ /* Remove spaces */

  if (!oneLineDisplay) printf("#%ld %s\n", lattices, glattice1);

  n = strlen(glattice1);

  if (!noErrorCheck) {
    /* The calling routine should ensure this */
    if (strchr(glattice1, ' ') != NULL) bug(3);

    /* glattice1 can't be a blank line */
    if (glattice1[0] == 0) {
      fprintf(stderr, "#%ld: %s\n", lattices, glattice1);
      fprintf(stderr, "?Error: Blank lines are not allowed\n");
      if (noAbortOnError) goto ERROR_RETURN; else exit(1);
    }

    /* glattice1 must have ending period for new (Brendan) compact standard */
    if (glattice1[n - 1] != '.') {
      fprintf(stderr, "#%ld: %s\n", lattices, glattice1);
      fprintf(stderr, "?Error: Last character should be a period\n");
      if (noAbortOnError) goto ERROR_RETURN; else exit(1);
    }

    /* if (instr(1, left(glattice1, n - 1), ".") != 0) { */
    if (strchr(glattice1, '.') != glattice1 + n - 1) {
      fprintf(stderr, "#%ld: %s\n", lattices, glattice1);
      fprintf(stderr, "?Error: Period can only be last character\n");
      if (noAbortOnError) goto ERROR_RETURN; else exit(1);
    }
    if (n == 1) {
      fprintf(stderr, "#%ld: %s\n", lattices, glattice1);
      fprintf(stderr, "?Error: Diagram must have at least one block\n");
      if (noAbortOnError) goto ERROR_RETURN; else exit(1);
    }
  }

  atoms = 0;
  blocks = 1;
  blockSize[blocks] = 0;
  extendedNotationOffset = 0; /* For + notation */
  for (i = 0; i < n; i++) {
    if (glattice1[i] == ',' || glattice1[i] == '.') {
      /* End of block */
      if (blockSize[blocks] < MIN_BLOCK_SIZE) {
        fprintf(stderr, "#%ld: %s\n", lattices, glattice1);
        fprintf(stderr,
            "?Error: Block %ld has %ld atoms, but minimum block size is %ld\n",
             blocks, blockSize[blocks], (long)MIN_BLOCK_SIZE);
        if (noAbortOnError) goto ERROR_RETURN; else exit(1);
      }
      if (glattice1[i] == ',') {
        /* Start of new block */
        blocks++;
        if (blocks > MAX_BLOCKS) {
          fprintf(stderr, "#%ld: %s\n", lattices, glattice1);
          fprintf(stderr,
   "?Error: Maximum blocks allowed is %ld.  Increase MAX_BLOCKS in program.\n",
                (long)MAX_BLOCKS);
          if (noAbortOnError) goto ERROR_RETURN; else exit(1);
        }
        blockSize[blocks] = 0;
      }
      continue;
    }

    if (glattice1[i] == '+') {
      /* Process the extended notation
         12...9A...Za...`{|}~+1+2...+|+}+~++1...++~+++1....
         From 23-Apr-2009 email to Mladen:
           "If in the far future we get to say 10000 atoms, that would be
           around 100 +'s per atom, obviously extremely inefficient.  But,
           we still have "0" unused, and can have an alternate (and
           compatible) notation where "0" is the start of a decimal number,
           with some non-digit, say ".", terminating it.  I'll leave a
           comment to that effect in the latticeg.c file, for when it
           becomes a problem for a future generation" */
      extendedNotationOffset += extendedNotationIncr;
      continue;
    }

    /* Get the atom number */
    /*j = instr(1, ATOM_MAP, chr(glattice1[i]));*/
    /*let(&str1, "");*/ /* Deallocate chr call */
    jptr = strchr(ATOM_MAP, glattice1[i]);
    if (jptr == NULL) {
      fprintf(stderr, "#%ld: %s\n", lattices, glattice1);
      fprintf(stderr, "?Error: Illegal character '%c' in diagram\n",
          glattice1[i]);
      if (noAbortOnError) goto ERROR_RETURN; else exit(1);
    }
    j = jptr - ATOM_MAP + 1; /* Atom number */
    j += extendedNotationOffset; /* For + notation */
    extendedNotationOffset = 0; /* For + notation - initialize for next atom */
    blockSize[blocks]++;
    if (blockSize[blocks] > MAX_BLOCK_SIZE) {
      fprintf(stderr, "#%ld: %s\n", lattices, glattice1);
      fprintf(stderr,
           "?Error: Block %ld has %ld atoms, but maximum block size is %ld \n",
           blocks, blockSize[blocks], (long)MIN_BLOCK_SIZE);
      if (noAbortOnError) goto ERROR_RETURN; else exit(1);
    }
    /* Assign the atom */
    block[blocks][blockSize[blocks]] = j;
    if (j > atoms) atoms = j; /* Maximum atom number */
  } /* next i */


  if (!noErrorCheck) {
    for (i = 1; i <= blocks; i++) {
      for (j = 1; j <= blockSize[i] - 1; j++) {
        for (k = j + 1; k <= blockSize[i]; k++) {
          if (block[i][j] == block[i][k]) {
            fprintf(stderr, "#%ld: %s\n", lattices, glattice1);
            fprintf(stderr,
                "?Error: Duplicate atom numbers in a block\n");
            if (noAbortOnError) goto ERROR_RETURN; else exit(1);
          }
        }
      }
    }
  }

  if (!criticalTestFlag) { /* Normal testing */
    result = state01Test(&backtrackCount);
    totalBacktrackCount += backtrackCount;
    if (oneLineDisplay) {
        /* #16 a32-b34 ((37)) passes:: 8HP,9KP,25A,23L,BCQ,5DN,7CL,67F,... */
        printf("#%ld a%ld-b%ld ((%ld)) %s:: %s\n", lattices, atoms,
              blocks, backtrackCount,
            result ? "fails" : "passes", glattice1);
    } else {
      printf("#%ld Backtrack count = %ld\n", lattices, backtrackCount);
      if (result) {
        printf("#%ld (atoms%ld-blocks%ld)%s\n", lattices, atoms,
              blocks, " Admits no {0,1} states");
      } else {
        printf("#%ld (atoms%ld-blocks%ld)%s\n", lattices, atoms,
              blocks, " Admits at least one {0,1} state");
      }
    }
  } else {
    /* Test for critical diagram:  the diagram must fail {0,1} assignment,
       but it must pass {0,1} assignment if any single block is removed */
    result = state01Test(&backtrackCount);
    totalBacktrackCount += backtrackCount;
    if (!result) { /* The original diagram didn't fail, so it isn't critical */
      if (oneLineDisplay) {
          /* #16 ((37)) passes:: 8HP,9KP,25A,23L,BCQ,5DN,7CL,9EN,67F,... */
          printf("#%ld a%ld-b%ld ((%ld)) %s:: %s\n", lattices, atoms,
              blocks, backtrackCount,
              "fails (admits state)", glattice1);
      } else {
        printf("#%ld Backtrack count = %ld\n", lattices, backtrackCount);
        printf("#%ld (atoms%ld-blocks%ld)%s\n", lattices, atoms, blocks,
            " is not critical because it admits at least one {0,1} state");
      }
    } else {
      /* The diagram doesn't admit a {0,1} state; now check that after any
         block removed, it does admit a {0,1} state. */
      /* First, save the original diagram */
      saveBlocks = blocks;
      for (i = 1; i <= blocks; i++) {
        saveBlockSize[i] = blockSize[i];
        for (j = 1; j <= blockSize[i]; j++) {
          saveBlock[i][j] = block[i][j];
        }
      }
      /* Next, remove each block and test again */
      for (n = 1; n <= saveBlocks; n++) {
        blocks = saveBlocks - 1;
        for (i = 1; i <= blocks; i++) {
          if (i < n) blockSize[i] = saveBlockSize[i];
          else blockSize[i] = saveBlockSize[i + 1];
          for (j = 1; j <= blockSize[i]; j++) {
            if (i < n) block[i][j] = saveBlock[i][j];
            else block[i][j] = saveBlock[i + 1][j];
          }
        }
        /* We assume that state01Test() can tolerate atom numbering gaps
           (I think it does), so we don't bother to renumber the atoms
           to remove gaps */
        result = state01Test(&backtrackCount);
        totalBacktrackCount += backtrackCount;
        if (result) break; /* A state couldn't be assigned; therefore
                              it isn't critical */
      } /* next n (next block removed) */
      /* We don't bother to restore from saveBlocks etc., since they
         aren't used again. */
      if (oneLineDisplay) {
          /* #16 ((37)) passes:: 8HP,9KP,25A,23L,BCQ,5DN,7CL,9EN,67F,... */
          printf("#%ld a%ld-b%ld ((%ld)) %s:: %s\n", lattices, atoms,
              saveBlocks, backtrackCount,
              result ? "fails (not critical)" : "passes (is critical)",
              glattice1);
      } else {
        printf("#%ld Backtrack count = %ld\n", lattices, backtrackCount);
        if (result) {
          printf("#%ld (atoms%ld-blocks%ld)%s\n", lattices, atoms,
              saveBlocks, " is not critical");
        } else {
          printf("#%ld (atoms%ld-blocks%ld)%s\n", lattices, atoms,
              saveBlocks, " is critical");
        }
      }
    } /* else the parent diagram does not admit a 01 state */
  } /* else we're doing the special critical test */



  /* Deallocate strings */
  /*let(&glattice1, "");*/

  /* The caller must deallocate str1 */
  return str1;

 ERROR_RETURN:  /* Gets here is noAbortOnError is set and an error occurred */
  if (oneLineDisplay) {
      /* #16 a32-b34 ((37)) passes:: 8HP,9KP,25A,23L,BCQ,5DN,7CL,67F,... */
      printf("#%ld a%ld-b%ld ((%ld)) %s:: %s\n", lattices, atoms,
            blocks, backtrackCount,
          "error", glattice1);
  } else {
    printf("#%ld Backtrack count = %ld\n", lattices, backtrackCount);
    printf("#%ld (atoms%ld-blocks%ld)%s\n", lattices, atoms,
          blocks, " An error occurred");
  }
  /* The caller must deallocate str1 */
  return str1;


} /* states01 */


/* states01.c */
/* 7/25/03 */
/* Returns 0 if there is a {0,1} state, 1 if there is no {0,1} state */
char state01Test(long *backtrackCount)
{
  long i, j, k, l, m, n;
  char found;
  long blockSort[MAX_BLOCKS + 1]; /* Sort # vs. block # */
  long reverseBlockSort[MAX_BLOCKS + 1];
      /* Block # vs. sort #; 0 means block not sorted yet */
  long sortedBlockSize[MAX_BLOCKS + 1];
      /* Same as blockSize[] but sorted by clustering routine */
  long sortedBlock[MAX_BLOCKS + 1][MAX_BLOCKS + 1];
      /* Same as block[][] but sorted by clustering routine */
  long atomCommittedBy[MAX_ATOMS + 1];
      /* 0 means atom has is available for assignment */
      /* >0 means sorted block entry that first assigned atom */
  signed char atomValue[MAX_ATOMS + 1];  /* 0 or 1 or -1 if unassigned */
  long blockConnectedSize[MAX_BLOCKS + 1];
      /* Size of the block if unconnected atoms are removed */
  char blockAtomConnected[MAX_BLOCKS + 1][MAX_BLOCKS + 1];
      /* If 1, it means the atom in the block is connected to another block */

  long blocksConnected[MAX_ATOMS + 1];
      /* Number of blocks connected to this atom */
  long connectedBlockList[MAX_ATOMS + 1][MAX_BLOCKS + 1];
      /* List of the blocks connected to this atom */

  /* Variables for block "tightness" sorting */
  long maxBlockConnections;
  long maxConnectedBlock;
  long maxConnectedBlockSize;
  long thisBlockConnections;

  /* Variables for main backtracking scan */
  long lastAtomTried[MAX_BLOCKS + 1];
      /* The latest atom assigned to 1 for the sorted block # */
  char unconnectedAtomWasTried[MAX_BLOCKS + 1];
      /* Flag that we've already tried an unconnected atom, so don't again */
  char retVal;  /* Return value: 0 if {0,1} state found, 1 if not */
  char backtrackAgain;
  long atom1;
  char conflict;
  long atom;
  long connectedBlock;
  long onesInBlock;
  long unassignedInBlock;

  long backtrackCountx = 0; /* For informational purposes */
  long p, q;       /* For verbose mode */
  long iter = 0;   /* For verbose mode */
  long v;          /* For verbose mode */
  vstring tmp="";  /* For verbose mode */

  for (i = 1; i <= atoms; i++) {
    blocksConnected[i] = 0;
  }
  /* Scan blocks to determine blocks atoms are connected to */
  for (i = 1; i <= blocks; i++) {
    blockConnectedSize[i] = blockSize[i];
    for (j = 1; j <= blockSize[i]; j++) {

      /* Build the atom to block connection list while we're at it */
      blocksConnected[block[i][j]]++;
      connectedBlockList[block[i][j]][blocksConnected[block[i][j]]] = i;

      /* Now back to the "not connected" scan */
      found = 0;
      for (k = 1; k <= blocks; k++) {
        if (k == i) continue;
        for (l = 1; l <= blockSize[k]; l++) {
          if (block[i][j] == block[k][l]) {
            found = 1;
            break;
          }
        }
        if (found) break;
      } /* next k */
      if (found) {
        blockAtomConnected[i][j] = 1;
      } else {
        blockAtomConnected[i][j] = 0;
        blockConnectedSize[i]--;
      }
    } /* next j */
  } /* next i */

  /* Arrange blocks into a list sorted by "tightness" (clustering)
     to other blocks */
  if (skipClusterSortAlgorithm) {
    /* To bypass algorithm for experimentation, just assign the necessary
       arrays without sorting the blocks */
    for (n = 1; n <= blocks; n++) {
      blockSort[n] = n;
      reverseBlockSort[n] = n;
      /* Create sorted versions of blockSize[], block[][] for speedup */
      sortedBlockSize[n] = blockSize[n];
      for (i = 1; i <= blockSize[n]; i++) {
        sortedBlock[n][i] = block[n][i];
      }
    }
  } else {
    for (n = 1; n <= blocks; n++) {
      reverseBlockSort[n] = 0;
    }
    for (n = 1; n <= blocks; n++) {
      /* In remaining blocks, count the number of connections to blocks
         already in the list.  Put the "best" block (the one most tightly
         coupled to the list so far) next in the sorted list.  The idea
         is to identify infeasible solutions in tight areas more quickly
         and not have to iterate exponentially through long chains of
         blocks. */
      maxBlockConnections = 0;
      /* Worst-case algorithm for speed experiments */
      if (worstCaseAlgorithm) maxBlockConnections = 10000000;
      maxConnectedBlock = 0;
      maxConnectedBlockSize = -1; /* -1 instead of 0 will tolerate blocks
              with no connections (to fix bug 1015) */
      for (i = 1; i <= blocks; i++) {
        thisBlockConnections = 0;
        if (reverseBlockSort[i]) continue; /* Skip blocks already in list */
        for (j = 1; j <= blockSize[i]; j++) {
          if (!blockAtomConnected[i][j]) continue; /* Ignore unconnected atoms*/
          found = 0;
          for (k = 1; k <= blocks; k++) {
            if (k == i) continue;
                /* Ignore same block (actually redundant due to next 'if') */
            if (reverseBlockSort[k] == 0) continue;
                /* Look only at blocks already in list */
            for (l = 1; l <= blockSize[k]; l++) {
              if (block[i][j] == block[k][l]) {
                found = 1;
                break;
              }
            }
            if (found) break;
          } /* next k */
          if (found) thisBlockConnections++;
        } /* next j */
        if (worstCaseAlgorithm) { /* Worst-case algorithm for experiments */
          if (thisBlockConnections < maxBlockConnections
              || (thisBlockConnections == maxBlockConnections
                  && blockConnectedSize[i] <= maxConnectedBlockSize)) {
            /* The criterion for the preferred block to put next in sorted
               listed has been met */
            maxBlockConnections = thisBlockConnections;
            maxConnectedBlockSize = blockConnectedSize[i];
            maxConnectedBlock = i;
          }
        } else {   /* Use normal intended algorithm (best case) */
          if (thisBlockConnections > maxBlockConnections
              || (thisBlockConnections == maxBlockConnections
                  /* 27-Oct-04 nm Changed comparison criteria - seems to
                     reduce average backtracks (based on limited testing) */
                  && (blockConnectedSize[i] > maxConnectedBlockSize ||
                  /* 27-Oct-04 nm Old algorithm can be invoked if desired */
                     (version1_0Algorithm &&
                         blockConnectedSize[i] >= maxConnectedBlockSize)))) {
            /* The criterion for the preferred block to put next in sorted
               listed has been met */
            maxBlockConnections = thisBlockConnections;
            maxConnectedBlockSize = blockConnectedSize[i];
            maxConnectedBlock = i;
          }
        }
      } /* next i */
      if (maxConnectedBlock <= 0) {
        bug(1015);
      }
      /* Add block to sorted list */
      blockSort[n] = maxConnectedBlock;
      reverseBlockSort[maxConnectedBlock] = n;
      /* Create sorted versions of blockSize[], block[][] for speedup */
      sortedBlockSize[n] = blockSize[maxConnectedBlock];
      for (i = 1; i <= blockSize[maxConnectedBlock]; i++) {
        sortedBlock[n][i] = block[maxConnectedBlock][i];
      }

    } /* next n */
  } /* if !skipClusterSortAlgorithm */

  /* Consistency check */
  for (i = 1; i <= blocks; i++) {
    if (blockSort[reverseBlockSort[i]] != i) {
      printf("i = %ld != blockSort[reverseBlockSort[i]] = %ld\n", i,
          blockSort[reverseBlockSort[i]]);
      bug(1016);
    }
    if (reverseBlockSort[blockSort[i]] != i) {
      printf("i = %ld != reverseBlockSort[blockSort[i]] = %ld\n", i,
          reverseBlockSort[blockSort[i]]);
      bug(1018);
    }
  }

  /* Scan the sorted list of blocks to try to assign a state */
  for (i = 1; i <= atoms; i++) {
    atomCommittedBy[i] = 0;
    atomValue[i] = -1; /* -1 value means it is unassigned and available */
  }
  /* Initialize the starting atom to "no previous atoms tried" */
  for (n = 1; n <= blocks; n++) {
    lastAtomTried[n] = 0;
    unconnectedAtomWasTried[n] = 0;
  }
  n = 1;

  if (verboseMode) {
    /* Print header above iteration display */
    printf("     ");
    for (p = 1; p <= blocks; p++) {
      for (q = 1; q <= blockSize[p]; q++) {
        printf("%c", ATOM_MAP[block[p][q] - 1] );
      }
      if (p < blocks) printf(",");
    }
    printf("\n");
    iter = 0; /* Iteration counter */
  }

  while (1) {

    if (verboseMode) {
      /* Print iteration line */
      if (iter > 0) {
        printf("%s", cat(space(4 - strlen(str(iter))), str(iter), " ", NULL));
        for (p = 1; p <= blocks; p++) {
          for (q = 1; q <= blockSize[p]; q++) {
            v = atomValue[block[p][q]];
            printf("%c", v == -1 ? '?' : str(v)[0]);
            let(&tmp,""); /* Purge temp strings to prevent overflow */
          }
          if (p < blocks) printf(",");
        }
        printf("\n");
      }
      iter++; /* Iteration counter */
    }

    if (n > blocks) {
      retVal = 0; /* A state was found */
      break;
    }
    /* Try assigning a value 1 to atoms in the block, until an assignment
       without conflict is found */
    atom1 = 0; /* This is the atom to which the value=1 is assigned, or 0 if
                  no value=1 assignment is possible without conflict */
    for (j = lastAtomTried[n] + 1; j <= sortedBlockSize[n]; j++) {
      if (atomValue[sortedBlock[n][j]] != 0) {
        /* The trial atom j is either already 1 or uncommitted, so we can
           try it */

        /* See if all other atoms in the block are either 0 or uncommitted;
           if not, we have a conflict and we'll try the next j */
        conflict = 0;
        for (k = 1; k <= sortedBlockSize[n]; k++) {
          if (k == j) continue; /* Skip the "1" atom */
          if (atomValue[sortedBlock[n][k]] == 1) {
            /* There's a conflict; try the next j */
            conflict = 1;
            break;
          }
        }
        if (conflict) continue;

        /* Assign the jth atom to 1, and assign the other atoms in the
           block to 0 */
        for (k = 1; k <= sortedBlockSize[n]; k++) {
          if (atomCommittedBy[sortedBlock[n][k]] == 0) {
            atomCommittedBy[sortedBlock[n][k]] = n;
            if (atomValue[sortedBlock[n][k]] != -1) bug(1017);
            if (k == j) {
              atomValue[sortedBlock[n][k]] = 1;
            } else {
              atomValue[sortedBlock[n][k]] = 0;
            }
          } else {
            if (k == j) {
              if (atomValue[sortedBlock[n][k]] != 1) bug(1018);
            } else {
              if (atomValue[sortedBlock[n][k]] != 0) bug(1019);
            }
          }
        }

        /* See if the assignment has caused a conflict */
        conflict = 0;
        for (k = 1; k <= sortedBlockSize[n]; k++) {
          atom = sortedBlock[n][k];
          for (l = 1; l <= blocksConnected[atom]; l++) {
            connectedBlock = connectedBlockList[atom][l];
            onesInBlock = 0;
            unassignedInBlock = 0;
            for (m = 1; m <= blockSize[connectedBlock]; m++) {
              if (atomValue[block[connectedBlock][m]] == 1) {
                onesInBlock++;
              } else {
                if (atomValue[block[connectedBlock][m]] == -1) {
                  unassignedInBlock++;
                }
              }
            }
            if (onesInBlock > 1
                || (onesInBlock == 0 && unassignedInBlock == 0)) {
              /* The assignment caused a conflict - can't use it */
              conflict = 1;
              break;
            }
          }
          if (conflict) break;
        }

        if (conflict) {
          /* If there was a conflict, remove the new assignment */
          for (k = 1; k <= sortedBlockSize[n]; k++) {
            if (atomCommittedBy[sortedBlock[n][k]] == n) {
              atomCommittedBy[sortedBlock[n][k]] = 0;
              atomValue[sortedBlock[n][k]] = -1;
            }
          }
          continue; /* Try the next j */
        } else {  /* No conflict */
          /* We found an uncommitted atom without a conflict, and
             we have assigned it */
          atom1 = j;
          break;
        }
      } /* end if block's jth atom value = 1 or unassigned */
    } /* next j */


    if (atom1 > 0) {
      lastAtomTried[n] = atom1;
      if (!blockAtomConnected[sortedBlock[n][j]]) {
        unconnectedAtomWasTried[n] = 1;
      }
      n++; /* Go to next block in sorted list */
      continue;
    }

    /* We have exhausted possibilities for finding a 1, so we must backtrack */
    retVal = 0;
    if (atomCommittedBy[sortedBlock[n][1]] >= n) bug(1022);
    lastAtomTried[n] = 0;  /* Start over next time around */
    unconnectedAtomWasTried[n] = 0;
    n--;
    backtrackCountx++;
    if (n == 0) {
      retVal = 1; /* No {0,1} state is possible */
      break;
    }
    backtrackAgain = 0;
    for (j = 1; j <= sortedBlockSize[n]; j++) {
      if (atomCommittedBy[sortedBlock[n][j]] == 0) bug(1023);
      if (atomCommittedBy[sortedBlock[n][j]] == n) {
        /* Uncommit any values assigned by previous block */
        atomCommittedBy[sortedBlock[n][j]] = 0;
        atomValue[sortedBlock[n][j]] = -1;
      } else {
        if (atomCommittedBy[sortedBlock[n][j]] > n) bug(1024);
      }
    }
    if (retVal == 1) break;
  } /* while 1 */

  /* backtrackCount is for informational purposes */
  /*if (!oneLineDisplay) printf("Backtrack count = %ld\n", backtrackCountx);*/
  *backtrackCount = backtrackCountx;  /* return argument */

  return retVal;  /* 0 if {0,1} state found, 1 if not */
} /* state01Test */


/******************* End of main program ********************************/


/***********************************************************************/
/************ Start of "vstring" body stuff ****************************/
/************ Do not touch anything from here to end of program ********/
/***********************************************************************/

/*****************************************************************************/
/*       Copyright (C) 1999  NORMAN D. MEGILL  <nm at alum.mit.edu>          */
/*             License terms:  GNU General Public License                    */
/*****************************************************************************/

/**************************************************************************

Variable-length string handler
------------------------------

     This collection of string-handling functions emulate most of the
string functions of VMS BASIC.  The objects manipulated by these functions
are strings of a special type called 'vstring' which
have no pre-defined upper length limit but are dynamically allocated
and deallocated as needed.  To use the vstring functions within a program,
all vstrings must be initially set to the null string when declared or
before first used, for example:

        vstring string1 = "";
        vstring stringArray[] = {"","",""};

        vstring bigArray[100][10]; /- Must be initialized before using -/
        int i,j;
        for (i=0; i<100; i++)
          for (j=0; j<10; j++)
            bigArray[i][j] = ""; /- Initialize -/


     After initialization, vstrings should be assigned with the 'let(&'
function only; for example the statements

        let(&string1,"abc");
        let(&string1,string2);
        let(&string1,left(string2,3));

all assign the second argument to 'string1'.  The 'let(&' function must
not be used to initialize a vstring for the first time.

     The 'cat' function emulates the '+' concatenation operator in BASIC.
It has a variable number of arguments, and the last argument should always
be NULL.  For example,

        let(&string1,cat("abc","def",NULL));

assigns "abcdef" to 'string1'.  Warning: 0 will work instead of NULL on the
VAX but not on the Macintosh, so always use NULL.

     All other functions are generally used exactly like their BASIC
equivalents.  For example, the BASIC statement

        let string1$=left$("def",len(right$("xxx",2)))+"ghi"+string2$

is emulated in c as

        let(&string1,cat(left("def",len(right("xxx",2))),"ghi",string2,NULL));

Note that ANSI c does not allow "$" as part of an identifier
name, so the names in c have had the "$" suffix removed.

     The string arguments of the vstring functions may be either standard c
strings or vstrings (except that the first argument of the 'let(&' function
must be a vstring).  The standard c string functions may use vstrings or
vstring functions as their string arguments, as long as the vstring variable
itself (which is a char * pointer) is not modified and no attempt is made to
increase the length of a vstring.  Caution must be excercised when
assigning standard c string pointers to vstrings or the results of
vstring functions, as the memory space may be deallocated when the
'le(&t' function is next executed.  For example,

        char *stdstr; /- A standard c string pointer -/
         ...
        stdstr=left("abc",2);  /- DO NOT DO THIS -/

will assign "ab" to 'stdstr', but this assignment will be lost when the
next 'let(&' function is executed.  To be safe, use 'strcpy':

        char stdstr1[80]; /- A fixed length standard c string -/
         ...
        strcpy(stdstr1,left("abc",2));

Here, of course, the user must ensure that the string copied to 'stdstr1'
does not exceed 79 characters in length.  IT IS SAFEST NOT TO USE ANY
STANDARD C STRING FUNCTIONS WITH VSTRINGS OR VSTRING FUNCTIONS UNLESS YOU
REALLY UNDERSTAND WHAT YOU ARE DOING.

     The vstring functions ('left', 'right', 'cat', etc.) allocate temporary
memory whenever they are called.  This temporary memory is deallocated
whenever a 'let(&' assignment is made.  The user should be aware of this
when using vstring functions outside of 'let(&' assignments; for example

        for (i=0; i<10000; i++)
          print2("%s\n",left(string1,70));

will allocate another 70 bytes or so of memory each 'left' call
and eventually overflow the temporary string stack.
If necessary, dummy 'let(&' assignments can be made periodically to clear
this temporary memory:

        for (i=0; i<10000; i++)
          {
          print2("%s\n",left(string1,70));
          let(&dummy,"");
          }

It should be noted that the 'linput' function assigns its target string
with 'let(&' and thus has the same deallocation effect as 'let(&'.

************************************************************************/


vstring tempAlloc(long size)    /* String memory allocation/deallocation */
{
  /* When "size" is >0, "size" bytes are allocated. */
  /* When "size" is 0, all memory previously allocated with this */
  /* function is deallocated. */
  /* EXCEPT:  When startTempAllocStack != 0, the freeing will start at
     startTempAllocStack. */
  int i;
  if (size) {
    if (tempAllocStackTop>=(MAX_ALLOC_STACK-1)) {
      print2("?Error: Temporary string stack overflow\n");
      bug(101);
    }
    if (!(tempAllocStack[tempAllocStackTop++]=malloc(size))) {
      print2("?Error: Temporary string allocation failed\n");
      bug(102);
    }
    return (tempAllocStack[tempAllocStackTop-1]);
  } else {
    for (i=startTempAllocStack; i<tempAllocStackTop; i++) {
      free(tempAllocStack[i]);
    }
    tempAllocStackTop=startTempAllocStack;
    return (NULL);
  }
}


/* Make string have temporary allocation to be released by next let() */
/* Warning:  after makeTempAlloc() is called, the genString may NOT be
   assigned again with let() */
void makeTempAlloc(vstring s)
{
    if (tempAllocStackTop>=(MAX_ALLOC_STACK-1)) {
      print2("?Error: Temporary string stack overflow\n");
      bug(103);
    }
    tempAllocStack[tempAllocStackTop++]=s;
}


void let(vstring *target,vstring source)        /* String assignment */
/* This function must ALWAYS be called to make assignment to */
/* a vstring in order for the memory cleanup routines, etc. */
/* to work properly.  If a vstring has never been assigned before, */
/* it is the user's responsibility to initialize it to "" (the */
/* null string). */
{
  long targetLength,sourceLength;

  sourceLength=strlen(source);  /* Save its length */
  targetLength=strlen(*target); /* Save its length */
  if (targetLength) {
    if (sourceLength) { /* source and target are both nonzero length */

      if (targetLength>=sourceLength) { /* Old string has room for new one */
        strcpy(*target,source); /* Re-use the old space to save CPU time */
      } else {
        /* Free old string space and allocate new space */
        free(*target);  /* Free old space */
        *target=malloc(sourceLength+1); /* Allocate new space */
        if (!*target) {
          print2("?Error: String memory couldn't be allocated\n");
          bug(104);
        }
        strcpy(*target,source);
      }

    } else {    /* source is 0 length, target is not */
      free(*target);
      *target= "";
    }
  } else {
    if (sourceLength) { /* target is 0 length, source is not */
      *target=malloc(sourceLength+1);   /* Allocate new space */
      if (!*target) {
        print2("?Error: Could not allocate string memory\n");
        bug(105);
      }
      strcpy(*target,source);
    } else {    /* source and target are both 0 length */
      *target= "";
    }
  }

  tempAlloc(0); /* Free up temporary strings used in expression computation */

}




vstring cat(vstring string1,...)        /* String concatenation */
#define MAX_CAT_ARGS 30
{
  va_list ap;   /* Declare list incrementer */
  vstring arg[MAX_CAT_ARGS];    /* Array to store arguments */
  long argLength[MAX_CAT_ARGS]; /* Array to store argument lengths */
  int numArgs=1;        /* Define "last argument" */
  int i;
  long j;
  vstring ptr;

  arg[0]=string1;       /* First argument */

  va_start(ap,string1); /* Begin the session */
  while ((arg[numArgs++]=va_arg(ap,char *)))
        /* User-provided argument list must terminate with 0 */
    if (numArgs>=MAX_CAT_ARGS-1) {
      print2("?Error: Too many cat() arguments\n");
      bug(106);
    }
  va_end(ap);           /* End var args session */

  numArgs--;    /* The last argument (0) is not a string */

  /* Find out the total string length needed */
  j=0;
  for (i=0; i<numArgs; i++) {
    argLength[i]=strlen(arg[i]);
    j=j+argLength[i];
  }
  /* Allocate the memory for it */
  ptr=tempAlloc(j+1);
  /* Move the strings into the newly allocated area */
  j=0;
  for (i=0; i<numArgs; i++) {
    strcpy(ptr+j,arg[i]);
    j=j+argLength[i];
  }
  return (ptr);

}


/* input a line from the user or from a file */
vstring linput(FILE *stream,vstring ask,vstring *target)
{
  /*
    BASIC:  linput "what";a$
    c:      linput(NULL,"what?",&a);

    BASIC:  linput #1,a$                        (error trap on EOF)
    c:      if (!linput(file1,NULL,&a)) break;  (break on EOF)

  */
  /* This function prints a prompt (if 'ask' is not NULL), gets a line from
    the stream, and assigns it to target using the let(&...) function.
    NULL is returned when end-of-file is encountered.  The vstring
    *target MUST be initialized to "" or previously assigned by let(&...)
    before using it in linput. */
  char f[10001]; /* Allow up to 10000 characters */
  if (ask) print2("%s",ask);
  if (stream == NULL) stream = stdin;
  if (!fgets(f,10000,stream)) {
    /* End of file */
    return NULL;
  }
  f[10000]=0;     /* Just in case */
  f[strlen(f)-1]=0;     /* Eliminate new-line character */
  /* Assign the user's input line */
  let(target,f);
  return *target;
}


/* Find out the length of a string */
long len(vstring s)
{
  return (strlen(s));
}


/* Extract sin from character position start to stop into sout */
vstring seg(vstring sin,long start,long stop)
{
  vstring sout;
  long len;
  if (start<1) start=1;
  if (stop<1) stop=0;
  len=stop-start+1;
  if (len<0) len=0;
  sout=tempAlloc(len+1);
  strncpy(sout,sin+start-1,len);
  sout[len]=0;
  return (sout);
}

/* Extract sin from character position start for length len */
vstring mid(vstring sin,long start,long len)
{
  vstring sout;
  if (start<1) start=1;
  if (len<0) len=0;
  sout=tempAlloc(len+1);
  strncpy(sout,sin+start-1,len);
/*??? Should db be substracted from if len > end of string? */
  sout[len]=0;
  return (sout);
}

/* Extract leftmost n characters */
vstring left(vstring sin,long n)
{
  vstring sout;
  if (n < 0) n = 0;
  sout=tempAlloc(n+1);
  strncpy(sout,sin,n);
  sout[n]=0;
  return (sout);
}

/* Extract after character n */
vstring right(vstring sin,long n)
{
  /*??? We could just return &sin[n-1], but this is safer for debugging. */
  vstring sout;
  long i;
  if (n<1) n=1;
  i = strlen(sin);
  if (n>i) return ("");
  sout = tempAlloc(i - n + 2);
  strcpy(sout,&sin[n-1]);
  return (sout);
}

/* Emulate VMS BASIC edit$ command */
vstring edit(vstring sin,long control)
#define isblank_(c) ((c==' ') || (c=='\t'))
    /* 11-Sep-2009 nm Added _ to fix '"isblank" redefined' compiler warning */
{
/*
EDIT$
  Syntax
         str-vbl = EDIT$(str-exp, int-exp)
     Values   Effect
     1        Trim parity bits
     2        Discard all spaces and tabs
     4        Discard characters: CR, LF, FF, ESC, RUBOUT, and NULL
     8        Discard leading spaces and tabs
     16       Reduce spaces and tabs to one space
     32       Convert lowercase to uppercase
     64       Convert [ to ( and ] to )
     128      Discard trailing spaces and tabs
     256      Do not alter characters inside quotes

     (non-BASIC extensions)
     512      Convert uppercase to lowercase
     1024     Tab the line (convert spaces to equivalent tabs)
     2048     Untab the line (convert tabs to equivalent spaces)
     4096     Convert VT220 screen print frame graphics to -,|,+ characters

     (Added 10/24/03:)
     8192     Discard CR only (to assist DOS-to-Unix conversion)
*/
  vstring sout;
  long i,j,k;
  int last_char_is_blank;
  int trim_flag,discardctrl_flag,bracket_flag,quote_flag,case_flag;
  int alldiscard_flag,leaddiscard_flag,traildiscard_flag,reduce_flag;
  int processing_inside_quote=0;
  int lowercase_flag, tab_flag, untab_flag, screen_flag, discardcr_flag;
  unsigned char graphicsChar;

  /* Set up the flags */
  trim_flag=control & 1;
  alldiscard_flag=control & 2;
  discardctrl_flag=control & 4;
  leaddiscard_flag=control & 8;
  reduce_flag=control & 16;
  case_flag=control & 32;
  bracket_flag=control & 64;
  traildiscard_flag=control & 128;
  quote_flag=control & 256;

  /* Non-BASIC extensions */
  lowercase_flag = control & 512;
  tab_flag = control & 1024;
  untab_flag = control & 2048;
  screen_flag = control & 4096; /* Convert VT220 screen prints to |,-,+
                                   format */
  discardcr_flag=control & 8192; /* Discard CR's */

  /* Copy string */
  i = strlen(sin) + 1;
  if (untab_flag) i = i * 7;
  sout=tempAlloc(i);
  strcpy(sout,sin);

  /* Discard leading space/tab */
  i=0;
  if (leaddiscard_flag)
    while ((sout[i]!=0) && isblank_(sout[i]))
      sout[i++]=0;

  /* Main processing loop */
  while (sout[i]!=0) {

    /* Alter characters inside quotes ? */
    if (quote_flag && ((sout[i]=='"') || (sout[i]=='\'')))
       processing_inside_quote=~processing_inside_quote;
    if (processing_inside_quote) {
       /* Skip the rest of the code and continue to process next character */
       i++; continue;
    }

    /* Discard all space/tab */
    if ((alldiscard_flag) && isblank_(sout[i]))
        sout[i]=0;

    /* Trim parity (eighth?) bit */
    if (trim_flag)
       sout[i]=sout[i] & 0x7F;

    /* Discard CR,LF,FF,ESC,BS */
    if ((discardctrl_flag) && (
         (sout[i]=='\015') || /* CR  */
         (sout[i]=='\012') || /* LF  */
         (sout[i]=='\014') || /* FF  */
         (sout[i]=='\033') || /* ESC */
         /*(sout[i]=='\032') ||*/ /* ^Z */ /* DIFFERENCE won't work w/ this */
         (sout[i]=='\010')))  /* BS  */
      sout[i]=0;

    /* Discard CR */
    if ((discardcr_flag) && (
         (sout[i]=='\015')))  /* CR  */
      sout[i]=0;

    /* Convert lowercase to uppercase */
    /*
    if ((case_flag) && (islower(sout[i])))
       sout[i]=toupper(sout[i]);
    */
    /* 13-Jun-2009 nm The upper/lower case C functions have odd behavior
       with characters > 127, at least in lcc.  So this was rewritten to
       not use them. */
    if ((case_flag) && (sout[i] >= 'a' && sout[i] <= 'z'))
       sout[i]=sout[i] - ('a' - 'A');

    /* Convert [] to () */
    if ((bracket_flag) && (sout[i]=='['))
       sout[i]='(';
    if ((bracket_flag) && (sout[i]==']'))
       sout[i]=')';

    /* Convert uppercase to lowercase */
    /*
    if ((lowercase_flag) && (isupper(sout[i])))
       sout[i]=tolower(sout[i]);
    */
    /* 13-Jun-2009 nm The upper/lower case C functions have odd behavior
       with characters > 127, at least in lcc.  So this was rewritten to
       not use them. */
    if ((lowercase_flag) && (sout[i] >= 'A' && sout[i] <= 'Z'))
       sout[i]=sout[i] + ('a' - 'A');

    /* Convert VT220 screen print frame graphics to +,|,- */
    if (screen_flag) {
      graphicsChar = sout[i]; /* Need unsigned char for >127 */
      /* vt220 */
      if (graphicsChar >= 234 && graphicsChar <= 237) sout[i] = '+';
      if (graphicsChar == 241) sout[i] = '-';
      if (graphicsChar == 248) sout[i] = '|';
      if (graphicsChar == 166) sout[i] = '|';
      /* vt100 */
      if (graphicsChar == 218 /*up left*/ || graphicsChar == 217 /*lo r*/
          || graphicsChar == 191 /*up r*/ || graphicsChar == 192 /*lo l*/)
        sout[i] = '+';
      if (graphicsChar == 196) sout[i] = '-';
      if (graphicsChar == 179) sout[i] = '|';
    }

    /* Process next character */
    i++;
  }
  /* sout[i]=0 is the last character at this point */

  /* Clean up the deleted characters */
  for (j=0,k=0; j<=i; j++)
    if (sout[j]!=0) sout[k++]=sout[j];
  sout[k]=0;
  /* sout[k]=0 is the last character at this point */

  /* Discard trailing space/tab */
  if (traildiscard_flag) {
    --k;
    while ((k>=0) && isblank_(sout[k])) --k;
    sout[++k]=0;
  }

  /* Reduce multiple space/tab to a single space */
  if (reduce_flag) {
    i=j=last_char_is_blank=0;
    while (i<=k-1) {
      if (!isblank_(sout[i])) {
        sout[j++]=sout[i++];
        last_char_is_blank=0;
      } else {
        if (!last_char_is_blank)
          sout[j++]=' '; /* Insert a space at the first occurrence of a blank */
        last_char_is_blank=1; /* Register that a blank is found */
        i++; /* Process next character */
      }
    }
    sout[j]=0;
  }

  /* Untab the line */
  if (untab_flag || tab_flag) {

    /*
    DEF FNUNTAB$(L$)      ! UNTAB LINE L$
    I9%=1%
    I9%=INSTR(I9%,L$,CHR$(9%))
    WHILE I9%
      L$=LEFT(L$,I9%-1%)+SPACE$(8%-((I9%-1%) AND 7%))+RIGHT(L$,I9%+1%)
      I9%=INSTR(I9%,L$,CHR$(9%))
    NEXT
    FNUNTAB$=L$
    FNEND
    */

    k = strlen(sout);
    for (i = 1; i <= k; i++) {
      if (sout[i - 1] != '\t') continue;
      for (j = k; j >= i; j--) {
        sout[j + 8 - ((i - 1) & 7) - 1] = sout[j];
      }
      for (j = i; j < i + 8 - ((i - 1) & 7); j++) {
        sout[j - 1] = ' ';
      }
      k = k + 8 - ((i - 1) & 7);
    }
  }

  /* Tab the line */
  if (tab_flag) {

    /*
    DEF FNTAB$(L$)        ! TAB LINE L$
    I9%=0%
    FOR I9%=8% STEP 8% WHILE I9%<LEN(L$)
      J9%=I9%
      J9%=J9%-1% UNTIL ASCII(MID(L$,J9%,1%))<>32% OR J9%=I9%-8%
      IF J9%<=I9%-2% THEN
        L$=LEFT(L$,J9%)+CHR$(9%)+RIGHT(L$,I9%+1%)
        I9%=J9%+1%
      END IF
    NEXT I9%
    FNTAB$=L$
    FNEND
    */

    i = 0;
    k = strlen(sout);
    for (i = 8; i < k; i = i + 8) {
      j = i;
      while (sout[j - 1] == ' ' && j > i - 8) j--;
      if (j <= i - 2) {
        sout[j] = '\t';
        j = i;
        while (sout[j - 1] == ' ' && j > i - 8 + 1) {
          sout[j - 1] = 0;
          j--;
        }
      }
    }
    i = k;
    /* sout[i]=0 is the last character at this point */
    /* Clean up the deleted characters */
    for (j = 0, k = 0; j <= i; j++)
      if (sout[j] != 0) sout[k++] = sout[j];
    sout[k] = 0;
    /* sout[k]=0 is the last character at this point */
  }

  return (sout);
}


/* Return a string of the same character */
vstring string(long n, char c)
{
  vstring sout;
  long j=0;
  if (n<0) n=0;
  sout=tempAlloc(n+1);
  while (j<n) sout[j++]=c;
  sout[j]=0;
  return (sout);
}


/* Return a string of spaces */
vstring space(long n)
{
  return (string(n,' '));
}


/* Return a character given its ASCII value */
vstring chr(long n)
{
  vstring sout;
  sout=tempAlloc(2);
  sout[0]= n & 0xFF;
  sout[1]=0;
  return(sout);
}


/* Search for string2 in string 1 starting at start_position */
long instr(long start_position,vstring string1,vstring string2)
{
   char *sp1,*sp2;
   long ls1,ls2;
   long found=0;
   if (start_position<1) start_position=1;
   ls1=strlen(string1);
   ls2=strlen(string2);
   if (start_position>ls1) start_position=ls1+1;
   sp1=string1+start_position-1;
   while ((sp2=strchr(sp1,string2[0]))!=0) {
     if (strncmp(sp2,string2,ls2)==0) {
        found=sp2-string1+1;
        break;
     } else
        sp1=sp2+1;
   }
   return (found);
}


/* Translate string in sin to sout based on table.
   Table must be 256 characters long!! <- not true anymore? */
vstring xlate(vstring sin,vstring table)
{
  vstring sout;
  long len_table,len_sin;
  long i,j;
  long table_entry;
  char m;
  len_sin=strlen(sin);
  len_table=strlen(table);
  sout=tempAlloc(len_sin+1);
  for (i=j=0; i<len_sin; i++)
  {
    table_entry= 0x000000FF & (long)sin[i];
    if (table_entry<len_table)
      if ((m=table[table_entry])!='\0')
        sout[j++]=m;
  }
  sout[j]='\0';
  return (sout);
}


/* Returns the ascii value of a character */
long ascii_(vstring c)
{
  return (long)((unsigned char)(c[0]));
}

/* Returns the floating-point value of a numeric string */
double val(vstring s)
{
  /*
  return (atof(s));
  */
  /* 12/10/98 - NDM - atof may corrupt memory when processing
     random character strings.
     The implementation below makes best guess of value of any
     random string, ignoring commas, etc. and tolerating numbers
     suffixed with sign.  "E" notation is not handled. */
  double v = 0;
  char signFound = 0;
  double power = 1.0;
  long i;
  /* Scan from lsd backwards to minimize rounding errors */
  for (i = strlen(s) - 1; i >= 0; i--) {
    switch (s[i]) {
      case '0': case '1': case '2': case '3': case '4':
      case '5': case '6': case '7': case '8': case '9':
        v = v + ((double)(s[i] - '0')) * power;
        power = 10.0 * power;
        break;
      case '.':
        v = v / power;
        power = 1.0;
        break;
      case '-':
        signFound = 1;
        break;
    }
  }
  if (signFound) v = - v;
  return v;
}


/* Returns current date as an ASCII string */
vstring date()
{
        vstring sout;
        struct tm *time_structure;
        time_t time_val;
        char *month[12];

        /* (Aggregrate initialization is not portable) */
        /* (It must be done explicitly for portability) */
        month[0]="Jan";
        month[1]="Feb";
        month[2]="Mar";
        month[3]="Apr";
        month[4]="May";
        month[5]="Jun";
        month[6]="Jul";
        month[7]="Aug";
        month[8]="Sep";
        month[9]="Oct";
        month[10]="Nov";
        month[11]="Dec";

        time(&time_val);                        /* Retrieve time */
        time_structure=localtime(&time_val); /* Translate to time structure */
        sout=tempAlloc(12);
        sprintf(sout,"%d-%s-%d",
                time_structure->tm_mday,
                month[time_structure->tm_mon],
                time_structure->tm_year);
        return(sout);
}

/* Return current time as an ASCII string */
vstring time_()
{
        vstring sout;
        struct tm *time_structure;
        time_t time_val;
        int i;
        char *format;
        char *format1="%d:%d %s";
        char *format2="%d:0%d %s";
        char *am_pm[2];
        /* (Aggregrate initialization is not portable) */
        /* (It must be done explicitly for portability) */
        am_pm[0]="AM";
        am_pm[1]="PM";

        time(&time_val);                        /* Retrieve time */
        time_structure=localtime(&time_val); /* Translate to time structure */
        if (time_structure->tm_hour>=12) i=1;
        else                             i=0;
        if (time_structure->tm_hour>12) time_structure->tm_hour-=12;
        if (time_structure->tm_hour==0) time_structure->tm_hour=12;
        sout=tempAlloc(12);
        if (time_structure->tm_min>=10)
          format=format1;
        else
          format=format2;
        sprintf(sout,format,
                time_structure->tm_hour,
                time_structure->tm_min,
                am_pm[i]);
        return(sout);

}


/* Return a number as an ASCII string */
vstring str(double f)
{
  /* This function converts a floating point number to a string in the */
  /* same way that %f in printf does, except that trailing zeroes after */
  /* the one after the decimal point are stripped; e.g., it returns 7 */
  /* instead of 7.000000000000000. */
  vstring s;
  long i;
  s = tempAlloc(50);
  sprintf(s, "%f", f);
  if (strchr(s, '.') != 0) {              /* the string has a period in it */
    for (i = strlen(s) - 1; i > 0; i--) { /* scan string backwards */
      if (s[i] != '0') break;             /* 1st non-zero digit */
      s[i] = 0;                           /* delete the trailing 0 */
    }
    if (s[i] == '.') s[i] = 0;            /* delete trailing period */
  }
  return (s);
}


/* Return a number as an ASCII string */
vstring num1(double f)
{
  return (str(f));
}


/* Return a number as an ASCII string surrounded by spaces */
vstring num(double f)
{
  return (cat(" ",str(f)," ",NULL));
}


/*** NEW FUNCTIONS ADDED 11/25/98 ***/

/* Emulate PROGRESS "entry" and related string functions */
/* (PROGRESS is a 4-GL database language) */

/* A "list" is a string of comma-separated elements.  Example:
   "a,b,c" has 3 elements.  "a,b,c," has 4 elements; the last element is
   an empty string.  ",," has 3 elements; each is an empty string.
   In "a,b,c", the entry numbers of the elements are 1, 2 and 3 (i.e.
   the entry numbers start a 1, not 0). */

/* Returns a character string entry from a comma-separated
   list based on an integer position. */
/* If element is less than 1 or greater than number of elements
   in the list, a null string is returned. */
vstring entry(long element, vstring list)
{
  vstring sout;
  long commaCount, lastComma, i, len;
  if (element < 1) return ("");
  lastComma = -1;
  commaCount = 0;
  i = 0;
  while (list[i] != 0) {
    if (list[i] == ',') {
      commaCount++;
      if (commaCount == element) {
        break;
      }
      lastComma = i;
    }
    i++;
  }
  if (list[i] == 0) commaCount++;
  if (element > commaCount) return ("");
  len = i - lastComma - 1;
  if (len < 1) return ("");
  sout = tempAlloc(len + 1);
  strncpy(sout, list + lastComma + 1, len);
  sout[len] = 0;
  return (sout);
}

/* Emulate PROGRESS lookup function */
/* Returns an integer giving the first position of an expression
   in a comma-separated list. Returns a 0 if the expression
   is not in the list. */
long lookup(vstring expression, vstring list)
{
  long i, exprNum, exprPos;
  char match;

  match = 1;
  i = 0;
  exprNum = 0;
  exprPos = 0;
  while (list[i] != 0) {
    if (list[i] == ',') {
      exprNum++;
      if (match) {
        if (expression[exprPos] == 0) return exprNum;
      }
      exprPos = 0;
      match = 1;
      i++;
      continue;
    }
    if (match) {
      if (expression[exprPos] != list[i]) match = 0;
    }
    i++;
    exprPos++;
  }
  exprNum++;
  if (match) {
    if (expression[exprPos] == 0) return exprNum;
  }
  return 0;
}


/* Emulate PROGRESS num-entries function */
/* Returns the number of items in a comma-separated list. */
long numEntries(vstring list)
{
  long i, commaCount;
  i = 0;
  commaCount = 0;
  while (list[i] != 0) {
    if (list[i] == ',') commaCount++;
    i++;
  }
  return (commaCount + 1);
}

/* Returns the character position of the start of the
   element in a list - useful for manipulating
   the list string directly.  1 means the first string
   character. */
/* If element is less than 1 or greater than number of elements
   in the list, a 0 is returned.  If entry is null, a 0 is
   returned. */
long entryPosition(long element, vstring list)
{
  long commaCount, lastComma, i;
  if (element < 1) return 0;
  lastComma = -1;
  commaCount = 0;
  i = 0;
  while (list[i] != 0) {
    if (list[i] == ',') {
      commaCount++;
      if (commaCount == element) {
        break;
      }
      lastComma = i;
    }
    i++;
  }
  if (list[i] == 0) {
    if (i == 0) return 0;
    if (list[i - 1] == ',') return 0;
    commaCount++;
  }
  if (element > commaCount) return (0);
  if (list[lastComma + 1] == ',') return 0;
  return (lastComma + 2);
}


void print2(char* fmt,...)
{
  /* This performs the same operations as printf, except that if a log file is
    open, the characters will also be printed to the log file. */
  va_list ap;
  char printBuffer[10001];

  va_start(ap, fmt);
  vsprintf(printBuffer, fmt, ap); /* Put formatted string into buffer */
  va_end(ap);

  printf("%s", printBuffer); /* Terminal */

  if (fplog != NULL) {
    fprintf(fplog, "%s", printBuffer);  /* Print to log file */
  }
  return;
}


/* Bug check */
void bug(int bugNum)
{
  print2("?Error: Program bug # %d\n", bugNum);
  exit(0);
}


/* Opens files with error message; opens output files with
   backup of previous version.   Mode must be "r" or "w". */
FILE *fSafeOpen(vstring fileName, vstring mode)
{
  FILE *fp;
  vstring prefix = "";
  vstring postfix = "";
  vstring bakName = "";
  vstring newBakName = "";
  long v;

  if (!strcmp(mode, "r")) {
    fp = fopen(fileName, "r");
    if (!fp) {
      print2("?Sorry, couldn't open the file \"%s\".\n", fileName);
    }
    return (fp);
  }

  if (!strcmp(mode, "w")) {
    /* See if the file already exists. */
    fp = fopen(fileName, "r");

    if (fp) {
      fclose(fp);

#define VERSIONS 9
      /* The file exists.  Rename it. */

#if defined __WATCOMC__ /* MSDOS */
      /* Make sure file name before extension is 8 chars or less */
      i = instr(1, fileName, ".");
      if (i) {
        let(&prefix, left(fileName, i - 1));
        let(&postfix, right(fileName, i));
      } else {
        let(&prefix, fileName);
        let(&postfix, "");
      }
      let(&prefix, cat(left(prefix, 5), "~", NULL));
      let(&postfix, cat("~", postfix, NULL));
      if (0) goto skip_backup; /* Prevent compiler warning */

#elif defined __GNUC__ /* Assume unix */
      let(&prefix, cat(fileName, "~", NULL));
      let(&postfix, "");

#elif defined THINK_C /* Assume Macintosh */
      let(&prefix, cat(fileName, "~", NULL));
      let(&postfix, "");

#elif defined VAXC /* Assume VMS */
      /* For debugging on VMS: */
      /* let(&prefix, cat(fileName, "-", NULL));
         let(&postfix, "-"); */
      /* Normal: */
      goto skip_backup;

#else /* Unknown; assume unix standard */
      /*if (1) goto skip_backup;*/  /* [if no backup desired] */
      let(&prefix, cat(fileName, "~", NULL));
      let(&postfix, "");

#endif


      /* See if the lowest version already exists. */
      let(&bakName, cat(prefix, str(1), postfix, NULL));
      fp = fopen(bakName, "r");
      if (fp) {
        fclose(fp);
        /* The lowest version already exists; rename all to lower versions. */

        /* If version VERSIONS exists, delete it. */
        let(&bakName, cat(prefix, str(VERSIONS), postfix, NULL));
        fp = fopen(bakName, "r");
        if (fp) {
          fclose(fp);
          remove(bakName);
        }

        for (v = VERSIONS - 1; v >= 1; v--) {
          let(&bakName, cat(prefix, str(v), postfix, NULL));
          fp = fopen(bakName, "r");
          if (!fp) continue;
          fclose(fp);
          let(&newBakName, cat(prefix, str(v + 1), postfix, NULL));
          rename(bakName, newBakName);
        }

      }
      let(&bakName, cat(prefix, str(1), postfix, NULL));
      rename(fileName, bakName);

      /***
      printLongLine(cat("The file \"", fileName,
          "\" already exists.  The old file is being renamed to \"",
          bakName, "\".", NULL), "  ", " ");
      ***/
    } /* End if file already exists */
   /*skip_backup:*/

    fp = fopen(fileName, "w");
    if (!fp) {
      print2("?Sorry, couldn't open the file \"%s\".\n", fileName);
    }

    let(&prefix, "");
    let(&postfix, "");
    let(&bakName, "");
    let(&newBakName, "");

    return (fp);
  } /* End if mode = "w" */

  bug(1510); /* Illegal mode */
  return(NULL);

}

/***********************************************************************/
/************ End of "vstring" body stuff ******************************/
/***********************************************************************/
